/*
 *  This code is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  This code is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
*/

#include "stdafx.h"
#include "IFOUtils.h"
#include "Utils.h"
#include "Cell.h"
#include "io.h"

char *IFONames[100] = {"VIDEO_TS.IFO","VTS_01_0.IFO","VTS_02_0.IFO","VTS_03_0.IFO","VTS_04_0.IFO","VTS_05_0.IFO","VTS_06_0.IFO","VTS_07_0.IFO","VTS_08_0.IFO","VTS_09_0.IFO","VTS_10_0.IFO","VTS_11_0.IFO","VTS_12_0.IFO","VTS_13_0.IFO","VTS_14_0.IFO","VTS_15_0.IFO","VTS_16_0.IFO","VTS_17_0.IFO","VTS_18_0.IFO","VTS_19_0.IFO","VTS_20_0.IFO","VTS_21_0.IFO","VTS_22_0.IFO","VTS_23_0.IFO","VTS_24_0.IFO","VTS_25_0.IFO","VTS_26_0.IFO","VTS_27_0.IFO","VTS_28_0.IFO","VTS_29_0.IFO","VTS_30_0.IFO","VTS_31_0.IFO","VTS_32_0.IFO","VTS_33_0.IFO","VTS_34_0.IFO","VTS_35_0.IFO","VTS_36_0.IFO","VTS_37_0.IFO","VTS_38_0.IFO","VTS_39_0.IFO","VTS_40_0.IFO","VTS_41_0.IFO","VTS_42_0.IFO","VTS_43_0.IFO","VTS_44_0.IFO","VTS_45_0.IFO","VTS_46_0.IFO","VTS_47_0.IFO","VTS_48_0.IFO","VTS_49_0.IFO","VTS_50_0.IFO","VTS_51_0.IFO","VTS_52_0.IFO","VTS_53_0.IFO","VTS_54_0.IFO","VTS_55_0.IFO","VTS_56_0.IFO","VTS_57_0.IFO","VTS_58_0.IFO","VTS_59_0.IFO","VTS_60_0.IFO","VTS_61_0.IFO","VTS_62_0.IFO","VTS_63_0.IFO","VTS_64_0.IFO","VTS_65_0.IFO","VTS_66_0.IFO","VTS_67_0.IFO","VTS_68_0.IFO","VTS_69_0.IFO","VTS_70_0.IFO","VTS_71_0.IFO","VTS_72_0.IFO","VTS_73_0.IFO","VTS_74_0.IFO","VTS_75_0.IFO","VTS_76_0.IFO","VTS_77_0.IFO","VTS_78_0.IFO","VTS_79_0.IFO","VTS_80_0.IFO","VTS_81_0.IFO","VTS_82_0.IFO","VTS_83_0.IFO","VTS_84_0.IFO","VTS_85_0.IFO","VTS_86_0.IFO","VTS_87_0.IFO","VTS_88_0.IFO","VTS_89_0.IFO","VTS_90_0.IFO","VTS_91_0.IFO","VTS_92_0.IFO","VTS_93_0.IFO","VTS_94_0.IFO","VTS_95_0.IFO","VTS_96_0.IFO","VTS_97_0.IFO","VTS_98_0.IFO","VTS_99_0.IFO"};

#define MAX_IFO_SIZE_INCREASE	100		// Maximum number of LBAs, the IFO size can be increased.
#define MAX_NUM_VOBUS 64000

// Used to parse commands...
#define CMD (OpCode & 0x000F)
#define TYP (OpCode >> 13)
#define DIR ((OpCode >> 12) & 0x0001)


CIFOLoader::CIFOLoader(int VTS)
{
	m_Buffer = NULL;
	m_IFOfd = NULL;
	m_NumVOBUs = 0;
	m_LoadedOK = 0;
	m_VOBUStartLBA = new uint[MAX_NUM_VOBUS];
	LoadIFO(VTS);
}

// These are safer versions than the ones in Cell.cpp (which are called here). I check that the address isn't outside
// of the boundaries of the IFO size. 
void CIFOLoader::PutInt(uchar* Buffer, int Offset, uint Value)
{
	if(Buffer+Offset < m_MinAddress || Buffer+Offset+sizeof(int) > m_MaxAddress) 
	{
		MESSAGE4("Out-of-bound address in PutInt %d %d %d",Buffer+Offset,m_MinAddress,m_MaxAddress);
		m_AddressError = true;
		return;
	}
	::PutInt(Buffer,Offset,Value);
}

void CIFOLoader::PutShort(uchar* Buffer, int Offset, ushort Value)
{
	if(Buffer+Offset < m_MinAddress || Buffer+Offset+sizeof(short) > m_MaxAddress) 
	{
		MESSAGE4("Out-of-bound address in PutShort %d %d %d",Buffer+Offset,m_MinAddress,m_MaxAddress);
		m_AddressError = true;
		return;
	}
	::PutShort(Buffer,Offset,Value);
}

uint CIFOLoader::GetInt(uchar* Buffer, int Offset)
{
	if(Buffer+Offset < m_MinAddress || Buffer+Offset+sizeof(int) > m_MaxAddress) 
	{
		MESSAGE4("Out-of-bound address in GetInt %d %d %d",Buffer+Offset,m_MinAddress,m_MaxAddress);
		m_AddressError = true;
		return 0;
	}
	return ::GetInt(Buffer,Offset);
}

ushort CIFOLoader::GetShort(uchar* Buffer, int Offset)
{
	if(Buffer+Offset < m_MinAddress || Buffer+Offset+sizeof(short) > m_MaxAddress)
	{
		MESSAGE4("Out-of-bound address in GetShort %d %d %d",Buffer+Offset,m_MinAddress,m_MaxAddress);
		m_AddressError = true;
		return 0;
	}
	return ::GetShort(Buffer,Offset);
}

void CIFOLoader::AddShort(uchar* Buffer, int Offset, short IncValue)
{
	if(Buffer+Offset < m_MinAddress || Buffer+Offset+sizeof(short) > m_MaxAddress)
	{
		MESSAGE4("Out-of-bound address in AddShort %d %d %d",Buffer+Offset,m_MinAddress,m_MaxAddress);
		m_AddressError = true;
		return;
	}
	::AddShort(Buffer,Offset,IncValue);
}

void CIFOLoader::AddInt(uchar* Buffer, int Offset, int IncValue)
{
	if(Buffer+Offset < m_MinAddress || Buffer+Offset+sizeof(int) > m_MaxAddress)
	{
		MESSAGE4("Out-of-bound address in AddInt %d %d %d",Buffer+Offset,m_MinAddress,m_MaxAddress);
		m_AddressError = true;
		return;
	}
	::AddInt(Buffer,Offset,IncValue);
}



CIFOLoader::CIFOLoader(const char* IFOFileName)
{
	m_Buffer = NULL;
	m_IFOfd = NULL;
	m_IsManager = 0;
	m_IFOSizekB = 0;
	m_MaxIFOSize = 0;
	m_NumVOBUs = 0;
	m_LoadedOK = 0;
	m_MinAddress = m_MaxAddress = 0;
	m_VOBUStartLBA = new uint[MAX_NUM_VOBUS];
	LoadIFO(IFOFileName);
}

CIFOLoader::CIFOLoader()
{
	m_Buffer = 0;
	m_IFOfd = NULL;
	m_IsManager = 0;
	strcpy(m_IFOFileName,"");
	m_IFOSizekB = 0;
	m_MaxIFOSize = 0;
	// FIX: FIX THIS!
	m_NumVOBUs = 0;
	m_LoadedOK = 0;
	m_MinAddress = m_MaxAddress = 0;
	m_VOBUStartLBA = new uint[MAX_NUM_VOBUS];
	m_VTSNum = -1;
}

int CIFOLoader::LoadIFO(const char* IFOFileName)
{
	m_LoadedOK = false;
	if(m_IFOfd)
		fclose(m_IFOfd);
	m_IFOfd = fopen(IFOFileName,"rb+");
	// If you can't read/write, let's see if we can at least read.
	if(m_IFOfd == 0)
		m_IFOfd = fopen(IFOFileName,"rb");
	m_IsManager = strcmpi(GetFname(IFOFileName),"video_ts") == 0;
	delete[] m_Buffer;
	m_Buffer = NULL;
	strcpy(m_IFOFileName,IFOFileName);
	m_IFOSizekB = GetFileSizekB(IFOFileName);
	if(m_IFOSizekB == 0)
		return -1;
	m_MaxIFOSize = m_IFOSizekB*1024 + MAX_IFO_SIZE_INCREASE*SECTOR_SIZE;
	if(m_IFOSizekB > 0)
	{
		m_Buffer = new uchar[m_MaxIFOSize];
		memset(m_Buffer,0,m_MaxIFOSize);
		m_MinAddress = m_Buffer;
		m_MaxAddress = m_Buffer+m_IFOSizekB*1024-1;
	}
	if(m_IFOfd == NULL)
		return -1;
	fread(m_Buffer,m_IFOSizekB*1024,1,m_IFOfd);
	m_NumVOBUs = 0;
	int i;
	// Find which VTS this is...
	char* Temp = GetFnameExt(IFOFileName);
	for(i=0;i<100;i++)
	{
		if(strcmpi(Temp,IFONames[i]) == 0)
		{
			m_VTSNum = i;
			break;
		}
	}
	// In case these are messed up!
	FixEndOfTables();
	m_LoadedOK = true;
	return 0;
}	

int CIFOLoader::LoadIFO(int VTS)
{
	m_LoadedOK = false;
	if(m_IFOfd)
		fclose(m_IFOfd);
	m_IFOfd = fopen(IFONames[VTS],"rb+");
	// If you can't read/write, let's see if we can at least read.
	if(m_IFOfd == 0)
		m_IFOfd = fopen(IFONames[VTS],"rb");
	m_IsManager = VTS == 0;
	m_VTSNum = VTS;
	delete[] m_Buffer;
	m_Buffer = NULL;
	strcpy(m_IFOFileName,IFONames[VTS]);
	m_IFOSizekB = GetFileSizekB(IFONames[VTS]);
	if(m_IFOSizekB == 0)
		return -1;
	m_MaxIFOSize = m_IFOSizekB*1024 + MAX_IFO_SIZE_INCREASE*SECTOR_SIZE;
	if(m_IFOSizekB > 0)
	{
		m_Buffer = new uchar[m_MaxIFOSize];
		memset(m_Buffer,0,m_MaxIFOSize);
		m_MinAddress = m_Buffer;
		m_MaxAddress = m_Buffer+m_IFOSizekB*1024-1;
	}
	if(m_IFOfd == NULL)
		return -1;
	fread(m_Buffer,m_IFOSizekB*1024,1,m_IFOfd);
	m_NumVOBUs = 0;
	// In case these are messed up!
	FixEndOfTables();
	m_LoadedOK = true;
	return 0;
}

void CIFOLoader::CloseIFO()
{
	if(m_IFOfd)
		fclose(m_IFOfd);
	m_IFOfd = 0;
}


CIFOLoader::~CIFOLoader()
{
	if(m_IFOfd) fclose(m_IFOfd);
	delete[] m_Buffer;
	delete[] m_VOBUStartLBA;
}

uchar* CIFOLoader::FindTable(IFOTable Table)
{
	// Load main table.
	IFO_main* Main = (IFO_main*) m_Buffer;

	uint Offset;

	switch(Table)
	{
	case IFO_MAIN :			return m_Buffer; break;
	case VTSM_PGCI_UT :		Offset = (m_IsManager) ? GetInt(m_Buffer,0x00c8) :  GetInt(m_Buffer,0x00d0) ; break;
	case VTSM_C_ADT :		Offset = GetInt((uchar*)&Main->m_PVTSM_C_ADT,0) ; break;
	case VTSM_VOBU_ADMAP:	Offset = GetInt((uchar*)&Main->m_PVTSM_VOBU_ADMAP,0) ; break;
	case TT_SRPT:			assert(m_IsManager) ; Offset = GetInt((uchar*)&Main->m_PTT_SRPT,0) ; break;
	case VTS_PTT_SRPT:		assert(!m_IsManager) ; Offset = GetInt((uchar*)&Main->m_PVTS_PTT_SRPT,0) ; break;
	case VMGM_PGCI_UT:		assert(m_IsManager) ; Offset = GetInt((uchar*)&Main->m_PVTS_PTT_SRPT,0) ; break;
	case VTS_C_ADT :		Offset = GetInt((uchar*)&Main->m_PVTS_C_ADT,0) ; break;
	case VTS_PGCI :			assert(!m_IsManager); Offset = GetInt((uchar*)&Main->m_PVTS_PGCI,0) ; break;
	case VMG_PTL_MAIT :		assert(m_IsManager); Offset = GetInt((uchar*)&Main->m_PVTS_PGCI,0) ; break;
	case VTS_VOBU_ADMAP:	Offset = GetInt((uchar*)&Main->m_PVTS_VOBU_ADMAP,0) ; break;
	case VTS_TMAPTI:		assert(!m_IsManager) ; Offset = GetInt((uchar*)&Main->m_PVTS_TMAPTI,0) ; break;
	case VMG_TXTDT_MG:		assert(m_IsManager) ; Offset = GetInt((uchar*)&Main->m_PVTS_TMAPTI,0) ; break;
	case FP_PGC:			assert(m_IsManager) ; Offset = GetInt(m_Buffer,0x84) ; break;
	case VMG_VTS_ATRT:		assert(m_IsManager) ; Offset = GetInt(m_Buffer,0x00d0) ; break;


	default: assert(0); return 0;
	}
	if(Offset == 0)	// No table.
		return 0;
	if(Table != FP_PGC)
	{
		// Make sure the table does not send you past the end of the IFO table.
		if(2*Offset > (uint)m_IFOSizekB)
			return 0;
		return m_Buffer+Offset*SECTOR_SIZE;
	}
	else
	{
		if(Offset > (uint)m_IFOSizekB*1024)
			return 0;
		return m_Buffer+Offset;
	}
}

uint CIFOLoader::GetEndAddress(IFOTable Table)
{
	uchar *Ptr = FindTable(Table);
	if(Ptr == 0)
		return 0;
	uint Ret = 0;
	if(Table != VTSM_VOBU_ADMAP && Table != VTS_VOBU_ADMAP && Table != VMG_TXTDT_MG && Table != FP_PGC)
		Ret = GetInt(Ptr,0x04);
	else if(Table == VMG_TXTDT_MG)
		Ret = GetInt(Ptr,0x10);
	else if(Table == FP_PGC)
	{
		// This is a lot more complicated.
		_ASSERT(m_IsManager);
		// Get command table:
		uint OffsetToCmds = GetShort(Ptr,0xe4);
		if(OffsetToCmds == 0) Ret = 0xeb;
		else Ret = OffsetToCmds + GetShort(Ptr,OffsetToCmds+6);
	}
	else
		Ret = GetInt(Ptr,0);
	Ret &= 0x00FFFFFF;
	if(Ptr - m_Buffer + Ret > (uint)m_IFOSizekB*1024) 
		Ret = 0;
	return Ret;
}

int CIFOLoader::FixOffset(uchar* Ptr, ushort Skip)
{
	if(Ptr == 0) return 0;
	uint Offset = GetInt(Ptr,Skip);
	PutInt(Ptr,Skip,Offset & 0x00FFFFFF);
	return (Offset & 0xFF000000);
}

void CIFOLoader::FixEndOfTables()
{
	uchar* Ptr;
	if(!m_IsManager)
	{
		Ptr = FindTable(VTS_PTT_SRPT);
		FixOffset(Ptr);
		Ptr = FindTable(VTS_PGCI);
		FixOffset(Ptr);
		Ptr = FindTable(VTSM_PGCI_UT);
		FixOffset(Ptr);
		Ptr = FindTable(VTS_TMAPTI);
		FixOffset(Ptr);
		Ptr = FindTable(VTSM_C_ADT);
		FixOffset(Ptr);
		Ptr = FindTable(VTSM_VOBU_ADMAP);
		FixOffset(Ptr,0);
		Ptr = FindTable(VTS_C_ADT);
		FixOffset(Ptr);
		Ptr = FindTable(VTS_VOBU_ADMAP);
		FixOffset(Ptr,0);
	}
	else
	{
		Ptr = FindTable(TT_SRPT);
		FixOffset(Ptr);
		Ptr = FindTable(VMGM_PGCI_UT);
		FixOffset(Ptr);
		Ptr = FindTable(VMG_PTL_MAIT);
		FixOffset(Ptr);
		Ptr = FindTable(VMG_VTS_ATRT);
		FixOffset(Ptr);
		Ptr = FindTable(VMG_TXTDT_MG);
		FixOffset(Ptr,10);
		Ptr = FindTable(VTSM_C_ADT);
		FixOffset(Ptr);
		Ptr = FindTable(VTSM_VOBU_ADMAP);
		FixOffset(Ptr,0);
	}
}


uint CIFOLoader::GetAbsoluteEndAddress(IFOTable Table)
{
	uchar *Ptr = FindTable(Table);
	uint Ret = 0;
	if(Ptr == 0)
		return Ret;
	if(Table != VTSM_VOBU_ADMAP && Table != VTS_VOBU_ADMAP && Table != VMG_TXTDT_MG && Table != FP_PGC)
		Ret = Ptr-m_Buffer+(GetInt(Ptr,0x04) & 0x00FFFFFF);
	else if(Table == VMG_TXTDT_MG)
		Ret = Ptr-m_Buffer+(GetInt(Ptr,0x10) & 0x00FFFFFF);
	else if(Table == FP_PGC)
		Ret = Ptr-m_Buffer+GetEndAddress(Table);
	else
		Ret = Ptr-m_Buffer+(GetInt(Ptr,0) & 0x00FFFFFF);

	if(Ret > (uint)m_IFOSizekB*1024) Ret = 0;
	return Ret;
}

void CIFOLoader::IncEndAddress(IFOTable Table, int IncBytes)
{
	uchar *Ptr = FindTable(Table);
	if(Ptr == 0) return;
	if(Table != VTSM_VOBU_ADMAP && Table != VTS_VOBU_ADMAP && Table != FP_PGC)
		AddInt(Ptr,0x04,IncBytes);
	else if(Table != FP_PGC)
		AddInt(Ptr,0,IncBytes);
}

bool CIFOLoader::TableSizeCanChange(IFOTable Table, int NumBytes)
{
	// Load main table.
	IFO_main* Main = (IFO_main*) m_Buffer;

	if(Table == VTSM_PGCI_UT || Table == VTSM_C_ADT || Table == VTSM_VOBU_ADMAP || 
		Table == VTS_PGCI || Table == VTS_C_ADT || Table == VTS_VOBU_ADMAP ||
		Table == VTS_TMAPTI || Table == FP_PGC || Table == VMG_VTS_ATRT || Table == TT_SRPT)
	{
		// I'm using the absolute end address to cover the FP_PGC table which is weird (does not start at LBA boundary).
		uint EndAddress = GetAbsoluteEndAddress(Table);
		if(EndAddress == 0) return false;
		// We can change the size if the incremented last byte is in the same sector.
		return ((EndAddress + NumBytes) / SECTOR_SIZE == EndAddress / SECTOR_SIZE);
	}
	assert(0);
	return 0;
}

void CIFOLoader::ChangeTableSize(IFOTable Table, int NumBytes)
{
	if(TableSizeCanChange(Table, NumBytes))
		return;
	uchar *Ptr = FindTable(Table);
	assert(Ptr);	// JUST SO I don't forget to check!
	if(Ptr==0)return;

	// We need to move stuff up or down!
#if 0
	// This does not work for First-Play PGC.
	uint EndAddress = GetEndAddress(Table);
	uchar* CurrentEndLBAPtr = Ptr + (EndAddress / SECTOR_SIZE + 1)*SECTOR_SIZE;
	uchar* DestEndLBAPtr = Ptr + ((EndAddress+NumBytes) / SECTOR_SIZE + 1)*SECTOR_SIZE;
#else
	// This does work for First-Play PGC...
	uint EndAddress = GetAbsoluteEndAddress(Table);
	uchar* CurrentEndLBAPtr = m_Buffer + (EndAddress / SECTOR_SIZE + 1)*SECTOR_SIZE;
	uchar* DestEndLBAPtr = m_Buffer + ((EndAddress+NumBytes) / SECTOR_SIZE + 1)*SECTOR_SIZE;
#endif
	int Inc = (int)DestEndLBAPtr - (int)CurrentEndLBAPtr;	// This is by how much we move.
	memmove(DestEndLBAPtr,CurrentEndLBAPtr,m_IFOSizekB*1024-(CurrentEndLBAPtr-m_Buffer));
	Inc /= SECTOR_SIZE;

	// Now update any pointer that might point past NextLBAPtr.
	IFO_main* Main = (IFO_main*) m_Buffer;
	uint ThisTableSector = (Ptr - m_Buffer) / SECTOR_SIZE;
	if(m_IsManager)
	{
		if(FixEndian(Main->m_PTT_SRPT) > ThisTableSector)
			AddInt((uchar*)&Main->m_PTT_SRPT,0,Inc);
	}

	if(FixEndian(Main->m_PVTS_PTT_SRPT) > ThisTableSector)
		AddInt((uchar*)&Main->m_PVTS_PTT_SRPT,0,Inc);

	if(FixEndian(Main->m_PVTS_PGCI) > ThisTableSector)
		AddInt((uchar*)&Main->m_PVTS_PGCI,0,Inc);

	if(FixEndian(Main->m_PVTSM_PGCI_UT) > ThisTableSector)
		AddInt((uchar*)&Main->m_PVTSM_PGCI_UT,0,Inc);

	if(FixEndian(Main->m_PVTS_TMAPTI) > ThisTableSector)
		AddInt((uchar*)&Main->m_PVTS_TMAPTI,0,Inc);

	if(FixEndian(Main->m_PVTSM_C_ADT) > ThisTableSector)
		AddInt((uchar*)&Main->m_PVTSM_C_ADT,0,Inc);

	if(FixEndian(Main->m_PVTSM_VOBU_ADMAP) > ThisTableSector)
		AddInt((uchar*)&Main->m_PVTSM_VOBU_ADMAP,0,Inc);

	if(FixEndian(Main->m_PVTS_C_ADT) > ThisTableSector)
		AddInt((uchar*)&Main->m_PVTS_C_ADT,0,Inc);

	if(FixEndian(Main->m_PVTS_VOBU_ADMAP) > ThisTableSector)
		AddInt((uchar*)&Main->m_PVTS_VOBU_ADMAP,0,Inc);
	
	// Also, modify last sector of IFO.
	AddInt((uchar*)&Main->m_IFOLastSector,0,Inc);

	// Finally, reflect change in size of file.
	m_IFOSizekB += Inc*SECTOR_SIZE/1024;
	m_MaxAddress = m_Buffer+m_IFOSizekB*1024-1;

}

void CIFOLoader::RemoveTable(IFOTable Table)
{
	uchar *Ptr = FindTable(Table);
	if(Ptr==0)return;
	// VMG_TXTDT_MG does not have an end address
	if(Table != VMG_TXTDT_MG)
	{
		uint EndAddress = GetEndAddress(Table);
		ChangeTableSize(Table, -(int)EndAddress-1);
	}
	// Now set the table pointer to 0.
	IFO_main* Main = (IFO_main*) m_Buffer;
	switch(Table)
	{
	case IFO_MAIN :			; break;
	case VTSM_PGCI_UT :		(m_IsManager) ? PutInt(m_Buffer,0x00c8,0) :  PutInt(m_Buffer,0x00d0,0) ; break;
	case VTSM_C_ADT :		PutInt((uchar*)&Main->m_PVTSM_C_ADT,0,0) ; break;
	case VTSM_VOBU_ADMAP:	PutInt((uchar*)&Main->m_PVTSM_VOBU_ADMAP,0,0) ; break;
	case TT_SRPT:			assert(m_IsManager) ; PutInt((uchar*)&Main->m_PTT_SRPT,0,0) ; break;
	case VTS_PTT_SRPT:		assert(!m_IsManager) ; PutInt((uchar*)&Main->m_PVTS_PTT_SRPT,0,0) ; break;
	case VMGM_PGCI_UT:		assert(m_IsManager) ; PutInt((uchar*)&Main->m_PVTS_PTT_SRPT,0,0) ; break;
	case VTS_C_ADT :		PutInt((uchar*)&Main->m_PVTS_C_ADT,0,0) ; break;
	case VTS_PGCI :			assert(!m_IsManager);  PutInt((uchar*)&Main->m_PVTS_PGCI,0,0) ; break;
	case VMG_PTL_MAIT :		assert(m_IsManager);  PutInt((uchar*)&Main->m_PVTS_PGCI,0,0) ; break;
	case VTS_VOBU_ADMAP:	PutInt((uchar*)&Main->m_PVTS_VOBU_ADMAP,0,0) ; break;
	case VTS_TMAPTI:		assert(!m_IsManager) ;  PutInt((uchar*)&Main->m_PVTS_TMAPTI,0,0) ; break;
	case VMG_TXTDT_MG:		assert(m_IsManager) ;  PutInt((uchar*)&Main->m_PVTS_TMAPTI,0,0) ; break;
	case FP_PGC:			assert(m_IsManager) ;  PutInt(m_Buffer,0x84,0) ; break;
	case VMG_VTS_ATRT:		assert(m_IsManager) ;  PutInt(m_Buffer,0x00d0,0) ; break;
	default: assert(0);		return;
	}
}


void CIFOLoader::MoveDataInTable(IFOTable Table, uchar* Ptr, int NumBytes)
{
	if(FindTable(Table) == 0 || Ptr == 0) return;
	if(NumBytes >= 0)
	{
		// We're growing, change table size first!
		ChangeTableSize(Table, NumBytes);
		uint EndAddress = GetEndAddress(Table);
		// Then move data. FindTable(Table) + GetEndAddress(Table) point to the last element of the table.
		int NumMove = FindTable(Table) + GetEndAddress(Table) - Ptr + 1;
		memmove(Ptr+NumBytes, Ptr, NumMove);
		IncEndAddress(Table,NumBytes);
	}
	else
	{
		// We're shrinking! Move things first, then change table size.
		uint EndAddress = GetEndAddress(Table);
		// Then move data. FindTable(Table) + GetEndAddress(Table) point to the last element of the table.
		int NumMove = FindTable(Table) + GetEndAddress(Table) - Ptr + 1;
		memmove(Ptr+NumBytes, Ptr, NumMove);
		ChangeTableSize(Table, NumBytes);
		IncEndAddress(Table,NumBytes);
	}
}

uchar* CIFOLoader::GoToLU(ushort LU)
{
	uchar*	Ptr = FindTable(VTSM_PGCI_UT);
	// Find offset to VTSM_LU
	Ptr =  Ptr?Ptr+GetInt(Ptr,12+(LU-1)*8) : 0;
	if((uint)(Ptr-m_Buffer) >= (uint)m_IFOSizekB*1024) Ptr = 0;
	return Ptr;
}	

uchar* CIFOLoader::GoToPCG(ushort LU, ushort PGC)
{
	uchar*	Ptr;
	if((LU == 0 && PGC == 0) || (m_IsManager && LU == 0))
	{
		// Special case: going to first-play PGC.
		Ptr = FindTable(FP_PGC);
	}
	else if(LU!=0)
	{
		Ptr = GoToLU(LU);
		if((uint)(Ptr+12+(PGC-1)*8-m_Buffer) >= (uint)m_IFOSizekB*1024) return 0;
		Ptr = Ptr ? Ptr + GetInt(Ptr,12+(PGC-1)*8) : 0;
	}
	else
	{
		Ptr = FindTable(VTS_PGCI);
		if((uint)(Ptr+12+(PGC-1)*8-m_Buffer) >= (uint)m_IFOSizekB*1024) return 0;
		Ptr = Ptr ? Ptr + GetInt(Ptr,12+(PGC-1)*8) : 0;
	}
	if((uint)(Ptr-m_Buffer) >= (uint)m_IFOSizekB*1024) Ptr = 0;
	return Ptr;
}


PGCType CIFOLoader::GetPGCType(ushort LU, ushort PGC, uchar* TitleNumber)
{
	if(LU == 0 && PGC == 0)
	{
		// Special case: first-play PGC.
		return PGCTypeFirstPlay;
	}
	if(LU!=0)
	{
		uchar*	Ptr = GoToLU(LU);
		if(Ptr == 0) return PGCTypeNone;
		uchar Category = Ptr[8+(PGC-1)*8];
		if((Category & 0x80)  == 0) return PGCTypeNone;
		return (PGCType) (Category & 0x0F);
	}
	else
	{
		uchar*	Ptr = FindTable(VTS_PGCI);
		if(Ptr == 0) return PGCTypeNone;
		uchar Category = Ptr[8+(PGC-1)*8];
		if((Category & 0x80)  == 0) return PGCTypeNone;
		if(TitleNumber) *TitleNumber = Category & 0x7F;
		return PGCTypeTitle;
	}
}

void CIFOLoader::GetNPUPGC(ushort LU, ushort PGC, ushort &NextPGC, ushort &PrevPGC, ushort &UpPGC)
{
	PGC_HEADER*	Ptr = GetPGCHeader(LU, PGC);
	if(Ptr)
	{
		NextPGC = FixEndian(Ptr->m_NextPGC);
		PrevPGC = FixEndian(Ptr->m_PrevPGC);
		UpPGC = FixEndian(Ptr->m_UpPGC);
	}
	else {NextPGC = PrevPGC = UpPGC = 0;}
}

	
ushort CIFOLoader::GetNumLU()
{
	uchar*	Ptr = FindTable(VTSM_PGCI_UT);
	return Ptr ? GetShort(Ptr,0) : 0;
}	

ushort CIFOLoader::GetNumPGC(ushort LU)
{
	if(LU!=0)
	{
		uchar*	Ptr = GoToLU(LU);
		// Go there and get the number of PGCs.
		return Ptr ? GetShort(Ptr,0) : 0;
	}
	else
	{
		// If video_ts, LU=0 is the first-play pgc...
		if(m_IsManager) 
			return GetInt(m_Buffer,0x84) ? 1 : 0;
		uchar*	Ptr = FindTable(VTS_PGCI);
		return Ptr ? GetShort(Ptr,0):0;
	}
}	

uchar CIFOLoader::GetNumCells(ushort LU, ushort PGC)
{
	uchar* Ptr = GoToPCG(LU, PGC);
	return Ptr ? Ptr[3] : 0;
}

uchar CIFOLoader::GetNumPrograms(ushort LU, ushort PGC)
{
	uchar* Ptr = GoToPCG(LU, PGC);
	return Ptr ? Ptr[2] : 0;
}

void CIFOLoader::SetNumPrograms(ushort LU, ushort PGC, uchar NumProg)
{
	uchar* Ptr = GoToPCG(LU, PGC);
	if(Ptr)
		Ptr[2] = NumProg;
}

uchar* CIFOLoader::GetProgramTable(ushort LU, ushort PGC)
{
	PGC_HEADER *Header = GetPGCHeader(LU,PGC);
	if(Header == 0) return 0;
	ushort Offset = FixEndian(Header->m_OffsetToPrograms);
	return (uchar*)Header+Offset;
}

ushort CIFOLoader::GetNumberVTSTitles()
{
	if(m_IsManager) return 0;
	uchar* Ptr = FindTable(VTS_PTT_SRPT);
	return Ptr ? GetShort(Ptr,0) : 0;
}

uint CIFOLoader::GetNumberChapters(ushort title)
{
	ushort NumVTSTitles =  GetNumberVTSTitles();
	if(title > NumVTSTitles)
		return 0;
	uchar* Ptr = FindTable(VTS_PTT_SRPT);
	if(Ptr == 0) return 0;
	int NumChapters;
	// You have to use the offset to the next title to get the number of chapters. If it's the last one, use the
	// end of table...
	if(title < NumVTSTitles-1)
		NumChapters = (GetInt(Ptr,8+(title+1)*4)-GetInt(Ptr,8+title*4))/4	;
	else
		NumChapters = ((GetInt(Ptr,4) & 0x00FFFFFF )+1-GetInt(Ptr,8+title*4))/4	;
	if(NumChapters > 1024)
	{
		// This is most certainly bogus!
		_ASSERT(0);
		return 0;
	}
	return NumChapters;
}

CELL_PLAYBACK_INFO *CIFOLoader::GetCellPlaybackInfo(ushort LU, ushort PGC, ushort Cell)
{
	if(GetNumCells(LU,PGC) < Cell)
		return 0;
	uchar* Ptr = GoToPCG(LU, PGC);
	if(Ptr == 0) return 0;
	// Now find offset to playback info for this cell.
	ushort Offset = GetShort(Ptr,0xe8);
	uchar* Ptr2 = Ptr + Offset + (Cell-1) * sizeof(CELL_PLAYBACK_INFO);
	return (CELL_PLAYBACK_INFO*) Ptr2;
}

CELL_POS_INFO *CIFOLoader::GetCellPosInfo(ushort LU, ushort PGC, ushort Cell)
{
	if(GetNumCells(LU,PGC) < Cell)
		return 0;
	uchar* Ptr = GoToPCG(LU, PGC);
	if(Ptr == 0) return 0;
	// Now find offset to playback info for this cell.
	ushort Offset = GetShort(Ptr,0xea);
	uchar* Ptr2 = Ptr + Offset + (Cell-1) * sizeof(CELL_POS_INFO);
	return (CELL_POS_INFO*) Ptr2;
}

int CIFOLoader::FindThisCell(ushort VOBID, uchar CELLID, ushort *LUOut, ushort *PGCOut, ushort *CELLOut)
{
	ushort NumPGC,NumCell;
	int Found = 0;
	if(LUOut != NULL)
	{
		ushort NumLU = GetNumLU();

		// Looping through all the LUs, all the PGCs and all the Cells, looking for the right VCID.
		ushort LU,PGC,CELL;
		for(LU=1;LU<=NumLU;LU++)
		{
			NumPGC = GetNumPGC(LU);
			for(PGC=1;PGC<=NumPGC;PGC++)
			{
				NumCell = GetNumCells(LU,PGC);
				for(CELL=1;CELL<=NumCell;CELL++)
				{
					CELL_POS_INFO *Temp = GetCellPosInfo(LU,PGC,CELL);
					if(FixEndian(Temp->m_VOBID) == VOBID && Temp->m_CELLID == CELLID)
					{
						// Found it!
						LUOut[Found] = LU;
						PGCOut[Found] = PGC;
						CELLOut[Found++] = CELL;
					}
				}
			}
		}
	}
	else
	{
		ushort PGC,CELL;
		NumPGC = GetNumPGC(0);
		for(PGC=1;PGC<=NumPGC;PGC++)
		{
			NumCell = GetNumCells(0,PGC);
			for(CELL=1;CELL<=NumCell;CELL++)
			{
				CELL_POS_INFO *Temp = GetCellPosInfo(0,PGC,CELL);
				if(FixEndian(Temp->m_VOBID) == VOBID && Temp->m_CELLID == CELLID)
				{
					// Found it!
					PGCOut[Found] = PGC;
					CELLOut[Found++] = CELL;
				}
			}
		}
	}
	return Found;
}

bool CIFOLoader::IsReferenced(ushort VOBID, uchar CELLID, bool Menu)
{
	int NumLU = Menu?GetNumLU():1;
	int j;
	ushort PGC,CELL;
	// Go through all LUs.
	for(j=1;j<=NumLU;j++)
	{
		int LU = (Menu)?j:0;
		int NumPGC = GetNumPGC(LU);
		for(PGC=1;PGC<=NumPGC;PGC++)
		{
			int NumCell = GetNumCells(LU,PGC);
			for(CELL=1;CELL<=NumCell;CELL++)
			{
				CELL_POS_INFO *Temp = GetCellPosInfo(LU,PGC,CELL);
				if(FixEndian(Temp->m_VOBID) == VOBID && Temp->m_CELLID == CELLID)
				{
					return true;
				}
			}
		}
	}
	return false;
}

bool CIFOLoader::AreAnyCellUnreferenced(bool menu)
{
	uchar* Ptr = FindTable(menu?VTSM_C_ADT:VTS_C_ADT);
	if(Ptr == 0) return false;

	int EndAdd = GetInt(Ptr,4);
	int NumVCID = (EndAdd+1-8)/12;
	int i;
	for(i =0;i <NumVCID;i ++)
	{
		ushort VID = GetShort(Ptr,8+i*12);
		uchar CID = Ptr[8+2+i*12];
		if(IsReferenced(VID,CID,menu) == false)
			return true;
	}
	return false;
}



void CIFOLoader::GetCLUT(uchar *CLUT, ushort LU, ushort PGC, bool RGB)
{
	PGC_HEADER *Header = GetPGCHeader(LU,PGC);
	assert(Header != 0);
	if(Header == 0)
		return;
	int i;
	uchar* Ptr = (uchar*) Header->m_Palette;
	for(i=0;i<16 && Ptr != 0;i++)
	{
		if(!RGB)
		{
			// Y Cr Cb
			CLUT[3*i] = Ptr[4*i+1];
			CLUT[3*i+1] = Ptr[4*i+2];
			CLUT[3*i+2] = Ptr[4*i+3];
		}
		else
		{
			// http://www.fourcc.org/fccyvrgb.php
#if 0
			float R = Ptr[4*i+1] + 1.402f*(Ptr[4*i+2]-128);
			CLUT[3*i] = (uchar) max(0,min(R,255));
			float G = Ptr[4*i+1] - 0.34414f*(Ptr[4*i+3]-128) - 0.71414f*(Ptr[4*i+2]-128);
			CLUT[3*i+1] = (uchar) max(0,min(G,255));
			float B = Ptr[4*i+1] + 1.772f*(Ptr[4*i+3]-128);
			CLUT[3*i+2] = (uchar) max(0,min(B,255));
#else		// These from r0lZ. They map Y=255 to a true white. The 0.5f acts as a round (a cast is a floor).
			float R = .5f + 1.164f*(Ptr[4*i+1] - 16) + 1.596f*(Ptr[4*i+2]-128);
			CLUT[3*i] = (uchar) max(0,min(R,255));
			float G = .5f + 1.164f*(Ptr[4*i+1] - 16) - 0.392f*(Ptr[4*i+3]-128) - 0.813f*(Ptr[4*i+2]-128);
			CLUT[3*i+1] = (uchar) max(0,min(G,255));
			float B = .5f + 1.164f*(Ptr[4*i+1] - 16) + 2.017f*(Ptr[4*i+3]-128);
			CLUT[3*i+2] = (uchar) max(0,min(B,255));
#endif
		}
	}
}

void CIFOLoader::SetCLUT(uchar *CLUT, ushort LU, ushort PGC, bool RGB)
{
	PGC_HEADER *Header = GetPGCHeader(LU,PGC);
	assert(Header != 0);
	if(Header == 0)
		return;
	int i;
	uchar* Ptr = (uchar*) Header->m_Palette;
	for(i=0;i<16 && Ptr != 0;i++)
	{
		if(!RGB)
		{
			// Y Cr Cb
			Ptr[4*i+1] = CLUT[3*i];
			Ptr[4*i+2] = CLUT[3*i+1];
			Ptr[4*i+3] = CLUT[3*i+2];
		}
		else
		{
			// http://www.fourcc.org/fccyvrgb.php . Or inv([1 1.402 0;1 -0.71414 -0.34414 ;1 0 1.772]) in matlab.
#if 0
			float Y = (0.2990f * CLUT[3*i]) + (0.5870f * CLUT[3*i+1]) + (0.1140f * CLUT[3*i+2]);
			float Cr =  (0.5000f * CLUT[3*i]) - (0.4187f * CLUT[3*i+1]) - (0.0813f * CLUT[3*i+2]) + 128;
			float Cb = -(0.1687f * CLUT[3*i]) - (0.3313f * CLUT[3*i+1]) + (0.5000f * CLUT[3*i+2]) + 128;
#else		// From r0lZ. The 0.5 acts as a round().
			float Y = (0.257f * CLUT[3*i]) + (0.504f * CLUT[3*i+1]) + (0.098f * CLUT[3*i+2]) + 16.5f;
			float Cr =  (0.439f * CLUT[3*i]) - (0.368f * CLUT[3*i+1]) - (0.071f * CLUT[3*i+2]) + 128.5f;
			float Cb = -(0.148f * CLUT[3*i]) - (0.291f * CLUT[3*i+1]) + (0.439f * CLUT[3*i+2]) + 128.5f;
#endif
			Ptr[4*i+1] = (uchar) max(0,min(Y,255));
			Ptr[4*i+2] = (uchar) max(0,min(Cr,255));
			Ptr[4*i+3] = (uchar) max(0,min(Cb,255));
		}
	}
}


int CIFOLoader::GetSubpicLangages(char* Languages, bool Menu)
{
	int NumLang = 0;
	if(Menu || m_IsManager)
	{
		NumLang = GetShort(m_Buffer,0x154);
		Languages[0] = m_Buffer[0x158];
		Languages[1] = m_Buffer[0x159];
	}
	else
	{
		NumLang = GetShort(m_Buffer,0x254);
		int i;
		for(i=0;i<NumLang;i++)
		{
			Languages[2*i] = m_Buffer[0x258+i*6];
			Languages[2*i+1] = m_Buffer[0x259+i*6];
		}
	}
	return NumLang;
}

bool CIFOLoader::IsSubpicStreamAvailable(ushort LU, ushort PGC, uchar Index)
{
	PGC_HEADER *Header = GetPGCHeader(LU,PGC);
	assert(Header != 0);
	if(Header == 0)
		return 0;
	int Attributes = FixEndian(Header->m_SubpicStreams[Index]);
	// Is the stream available?
	return ((Attributes & 0x80000000) != 0);
}

void CIFOLoader::FixSubpicAttributes(bool menu)
{
	if(!menu && m_IsManager) return;
	ushort LU,PGC;
	int i;
	uchar Used[32];
	memset(Used,0,32);
	ushort NumLU = menu == true ? GetNumLU() : 1;
	bool AllDummies = true;
	for(LU=1;LU<=NumLU;LU++)
	{
		ushort TheLU = menu ? LU : 0;
		ushort NumPGC = GetNumPGC(TheLU);
		for(PGC=1;PGC<=NumPGC;PGC++)
		{
			if(GetNumCells(LU,PGC) > 0) AllDummies = false;
			for(i=0;i<32;i++)
			{
				if(IsSubpicStreamAvailable(TheLU, PGC, i) != 0) Used[i] = 1;
			}
		}
	}
	// Let's get the top index of the last used stream.
	for(i=31;i>=0;i--)
		if(Used[i] == 1) break;
	ushort NumStreams = i+1;
	if(AllDummies == false && menu) NumStreams = max(NumStreams,1);
	// Now adjust that in the IFO.
	PutShort(m_Buffer,(menu ? 0x154 : 0x254),NumStreams);
}



bool CIFOLoader::GetPgcSubpicLanguage(char* Languages, ushort LU, ushort PGC, uchar StreamID, uchar* DispModes)
{
	char DLanguages[128];
	// First get the languages defined in the domain.
	int NumLang = GetSubpicLangages(DLanguages,LU!=0?true:false);
	// Now go the the PGC.
	PGC_HEADER *Header = GetPGCHeader(LU,PGC);
	assert(Header != 0);
	if(Header == 0)
		return 0;

	// Get video attributes.
	uchar VideoAttributes = m_Buffer[LU==0?0x200:0x100];
	uchar  VideoDisModes = 0;
	if((VideoAttributes & 0xc) == 0xc)
		VideoDisModes = (~(VideoAttributes & 0x3) << 2) | 0x02 ;	// 16:9 video.
	else if((VideoAttributes & 0xc) == 0)
		VideoDisModes = 0x01;										// 4:3 video
	int i;
	bool Done = false;
	if(DispModes) *DispModes = 0;
	for(i=0;i<32;i++)
	{
		int Attributes = FixEndian(Header->m_SubpicStreams[i]);
		// Is the stream available?
		if((Attributes & 0x80000000) == 0)
			continue;
		// The stream number in Attributes is between 0 and 32, and corresponds to streams
		// 0x20 to 0x3F
		if(((Attributes>>24) & 0x1F) == StreamID-0x20) 
		{
			// 4:3
			if(DispModes) *DispModes |= 1;
			Done = true;	
		}
		if(((Attributes>>16) & 0x1F) == StreamID-0x20) 
		{
			// Wide
			if(DispModes) *DispModes |= 2;
			Done = true;
		}
		if(((Attributes>>8) & 0x1F) == StreamID-0x20) 
		{
			// Letterbox
			if(DispModes) *DispModes |= 4;
			Done = true;
		}
		if(((Attributes) & 0x1F) == StreamID-0x20) 
		{
			// Pan&Scan
			if(DispModes) *DispModes |= 8;
			Done = true;
		}
		if(Done) break;
	}
	if(i==32)
	{
		Languages[0] = ' ';
		Languages[1] = ' ';
		return false;
	}
	Languages[0] = DLanguages[2*i];
	Languages[1] = DLanguages[2*i+1];
	if(Languages[0] == 0) Languages[0] = ' ';
	if(Languages[1] == 0) Languages[1] = ' ';
	if(DispModes)
		*DispModes &= VideoDisModes;
	return true;
}

PGC_HEADER* CIFOLoader::GetPGCHeader(ushort LU, ushort PGC)
{
	uchar* Ptr = GoToPCG(LU, PGC);
	return (PGC_HEADER*) Ptr;
}

int CIFOLoader::WriteIFO(char* WritePath)
{
	if(WritePath == NULL)
	{
		// Overwrite input IFO file.
		fseek(m_IFOfd,0,SEEK_SET);
		// I need to change the size of that file, in case we shortened it!
		int filehandle = _fileno(m_IFOfd);
		if(filehandle) _chsize(filehandle,m_IFOSizekB*1024);
		if(fwrite(m_Buffer,m_IFOSizekB*1024,1,m_IFOfd) == 0) return -1;
		fflush(m_IFOfd);
		FILE* fd = fopen(ChangeExt(m_IFOFileName,".BUP"),"wb");
		if(fd)
		{
			if(fwrite(m_Buffer,m_IFOSizekB*1024,1,fd) == 0) return -1;
			fclose(fd);
			return 0;
		}
		return -1;
	}
	else
	{
		FILE* fd = fopen(WritePath,"wb");
		if(fd)
		{
			if(fwrite(m_Buffer,m_IFOSizekB*1024,1,fd) == 0) return -1;
			fclose(fd);
		}
		else return -1;
		fd = fopen(ChangeExt(WritePath,".BUP"),"wb");
		if(fd)
		{
			if(fwrite(m_Buffer,m_IFOSizekB*1024,1,fd) == 0) return -1;
			fclose(fd);
			return 0;
		}
		else return -1;
	}
}

uchar* CIFOLoader::GetNumAudioStreams(bool menu)
{
	if(!menu && m_IsManager) return 0;
	return (m_Buffer+(menu ? 0x102 : 0x202));
}

AUDIO_ATTRIBUTES* CIFOLoader::GetAudioAttributes(uint ID, bool menu)
{
	if(!menu && m_IsManager) return 0;
	uchar *Ptr = m_Buffer+ (menu ? 0x104+8*ID : 0x204+8*ID);
	return (AUDIO_ATTRIBUTES*) Ptr;
}

void CIFOLoader::FixAudioAttributes(bool menu)
{
	if(!menu && m_IsManager) return;
	ushort LU,PGC;
	int i;
	uchar Used[8] = {0,0,0,0,0,0,0,0};
	ushort NumLU = menu == true ? GetNumLU() : 1;
	for(LU=1;LU<=NumLU;LU++)
	{
		ushort TheLU = menu ? LU : 0;
		ushort NumPGC = GetNumPGC(TheLU);
		for(PGC=1;PGC<=NumPGC;PGC++)
		{
			for(i=0;i<8;i++)
			{
				if(GetPGCAudioInfo(TheLU, PGC, i, 0, 0, 0, 0) == 0) Used[i] = 1;
			}
		}
	}
	// Let's get the top index of the last used stream.
	for(i=7;i>=0;i--)
		if(Used[i] == 1) break;
	ushort NumStreams = i+1;
	// Now adjust that in the IFO.
	PutShort(m_Buffer,(menu ? 0x102 : 0x202),NumStreams);
}


int CIFOLoader::GetPGCAudioInfo(ushort LU, ushort PGC, uchar Index, char* Lang, char* Encoding, uchar* NumChans, uchar* StreamID)
{
	int NumAudioStream = GetShort(GetNumAudioStreams(LU != 0),0);
	PGC_HEADER* Header = GetPGCHeader(LU,PGC);
	if(Header == 0) return -1;
	ushort AS = FixEndian(Header->m_AudioStreams[Index]);
	if((AS & 0x8000) != 0)
	{
		// This stream is available.
		AUDIO_ATTRIBUTES* Attr = GetAudioAttributes(Index,LU!=0);
		uchar ID = (AS&0x0700)>>8;
		uchar StreamType = Attr->m_CodingMode >> 5;
		char* CodeName;
		switch(StreamType)
		{
		case 0 : CodeName = "AC3"; ID += 0x80; break;
		case 2 : CodeName = "Mpeg1"; ID += 0x80; break;
		case 3 : CodeName = "Mpeg2"; ID += 0xC0; break;
		case 4 : CodeName = "LPCM"; ID += 0xA0; break;
		case 6 : CodeName = "DTS"; ID += 0x88; break;
		default : CodeName = "?"; ID += 0; 
		}
		if(Lang) strncpy(Lang,(const char*)Attr->m_Language,2), Lang[2]=0;
		if(Encoding) strcpy(Encoding,CodeName);
		if(NumChans) *NumChans = (Attr->m_Quantization & 0x07) + 1;
		if(StreamID) *StreamID = ID;
		return 0;
	}
	return -1;
}


uchar* CIFOLoader::GetVideoAttribute()
{
	// TODO: I should really ask whether you want the attributes for the menus or the titles.
	uchar* Ptr = m_Buffer+ (m_IsManager ? 0x0100 : 0x0200);
	return Ptr;
}

uchar* CIFOLoader::GoToCommands(ushort LU, ushort PGC)
{
	PGC_HEADER *Header = GetPGCHeader(LU, PGC);
	if(Header == 0) return 0;
	ushort Offset = FixEndian(Header->m_OffsetToCommands);
	if(Offset == 0)
		return 0;
	return (uchar*)Header + Offset;
}

ushort CIFOLoader::GetNumCommands(ushort LU, ushort PGC, int Type)
{
	uchar* Temp = GoToCommands(LU,PGC);
	if(Temp == 0) return 0;
	ushort NumCommands = 0;
	switch(Type)
	{
	case 0: NumCommands = GetShort(Temp,0); break;
	case 1: NumCommands = GetShort(Temp,2); break;
	case 2: NumCommands = GetShort(Temp,4); break;
	case 3: NumCommands = min(128,GetShort(Temp,0)) + min(128,GetShort(Temp,2)) + min(128,GetShort(Temp,4));
	}
	if(Type != 3 && NumCommands > 128)
	{
		MESSAGE2("Abnormal number of commands! %d",NumCommands);
		NumCommands = 128;
	}
	
	return NumCommands;
}

uchar* CIFOLoader::GetCommands(ushort LU, ushort PGC, int Type, uchar** Commands)
{

	uchar* Ptr = GoToCommands(LU,PGC);
	if(Ptr == 0)
	{
		// 0 indicates there's no command whatsoever.
		if(Commands)
			*Commands = 0;
		return 0;
	}
	if(Commands)
	{
		switch(Type)
		{
		case 0: *Commands = Ptr + 8; break;
		case 1: *Commands = Ptr + 8 + 8*min(128,GetShort(Ptr,0)); break; 
		case 2: *Commands = Ptr + 8 + 8*(min(128,GetShort(Ptr,0)) + min(128,GetShort(Ptr,2))); break;
		default: _ASSERT(0); break;
		}
	}
	return Ptr + Type*2;
}

int CIFOLoader::InsertCommand(ushort LU, ushort PGC, uchar* Command, int Type)
{
	uchar Temp[8] ={0,0,0,0,0,0,0,0};
	// Add command to the end.
	AddCommand(LU,PGC,Temp,Type);
	// Then copy all commands down.
	uchar *Ptr;
	uchar* NumCommands = GetCommands(LU, PGC, Type, &Ptr);
	if(Ptr == 0 || NumCommands == 0) return -1;
	int NumCom = GetShort(NumCommands,0);
	memmove(Ptr+8,Ptr,8*(NumCom - 1));
	// Add new command at the top.
	memcpy(Ptr,Command,8);
	// Now renumber gotos...
	int i;
	for(i=1;i<NumCom;i++)
	{
		memcpy(Temp,Ptr+i*8,8);
		ushort OpCode = ::GetShort(Temp,0);
		// A goto has TYP=0, CMD=1
		if(TYP == 0 && CMD == 1)
		{
			Temp[7] ++;
			memcpy(Ptr+i*8,Temp,8);
		}
	}
	return 0;
}
	

int CIFOLoader::AddCommand(ushort LU, ushort PGC, uchar* Command, int Type)
{
	ushort TotNumCommands = GetNumCommands(LU,PGC,3);
	if(TotNumCommands >= 128)
		return -1;


	// Change size of VTSM_PGCI_UT table. This will move other tables down if needed.
	uchar *Ptr;
	// This holds by how much we're moving things below us. It's normally 8 bytes.
	int Offset = 8;
	if(TotNumCommands == 0)
	{
		// No commands! There's no space for the numbers of pre/post/cell commands and the end address
		// Move things down by 8 bytes for the command table.
		// Adding command ro first-play is difficult. Ignore for now!
		_ASSERT(LU != 0 || PGC != 0);
		if(LU == 0 && PGC == 0) return -1;
		PGC_HEADER *Header = GetPGCHeader(LU, PGC);
		if(Header == 0) return -1;
		MoveDataInTable(LU==0 ? VTS_PGCI : VTSM_PGCI_UT,(uchar*)Header+0xec,8);
		// Set the offset to the command tables.
		Header->m_OffsetToCommands = FixEndian((ushort)0xec);
		// Zero-out the number of pre, post and cell commands, and set the end byte to 0x7.
		PutShort((uchar*)Header,0xec,0);
		PutShort((uchar*)Header,0xec+2,0);
		PutShort((uchar*)Header,0xec+4,0);
		PutShort((uchar*)Header,0xec+6,0x07);
		// Now, every other PCG and LU tables below us will have to be moved by 8 more bytes.
		Offset += 8;
	}
	// Go there
	uchar *TableStart = FindTable(LU==0 ? VTS_PGCI : VTSM_PGCI_UT);
	uchar* NumCommands = GetCommands(LU, PGC, Type, &Ptr);
	if(Ptr == 0) return 0;

	// Move everything down below the command by 8 bytes. This will resize the table, and move others if needed.
	Ptr += 8*GetShort(NumCommands,0);
	if(LU == 0 && PGC == 0)
		MoveDataInTable(FP_PGC,Ptr,8);
	else
		MoveDataInTable(LU==0 ? VTS_PGCI : VTSM_PGCI_UT,Ptr,8);

	// Add the command...
	memcpy(Ptr,Command,8);

	// Now increase the number of commands and update the end address (offset 0x6 relative to
	// the number of precommands).
	AddShort(NumCommands,0,1); 
	AddShort(GetCommands(LU, PGC, 0, 0),6,8);	// Increase by 8 bytes.

	// Now update the program map, cell playback and cell position offsets.
	PGC_HEADER *Header = GetPGCHeader(LU, PGC);
	if(Header->m_OffsetToPrograms != 0)
		AddShort((uchar*)&Header->m_OffsetToPrograms,0,Offset);
	if(Header->m_OffsetToCellPB != 0)
		AddShort((uchar*)&Header->m_OffsetToCellPB,0,Offset);
	if(Header->m_OffsetToCellPos != 0)
		AddShort((uchar*)&Header->m_OffsetToCellPos,0,Offset);

	if(LU)
	{
		// Now update the PGC offsets defined in VTSM_LU for the PGCs after us.
		ushort NumPGC = GetNumPGC(LU);
		Ptr = GoToLU(LU);
		AddInt(Ptr,4,Offset); // Update end address!
		ushort pgc;
		for(pgc=PGC;pgc<NumPGC;pgc++)
		{
			AddInt(Ptr,12+pgc*8,Offset);
		}

		// Update the LU offsets defined in VTSM_PGCI_UT for the LUs after us
		ushort NumLU = GetNumLU();
		Ptr = FindTable(VTSM_PGCI_UT);
//		AddInt(Ptr,0x4,Offset); // Update end address!	NO MoveDataInTable DOES IT FOR US!
		ushort lu;
		for(lu=LU;lu<NumLU && Ptr != 0;lu++)
		{
			AddInt(Ptr,12+lu*8,Offset);
		}
	}
	else if(PGC != 0)
	{
		_ASSERT(m_VTSNum != 0);
		ushort NumPGC = GetNumPGC(LU);
		Ptr = FindTable(VTS_PGCI);
//		AddInt(Ptr,0x4,Offset); // Update end address! NO MoveDataInTable DOES IT FOR US!
		ushort pgc;
		for(pgc=PGC;pgc<NumPGC && Ptr != 0;pgc++)
		{
			AddInt(Ptr,12+pgc*8,Offset);
		}
	}
	else	// Special case for the first-play...
		AddInt(m_Buffer,0x80,8);

	return (int)GetShort(NumCommands,0);
}

int CIFOLoader::RemoveCommand(ushort LU, ushort PGC, int Type, int CommandIndex)
{
	ushort TotNumCommands = GetNumCommands(LU,PGC,3);
	if(TotNumCommands == 0)
		return -1;

	// Change size of VTSM_PGCI_UT table. This will move other tables down if needed.
	uchar *Ptr;
	// This holds by how much we're moving things below us. It's normally 8 bytes.
	int Offset = -8;
	// Go there
	IFOTable Table = LU==0 ? VTS_PGCI : VTSM_PGCI_UT;
	// Removing a command from first-play is difficult!
	_ASSERT(LU != 0 || PGC != 0);
	if(LU == 0 && PGC == 0) return -1;
	
	uchar *TableStart = FindTable(Table);
	uchar* NumCommands = GetCommands(LU, PGC, Type, &Ptr);

	if(NumCommands == 0 || CommandIndex > GetShort(NumCommands,0) || Ptr == 0)
		return -1;
	// Move everything up below the command by 8 bytes. This will resize the table, and move others if needed.
	Ptr += 8*CommandIndex;
	MoveDataInTable(Table,Ptr,-8);

	// Now decrease the number of commands and update the end address (offset 0x6 relative to
	// the number of precommands).
	AddShort(NumCommands,0,-1); 
	AddShort(GetCommands(LU, PGC, 0, 0),6,-8);	// Decrease by 8 bytes.

	// Now update the program map, cell playback and cell position offsets.
	PGC_HEADER *Header = GetPGCHeader(LU, PGC);
	AddShort((uchar*)&Header->m_OffsetToPrograms,0,Offset);
	AddShort((uchar*)&Header->m_OffsetToCellPB,0,Offset);
	AddShort((uchar*)&Header->m_OffsetToCellPos,0,Offset);

	if(LU)
	{
		// Now update the PGC offsets defined in VTSM_LU for the PGCs after us.
		ushort NumPGC = GetNumPGC(LU);
		Ptr = GoToLU(LU);
		AddInt(Ptr,4,Offset); // Update end address!
		ushort pgc;
		for(pgc=PGC;pgc<NumPGC;pgc++)
		{
			AddInt(Ptr,12+pgc*8,Offset);
		}

		// Update the LU offsets defined in VTSM_PGCI_UT for the LUs after us
		ushort NumLU = GetNumLU();
		Ptr = FindTable(VTSM_PGCI_UT);
		ushort lu;
		for(lu=LU;lu<NumLU && Ptr != 0;lu++)
		{
			AddInt(Ptr,12+lu*8,Offset);
		}
	}
	else
	{
		ushort NumPGC = GetNumPGC(LU);
		Ptr = FindTable(VTS_PGCI);
		ushort pgc;
		for(pgc=PGC;pgc<NumPGC && Ptr != 0;pgc++)
		{
			AddInt(Ptr,12+pgc*8,Offset);
		}
	}

	return (int)GetShort(NumCommands,0);
}


void CIFOLoader::UpdateVTS_C_ADT(CCell* TheCell, bool AddToVTSM_C_ADT, bool Menu)
{
	int i;
	uint LastVOBUEndLBA = TheCell->m_StartLBA + TheCell->m_NumPacks-1;

	VTSM_C_ADTTable* Table = (VTSM_C_ADTTable*) FindTable(Menu?VTSM_C_ADT:VTS_C_ADT);
	if(Table == 0)
		return;
	ushort NumCells = (FixEndian(Table->m_LastByte) - 8 + 1) / sizeof(VTSM_C_ADTItem);
	VTSM_C_ADTItem* ThisItem;
	for(i=0;i<(int)NumCells;i++)
	{
		ThisItem = &Table->m_Items[i];
		if(FixEndian(ThisItem->m_VOBID) != TheCell->m_VOBID || ThisItem->m_CELLID != TheCell->m_CELLID)\
			continue;
		// Found the cell, let's update the start and end LBAs.
		ThisItem->m_StartLBA = FixEndian((uint)TheCell->m_StartLBA);
		ThisItem->m_EndLBA = FixEndian(LastVOBUEndLBA);
		break;
	}
	if(i==(int)NumCells && AddToVTSM_C_ADT)
	{
		// The cell wasn't found in the VTSM_C_ADT, let's add one entry for it.
		ChangeTableSize(Menu?VTSM_C_ADT:VTS_C_ADT,sizeof(VTSM_C_ADTItem));
		ThisItem = &Table->m_Items[NumCells];
		ThisItem->m_VOBID = FixEndian(TheCell->m_VOBID);
		ThisItem->m_CELLID = TheCell->m_CELLID;
		ThisItem->m_StartLBA = FixEndian((uint)TheCell->m_StartLBA);
		ThisItem->m_EndLBA = FixEndian(LastVOBUEndLBA);
		// Change the last byte.
		AddInt((uchar*)&Table->m_LastByte,0,sizeof(VTSM_C_ADTItem));
	}
}

void CIFOLoader::RemoveFromVTS_C_ADT(CCell* TheCell, bool Menu)
{
	int i,j;

	VTSM_C_ADTTable* Table = (VTSM_C_ADTTable*) FindTable(Menu?VTSM_C_ADT:VTS_C_ADT);
	if(Table == 0)
		return;
	ushort NumCells = (FixEndian(Table->m_LastByte) - 8 + 1) / sizeof(VTSM_C_ADTItem);
	VTSM_C_ADTItem* ThisItem;
	for(i=j=0;i<(int)NumCells;i++)
	{
		Table->m_Items[j] = Table->m_Items[i];
		ThisItem = &Table->m_Items[i];
		if(FixEndian(ThisItem->m_VOBID) != TheCell->m_VOBID || ThisItem->m_CELLID != TheCell->m_CELLID)\
			j++;
	}
	if(j!=i)
	{
		// Change the table size.
		ChangeTableSize(Menu?VTSM_C_ADT:VTS_C_ADT,-(int)sizeof(VTSM_C_ADTItem));
		AddInt((uchar*)&Table->m_LastByte,0,-(int)sizeof(VTSM_C_ADTItem));
	}
	ushort Count=0;
	for(i=0;i<NumCells;i++)
	{
		// Go through list, and if you don't see this VOBID in the preceding cells, increment count.
		for(j=i-1;j>=0;j--)
		{
			if(Table->m_Items[j].m_VOBID == Table->m_Items[i].m_VOBID)
				break;
		}
		if(j==-1) Count++;
	}
	// Write that back into the table.
	Table->m_NumVOBID = FixEndian(Count);
}

void CIFOLoader::CleanupVTS_C_ADT(bool Menu)
{
	int i,j;

	VTSM_C_ADTTable* Table = (VTSM_C_ADTTable*) FindTable(Menu?VTSM_C_ADT:VTS_C_ADT);
	if(Table == 0)
		return;
	ushort NumCells = (FixEndian(Table->m_LastByte) - 8 + 1) / sizeof(VTSM_C_ADTItem);
	VTSM_C_ADTItem* ThisItem;
	ushort LUList[1024];
	ushort PGCList[1024];
	ushort CELLList[1024];
	for(i=j=0;i<(int)NumCells;i++)
	{
		// Go through list of cells, see if it's referenced in any of the PGCs...
		Table->m_Items[j] = Table->m_Items[i];
		ThisItem = &Table->m_Items[i];
		if(FindThisCell(FixEndian(ThisItem->m_VOBID),ThisItem->m_CELLID,
			(Menu)?LUList:0,PGCList,CELLList) != 0)
			j++;
	}
	if(j!=i)
	{
		// Change the table size.
		ChangeTableSize(Menu?VTSM_C_ADT:VTS_C_ADT,(j-i)*(int)sizeof(VTSM_C_ADTItem));
		AddInt((uchar*)&Table->m_LastByte,0,(j-i)*sizeof(VTSM_C_ADTItem));
	}
	ushort Count=0;
	for(i=0;i<NumCells;i++)
	{
		// Go through list, and if you don't see this VOBID in the preceding cells, increment count.
		for(j=i-1;j>=0;j--)
		{
			if(Table->m_Items[j].m_VOBID == Table->m_Items[i].m_VOBID)
				break;
		}
		if(j==-1) Count++;
	}
	// Write that back into the table.
	Table->m_NumVOBID = FixEndian(Count);

}

void CIFOLoader::RebuildVTS_C_ADT(VTSM_C_ADTItem *Items, int NumCells, bool Menu)
{
	int i;

	VTSM_C_ADTTable* Table = (VTSM_C_ADTTable*) FindTable(Menu?VTSM_C_ADT:VTS_C_ADT);
	if(Table == 0 || Table->m_LastByte == 0)
		return;
	ushort OrigNumCells = (FixEndian(Table->m_LastByte) - 8 + 1) / sizeof(VTSM_C_ADTItem);

	// Clean up original table.
	memset(&Table->m_Items[0],0,OrigNumCells*sizeof(VTSM_C_ADTItem));

	// Then grow or shrink it!
	if(OrigNumCells != NumCells)
	{
		ChangeTableSize(Menu?VTSM_C_ADT:VTS_C_ADT,(NumCells-OrigNumCells)*(int)sizeof(VTSM_C_ADTItem));
		PutInt((uchar*)&Table->m_LastByte,0,NumCells*sizeof(VTSM_C_ADTItem)+8-1);
	}
	if(NumCells == 0)
	{
		// If there's no cell, it's important to actually remove the vmgm_c_adt or vtsm_c_adt or vts_c_adt.
		// table entirely. If you don't Shrink will possibly crash while loading, or processing.
		PutInt(m_Buffer, Menu ? 0xD8 : 0xE0,0);
		return;
	}

	// This needs to be fixed for ILVU VTS because the cells have to be sorted by VOBID...
	// Then copy info into table...
	int j;
	ushort Count=0;
	int OutIndex = 0;
	// In this version, the cells are stored in VOBID/CID order, de-interleaved as they would if there
	// were ILVUs.
	for(i=0;i<NumCells;i++)
	{
		// Find if we already delt with this VOBID.
		for(j=i-1;j>=0;j--)
		{
			if(Items[j].m_VOBID == Items[i].m_VOBID)
				break;
		}
		if(j==-1)
		{
			// No, this is a new VOBID. Increment VOBID count, then write cell.
			Count++;
			Table->m_Items[OutIndex].m_VOBID = FixEndian(Items[i].m_VOBID);
			Table->m_Items[OutIndex].m_CELLID = Items[i].m_CELLID;
			Table->m_Items[OutIndex].m_StartLBA = FixEndian(Items[i].m_StartLBA);
			Table->m_Items[OutIndex++].m_EndLBA = FixEndian(Items[i].m_EndLBA);
			// Now look until the end of the input array and write all the cells have have the same
			// VOBID.
			for(j=i+1;j<NumCells;j++)
			{
				if(Items[j].m_VOBID == Items[i].m_VOBID)
				{
					// Same VOBID. Write it out....
					Table->m_Items[OutIndex].m_VOBID = FixEndian(Items[j].m_VOBID);
					Table->m_Items[OutIndex].m_CELLID = Items[j].m_CELLID;
					Table->m_Items[OutIndex].m_StartLBA = FixEndian(Items[j].m_StartLBA);
					Table->m_Items[OutIndex++].m_EndLBA = FixEndian(Items[j].m_EndLBA);
				}
			}
		}
	}
	// In the end, we should have written exactly how many we had on input.
	assert(OutIndex == NumCells);

	// Write the number of distinct VOBID back into the table.
	Table->m_NumVOBID = FixEndian(Count);
	return;
}

void CIFOLoader::UpdateCellInfo(VTSM_C_ADTItem *Items, int NumItems, bool Menu)
{
	// Go through all LU, PGC, and cells and correct cell pointers if needed.
	ushort LU,PGC,CELL;
	ushort NumLU = Menu == true ? GetNumLU() : 1;
	for(LU=1;LU<=NumLU;LU++)
	{
		ushort TheLU = Menu ? LU : 0;
		ushort NumPGC = GetNumPGC(TheLU);
		int iteration;
		ushort GoodCell = 0;
		ushort GoodPGC = 0;
		for(iteration=0;iteration<2;iteration++)
		{
			// This will point to the first "good" cell (i.e. referenced).
			int NotFound = 0;
			for(PGC=1;PGC<=NumPGC;PGC++)
			{
				ushort NumCell = GetNumCells(TheLU,PGC);
				for(CELL=1;CELL<=NumCell;CELL++)
				{
					CELL_POS_INFO *Pos = GetCellPosInfo(TheLU,PGC,CELL);
					CELL_PLAYBACK_INFO *PB = GetCellPlaybackInfo(TheLU,PGC,CELL);
					// Get the LBA pointers.
					uint FirstVOBUStart = FixEndian(PB->m_FirstVOBUStartLBA);
					uint FirstILVUEnd = FixEndian(PB->m_FirstILVUEndLBA);
					uint LastVOBUStart = FixEndian(PB->m_LastVOBUStartLBA);
					uint LastVOBUEnd = FixEndian(PB->m_LastVOBUEndLBA);
					// Get the VCID and find it in our table.
					ushort VOBID = FixEndian(Pos->m_VOBID);
					uchar CELLID = Pos->m_CELLID;
					int j;
					for(j=0;j<NumItems;j++)
					{
						if(Items[j].m_CELLID == CELLID && Items[j].m_VOBID == VOBID)
							break;
					}
					if(j == NumItems)
					{
						// VCID not found!!!
						if(iteration == 0)
							MESSAGE5("PGC %d Cell %d references VCID %d/%d not found in VOB",PGC,CELL,VOBID,CELLID);
						NotFound = 1;
						if(iteration != 0 && GoodCell)
						{
							// NO! Try to find the cell these LBAs belong to?
							for(j=0;j<NumItems;j++)
							{
								if(Items[j].m_StartLBA <= FirstVOBUStart && Items[j].m_EndLBA >= LastVOBUEnd)
									break;
							}
							if(j<NumItems)
							{
								// We found the VCID these pointers refer to. Let's adjust that.
								Pos->m_CELLID = Items[j].m_CELLID;
								Pos->m_VOBID = FixEndian(Items[j].m_VOBID);
								MESSAGE5("Correcting VCID of PGC %d Cell %d to %d/%d",PGC,CELL,Items[j].m_VOBID,Items[j].m_CELLID);
							}
							else
							{
								// This cell is seriously broken. The start/end lbas don't match any cell.
								// Copy info from good cell to this one!
								CELL_POS_INFO *PosGood = GetCellPosInfo(TheLU,GoodPGC,GoodCell);
								CELL_PLAYBACK_INFO *PBGood = GetCellPlaybackInfo(TheLU,GoodPGC,GoodCell);
								memcpy(Pos,PosGood,sizeof(CELL_POS_INFO));
								memcpy(PB,PBGood,sizeof(CELL_PLAYBACK_INFO));
								MESSAGE5("Remapping PGC %d Cell %d to %d/%d",PGC,CELL,FixEndian(PosGood->m_VOBID),PosGood->m_CELLID);
							}

						}
						continue;
					}
					if(GoodCell == 0) {GoodCell = CELL; GoodPGC = PGC;}
					// Only do that if angle flag isn't set!
					if((PB->m_Flags & 0x30) != 0 || (PB->m_Flags & 0x04) != 0)
						continue;
					// We found the cell referenced by this PGC Cell. Let's check that our LBAs are valid.
					if(FirstVOBUStart < Items[j].m_StartLBA || FirstVOBUStart >= Items[j].m_EndLBA) 
					{
						FirstVOBUStart = Items[j].m_StartLBA;
						MESSAGE3("PGC %d - Cell %d, adjusting First Vobu Start LBA",PGC,CELL);
					}
					if(LastVOBUStart < Items[j].m_StartLBA || LastVOBUStart > Items[j].m_EndLBA) 
					{
						LastVOBUStart = Items[j].m_StartLBA;
						MESSAGE3("PGC %d - Cell %d, Adjusting Last Vobu Start LBA",PGC,CELL);
					}
					if(LastVOBUEnd < Items[j].m_StartLBA || LastVOBUEnd > Items[j].m_EndLBA) 
					{
						LastVOBUEnd = Items[j].m_EndLBA;
						MESSAGE3("PGC %d - Cell %d, Adjusting Last Vobu End LBA",PGC,CELL);
					}
					// Now put the possibly modified values back.
					PB->m_FirstVOBUStartLBA = FixEndian(FirstVOBUStart);
					PB->m_FirstILVUEndLBA = FixEndian(FirstILVUEnd);
					PB->m_LastVOBUStartLBA = FixEndian(LastVOBUStart);
					PB->m_LastVOBUEndLBA = FixEndian(LastVOBUEnd);
				}
			}
			// Do that again, if we found a good cells and some of the cells were not found in the vob,
			// they will be modified...
			if(GoodCell == 0 || NotFound == 0) break;
		}
	}
}

void CIFOLoader::UpdateCellInfo(CCell* TheCell, bool AddToVTSM_C_ADT, bool Menu)
{
	uint FirstVOBUStartLBA = TheCell->m_StartLBA;
	int LastNav = TheCell->FindPrevNavPack(TheCell->m_NumPacks);
	uint LastVOBUStartLBA = LastNav+TheCell->m_StartLBA;
	uint LastVOBUEndLBA = TheCell->m_StartLBA + TheCell->m_NumPacks-1;

	// Now, find every occurence of this cell.
	ushort LU[512], PGC[512], CELL[512];
	int NumFound = FindThisCell(TheCell->m_VOBID, TheCell->m_CELLID, (Menu)?LU:0, PGC, CELL);

	int i;
	for(i=0;i<NumFound;i++)
	{
		// Correct sector pointers for this PGC Cell.
		CELL_PLAYBACK_INFO* PBInfo = GetCellPlaybackInfo((Menu)?LU[i]:0,PGC[i],CELL[i]);
		PutInt((uchar*)&PBInfo->m_FirstVOBUStartLBA,0,FirstVOBUStartLBA);
		PutInt((uchar*)&PBInfo->m_LastVOBUStartLBA,0,LastVOBUStartLBA);
		PutInt((uchar*)&PBInfo->m_LastVOBUEndLBA,0,LastVOBUEndLBA);
		// Update the cell duration.
		CDVDTime CellDur(TheCell->m_EndPTS - TheCell->m_StartPTS);
		CellDur.WriteToBuffer((uchar*)&PBInfo->m_PBTime);
		// And the PGC time.
		UpdatePGCPlayBackTime((Menu)?LU[i]:0,PGC[i]);
	}

	// Now, look for the cell in the VTSM_C_ADT table..
	UpdateVTS_C_ADT(TheCell, AddToVTSM_C_ADT, Menu);

	// Add the cell's VOBUs to the list in the IFO.
	int NextNavPack = 0;
	while(NextNavPack != -1)
	{
		m_VOBUStartLBA[m_NumVOBUs] = TheCell->m_StartLBA + NextNavPack;
		m_NumVOBUs++;
		assert(MAX_NUM_VOBUS >= m_NumVOBUs);
		NextNavPack = TheCell->FindNextNavPack(NextNavPack);
	}
}

static int InNavPackNotFound = false;	// Ugly!
int FindMatchingNavPack(int ToMatch, int *InNavPacks, int *OutNavPacks, int NumNavPacks,
						int OutMin=-1)
{
	int i;
	InNavPackNotFound = false;
	for(i=0;i<NumNavPacks;i++)
	{
		if(InNavPacks[i] >= ToMatch && OutNavPacks[i] > OutMin)
		{
			if(InNavPacks[i] != ToMatch)
				InNavPackNotFound = true ;
			return OutNavPacks[i];
		}
	}
	return OutNavPacks[NumNavPacks-1];
}

void CIFOLoader::UpdateNavPackInfo(int *InNavPacks, int *OutNavPacks, int NumNavPacks, bool Menu)
{
	ushort LU = 0;
	ushort NumLUs = GetNumLU();
	if(!Menu) NumLUs = 1;

	for(LU=1;LU<=NumLUs;LU++)
	{
		ushort NumPGCs = GetNumPGC(Menu?LU:0);
		ushort PGC;
		for(PGC=1;PGC<=NumPGCs;PGC++)
		{
			ushort NumCells = GetNumCells(Menu?LU:0,PGC);
			ushort Cell;
			for(Cell=1;Cell<=NumCells;Cell++)
			{
				CELL_PLAYBACK_INFO* PBInfo = GetCellPlaybackInfo(Menu?LU:0,PGC,Cell);
				uint Temp = GetInt((uchar*)&PBInfo->m_FirstVOBUStartLBA,0);
				// Remember that the last navpack is a dummy one pointing after the end
				// of the VOB. We can't use that as a start LBA.
				Temp = FindMatchingNavPack(Temp,InNavPacks,OutNavPacks,NumNavPacks-1);
				PutInt((uchar*)&PBInfo->m_FirstVOBUStartLBA,0,Temp);
				uint FirstVobuStart = Temp;
				if(InNavPackNotFound)
					MESSAGE3("PGC %d - Cell %d, Original entry sector didn't point to a navpack. Fixed",
						PGC,Cell);

				Temp = GetInt((uchar*)&PBInfo->m_LastVOBUStartLBA,0);
				// Same here.
				Temp = FindMatchingNavPack(Temp,InNavPacks,OutNavPacks,NumNavPacks-1);
				uint LastVobuStart = Temp;
				assert(LastVobuStart>=FirstVobuStart);
				PutInt((uchar*)&PBInfo->m_LastVOBUStartLBA,0,Temp);
				if(InNavPackNotFound)
					MESSAGE3("PGC %d - Cell %d, Original last VOBU start LBA didn't point to a navpack. Fixed",
						PGC,Cell);

				Temp = GetInt((uchar*)&PBInfo->m_LastVOBUEndLBA,0)+1;
				// But here, we can use the last navpack because we're looking for an end
				// LBA, not a start one. Also, I want to make sure that this LBA will be after
				// LastVobuStart.
				Temp = FindMatchingNavPack(Temp,InNavPacks,OutNavPacks,NumNavPacks,LastVobuStart);
				if(Temp >= 1) Temp --;
				assert(Temp>=LastVobuStart);
				PutInt((uchar*)&PBInfo->m_LastVOBUEndLBA,0,Temp);
				if(InNavPackNotFound)
					MESSAGE3("PGC %d - Cell %d, Original last VOBU end LBA + 1 didn't point to a navpack. Fixed",
						PGC,Cell);

				Temp = GetInt((uchar*)&PBInfo->m_FirstILVUEndLBA,0)+1;
				if(Temp > 1)
				{
					// same here...
					Temp = FindMatchingNavPack(Temp,InNavPacks,OutNavPacks,NumNavPacks);
					if(Temp >= 1) Temp --;
					PutInt((uchar*)&PBInfo->m_FirstILVUEndLBA,0,Temp);
					if(InNavPackNotFound)
						MESSAGE3("PGC %d - Cell %d, Original first ILVU end LBA didn't point to a navpack. Fixed",
							PGC,Cell);
				}
			}
		}
	}

	int i;
	m_NumVOBUs = 0;
	assert(MAX_NUM_VOBUS >= NumNavPacks-1);
	for(i=0;i<NumNavPacks;i++)
	{
		// Don't write twice the same guy!
		if(i == 0 || OutNavPacks[i] != OutNavPacks[i-1])
			m_VOBUStartLBA[m_NumVOBUs++] = OutNavPacks[i];
	}
	// NOTE I'm getting rid of the last one, which by convention is supposed to point past
	// the end of the VOB, as if there was a navpack there.
	if(m_NumVOBUs > 0)
		m_NumVOBUs--;
	
	// Now update the VOBU_ADMAP table.
	UpdateVOBUADMAP(Menu);
}

void CIFOLoader::UpdatePGCPlayBackTime(ushort LU, ushort PGC, bool FixZeroTimes, bool AdjustFormat)
{
	// Find this PGC...
	PGC_HEADER*	Header = GetPGCHeader(LU, PGC);

	// Get number of cells in this PGC
	uchar NumCells = GetNumCells(LU, PGC);

	// This will hold the total playback time of the PGC.
	CDVDTime TotalPGCTime(0);
	// Now go through all the cells in this PGC and add their playback times.
	uchar Cell;
	VFormat Format = VideoFormat;
	for(Cell=0;Cell<NumCells;Cell++)
	{
		CELL_PLAYBACK_INFO*	PBInfo = GetCellPlaybackInfo(LU, PGC, Cell+1);
		CDVDTime CellPBTime;
		// Read time, and add.
		CellPBTime.ReadFromBuffer((uchar*)(&PBInfo->m_PBTime));
		if(!AdjustFormat) {Format = CellPBTime.m_format; TotalPGCTime.m_format = Format;}
		if(FixZeroTimes && CellPBTime.IsZero())
		{
			CellPBTime.m_frames = (Format == FORMAT_PAL) ? 12 : 18;
			CellPBTime.WriteToBuffer((uchar*)(&PBInfo->m_PBTime));
		}
		// Adjust elapsed time! But only if there's no angle, or this is the first block...
		if((PBInfo->m_Flags & 0x10) == 0 || 
			((PBInfo->m_Flags & 0x40) != 0 && (PBInfo->m_Flags & 0x80) == 0))
			TotalPGCTime += CellPBTime;
	}
	// Done, let's write the time back to the PGC info...
	TotalPGCTime.WriteToBuffer((uchar*)&Header->m_PlaybackTime);
}

// Renumbers all the VCIDs in all the pgc cells according to the map.
void CIFOLoader::RenumberVCID(VCIDMapItem* Map, int MapLength, bool Menu)
{
	int i,j,k,n;
	int NumLU = Menu?GetNumLU():1;
	// Go through all LUs.
	for(j=1;j<=NumLU;j++)
	{
		int LU = (Menu)?j:0;
		int NumPGC = GetNumPGC(LU);
		// And all PGCs.
		for(i=1;i<=NumPGC;i++)
		{
			int NumCell = GetNumCells(LU,i);
			// And all cells.
			for(k=1;k<=NumCell;k++)
			{
				// Now see if this VCID needs to be remapped.
				CELL_POS_INFO *Temp = GetCellPosInfo(LU,i,k);
				ushort VOBID = FixEndian(Temp->m_VOBID);
				uchar CELLID = Temp->m_CELLID;
				for(n=0;n<MapLength;n++)
				{
					if(Map[n].InVOBID == VOBID && Map[n].InCELLID == CELLID)
					{
						// Yes, remap it!
						Temp->m_VOBID = FixEndian(Map[n].OutVOBID);
						Temp->m_CELLID = Map[n].OutCELLID;
						break;
					}
				}
			}
		}
	}
}

void CIFOLoader::UpdateVOBUADMAP(bool Menu)
{
	uchar* Ptr;
	IFOTable VOBU_ADMAP = (Menu) ? VTSM_VOBU_ADMAP : VTS_VOBU_ADMAP;
	
	Ptr = FindTable(VOBU_ADMAP);
	if(Ptr == 0) return;
	
	// Find how many VOBUs were declared in the table.
	if(GetInt(Ptr,0)+Ptr > m_Buffer+m_IFOSizekB*1024) 
		return;
	int NumVOBUS = (GetInt(Ptr,0) + 1)/4 - 1;
	
	// Now grow it, or shrink it, as necessary.
	int Delta = m_NumVOBUs - NumVOBUS;
	// If we shrink the table, replace the old values by 0s.
	int i;
	for(i=m_NumVOBUs;i<NumVOBUS;i++)
		PutInt(Ptr,(i+1)*4,0);
	// Then change the table size...
	ChangeTableSize(VOBU_ADMAP,Delta*4);

	// And write the VOBU info:
	for(i=0;i<m_NumVOBUs;i++)
		PutInt(Ptr,(i+1)*4,m_VOBUStartLBA[i]);

	if(m_NumVOBUs == 0)
	{
		// If there's no cell, it's important to actually remove the vts(m)_vobu_admap
		// table entirely. If you don't Shrink will possibly crash while loading, or processing.
		PutInt(m_Buffer, Menu ? 0xDC : 0xE4,0);
		return;
	}

	// Finally, update the end byte of the table.
	PutInt(Ptr,0,3+m_NumVOBUs*4);
}

void CIFOLoader::UpdateVOBUADMAP(uint* StartLBA, int Num, bool Menu)
{
	uchar* Ptr;
	IFOTable VOBU_ADMAP = (Menu) ? VTSM_VOBU_ADMAP : VTS_VOBU_ADMAP;
	
	Ptr = FindTable(VOBU_ADMAP);
	if(Ptr == 0) return;
	
	// Find how many VOBUs were declared in the table.
	int NumVOBUS = (GetInt(Ptr,0) + 1)/4 - 1;
	
	// Now grow it, or shrink it, as necessary.
	int Delta = Num - NumVOBUS;
	// If we shrink the table, replace the old values by 0s.
	int i;
	for(i=Num;i<NumVOBUS;i++)
		PutInt(Ptr,(i+1)*4,0);
	// Then change the table size...
	ChangeTableSize(VOBU_ADMAP,Delta*4);

	// And write the VOBU info:
	for(i=0;i<Num;i++)
		PutInt(Ptr,(i+1)*4,StartLBA[i]);

	if(Num == 0)
	{
		// If there's no cell, it's important to actually remove the vts(m)_vobu_admap
		// table entirely. If you don't Shrink will possibly crash while loading, or processing.
		PutInt(m_Buffer, Menu ? 0xDC : 0xE4,0);
		return;
	}

	// Finally, update the end byte of the table.
	PutInt(Ptr,0,3+Num*4);
}


void CIFOLoader::UpdateTMAPTI(int *InNavPacks, int *OutNavPacks, int NumNavPacks)
{
	uchar* TMAPTI;
	TMAPTI = FindTable(VTS_TMAPTI);
	if(TMAPTI == 0) return;

	m_AddressError = 0;
	uchar* LastPtr = TMAPTI;

	ushort NumMaps = GetShort(TMAPTI,0);
	ushort i;

	for(i=0;i<NumMaps;i++)
	{
		// Find offset to map:
		uchar* Ptr = TMAPTI + GetInt(TMAPTI,8+4*i);
		ushort NumEntries = GetShort(Ptr,2);
		ushort j,k;
		uint LastEntry = 0xFFFFFFFF;
		for(j=k=0;j<NumEntries;j++)
		{
			// Get original entry, remember discontinuity, 
			uint OrigEntry = GetInt(Ptr,4+j*4);
			uint Discont = OrigEntry & 0x80000000;
			OrigEntry &= 0x7FFFFFFF;
			// Find matching output lba.
			uint NewEntry = FindMatchingNavPack(OrigEntry,InNavPacks,OutNavPacks,NumNavPacks-1);
			if(NewEntry != LastEntry)
			{
				// Put discontinuity back.
				LastEntry = NewEntry;
				NewEntry |= Discont;
				PutInt(Ptr,4+k*4,NewEntry);
				k++;
			}
			if(m_AddressError) return;
		}
		// Done... We might have fewer entries! Write that back.
		PutShort(Ptr,2,k);
		// This is the end of this table: Ptr+4+4*k+3. Remember the furthest...
		LastPtr = max(LastPtr,Ptr+4+4*k+3);
	}
	// Now put back the size of the table, and change the size...
	uint OrigSize = GetInt(TMAPTI,4);
	uint NewSize = LastPtr-TMAPTI;
	if(m_AddressError) return;
	ChangeTableSize(VTS_TMAPTI,(int)NewSize-(int)OrigSize);
	PutInt(TMAPTI,4,NewSize);
}

typedef struct
{
	// All sizes in sectors.
	int		m_VTSMenuFinalSize;
	int		m_IFOFinalSize;
	int		m_VTSTitleSize;
} VTSInfo;

void CIFOLoader::UpdateVTSSectors(bool CorrectNumVTS)
{
try{
	if(!m_IsManager)
	{
		// Now update the size of the IFO table. To do this, look for all the tables and compute size
		// in LBA.
		uint MaxEndLBA = 0;
		MaxEndLBA = max(MaxEndLBA, GetAbsoluteEndAddress(VTS_PTT_SRPT)/SECTOR_SIZE);
		MaxEndLBA = max(MaxEndLBA, GetAbsoluteEndAddress(VTS_PGCI)/SECTOR_SIZE);
		MaxEndLBA = max(MaxEndLBA, GetAbsoluteEndAddress(VTSM_PGCI_UT)/SECTOR_SIZE);
		MaxEndLBA = max(MaxEndLBA, GetAbsoluteEndAddress(VTS_TMAPTI)/SECTOR_SIZE);
		MaxEndLBA = max(MaxEndLBA, GetAbsoluteEndAddress(VTSM_C_ADT)/SECTOR_SIZE);
		MaxEndLBA = max(MaxEndLBA, GetAbsoluteEndAddress(VTSM_VOBU_ADMAP)/SECTOR_SIZE);
		MaxEndLBA = max(MaxEndLBA, GetAbsoluteEndAddress(VTS_C_ADT)/SECTOR_SIZE);
		MaxEndLBA = max(MaxEndLBA, GetAbsoluteEndAddress(VTS_VOBU_ADMAP)/SECTOR_SIZE);
		// Now put that in 0x1c
		PutInt(m_Buffer,0x1c,MaxEndLBA);
		m_IFOSizekB = (MaxEndLBA+1)*2;
		m_MaxAddress = m_Buffer+m_IFOSizekB*1024-1;
		// Set C0, first sector of VTS_xx_0.VOB.
		PutInt(m_Buffer,0xc0,m_IFOSizekB/2);

		// Set C4, first sector of VTS_xx_1.VOB.
		// Find size of VTS_xx_0.VOB
		int MenuSize = GetFileSizekB(ChangeExt(m_IFOFileName,".VOB"))/2;
		// If there's no video_ts.vob, put a 0 here.
		if(MenuSize == 0)
			PutInt(m_Buffer,0xc0,0);
		// Then find cumulative size of title vobs.
		int i;
		int TitleSize = 0;
		for(i=1;i<10;i++)
		{
			char Temp[512];
			sprintf(Temp,"%sVTS_%02d_%d.VOB",GetDirPath(m_IFOFileName),m_VTSNum,i);
			TitleSize += GetFileSizekB(Temp)/2;
		}
		PutInt(m_Buffer,0xc4,m_IFOSizekB/2+MenuSize);
		PutInt(m_Buffer,0x0c,2*m_IFOSizekB/2+MenuSize+TitleSize-1);
	}
	else
	{
		// Find how many VTS are present.
		int i,j;
		for(i=1;i<100;i++)
		{
			if(GetFileSizekB(CatStr(GetDirPath(m_IFOFileName),IFONames[i])) == 0)
				break;
		}
		int NumVTS = i-1;
		if(CorrectNumVTS)
			UpdateNumVTS(NumVTS);

		// Now update the size of the IFO table. To do this, look for all the tables and compute size
		// in LBA.
		uint MaxEndLBA = 0;
		MaxEndLBA = max(MaxEndLBA, GetAbsoluteEndAddress(TT_SRPT)/SECTOR_SIZE);
		MaxEndLBA = max(MaxEndLBA, GetAbsoluteEndAddress(VMGM_PGCI_UT)/SECTOR_SIZE);
		MaxEndLBA = max(MaxEndLBA, GetAbsoluteEndAddress(VMG_PTL_MAIT)/SECTOR_SIZE);
		MaxEndLBA = max(MaxEndLBA, GetAbsoluteEndAddress(VMG_VTS_ATRT)/SECTOR_SIZE);
		MaxEndLBA = max(MaxEndLBA, GetAbsoluteEndAddress(VMG_TXTDT_MG)/SECTOR_SIZE);
		MaxEndLBA = max(MaxEndLBA, GetAbsoluteEndAddress(VTSM_C_ADT)/SECTOR_SIZE);
		MaxEndLBA = max(MaxEndLBA, GetAbsoluteEndAddress(VTSM_VOBU_ADMAP)/SECTOR_SIZE);
		// Now put that in 0x1c
		PutInt(m_Buffer,0x1c,MaxEndLBA);
		m_IFOSizekB = (MaxEndLBA+1)*2;
		m_MaxAddress = m_Buffer+m_IFOSizekB*1024-1;
		// Set C0, first sector of video_ts.vob.
		PutInt(m_Buffer,0xc0,m_IFOSizekB/2);

		int MenuSize = GetFileSizekB(ChangeExt(m_IFOFileName,".VOB"))/2;
		PutInt(m_Buffer,0x0c,2*m_IFOSizekB/2+MenuSize-1);
		// If there's no video_ts.vob, put a 0 here.
		if(MenuSize == 0)
			PutInt(m_Buffer,0xc0,0);
		// Now update the rest of the pointers in the TT_SRPT. For this I need to gather all the
		// file sizes. Start with the manager (this one!).
		VTSInfo* VTSInfos = new VTSInfo[100];
		VTSInfos[0].m_IFOFinalSize = m_IFOSizekB/2;
		VTSInfos[0].m_VTSMenuFinalSize = MenuSize;
		VTSInfos[0].m_VTSTitleSize = 0;
		uint VMGSize = 2*VTSInfos[0].m_IFOFinalSize+VTSInfos[0].m_VTSMenuFinalSize;

		// Then the vobs.
		uint StartOfVTS[100];
		StartOfVTS[0] = 0;
		int LastIFO = 0;
		for(i=1;i<100;i++)
		{
			VTSInfos[i].m_IFOFinalSize = GetFileSizekB(CatStr(GetDirPath(m_IFOFileName),IFONames[i]))/2;
			if(VTSInfos[i].m_IFOFinalSize != 0) 
				LastIFO = i;
			VTSInfos[i].m_VTSMenuFinalSize = GetFileSizekB(ChangeExt(CatStr(GetDirPath(m_IFOFileName),IFONames[i]),".VOB"))/2;
			int Size = 0;
			for(j=1;j<10;j++)
			{
				char Temp[512];
				sprintf(Temp,"%sVTS_%02d_%d.VOB",GetDirPath(m_IFOFileName),i,j);
				int ThisSize = GetFileSizekB(Temp)/2;
				if(ThisSize == 0 && j > 0)
					break;
				Size += ThisSize;
			}
			VTSInfos[i].m_VTSTitleSize = Size;
			StartOfVTS[i] = StartOfVTS[i-1]+2*VTSInfos[i-1].m_IFOFinalSize+VTSInfos[i-1].m_VTSMenuFinalSize+
				VTSInfos[i-1].m_VTSTitleSize;
		}

		// LastIFO indicates the number of titlesets. Update that in the header.
		uchar* Ptr = FindTable(TT_SRPT);
		if(Ptr)
		{
			uint NumTitles = GetShort(Ptr,0);
			for(i=0;i<(int)NumTitles;i++)
			{
				// Get VTS number:
				uchar VTSNumber = Ptr[8+i*12+6];
				if(VTSNumber <= LastIFO)
					PutInt(Ptr,8+i*12+8,StartOfVTS[VTSNumber]);
			}
		}
		delete[] VTSInfos;
	}
}
	catch(...)
	{
		ERROR1("UVS ERROR");
	}
}

void CIFOLoader::UpdateNumVTS(int NumVTS)
{
	int i;
	// LastIFO indicates the number of titlesets. Update that in the header.
	PutShort(m_Buffer,0x3e,NumVTS);
	// Also change that in the parental management table.
	uchar* Ptr = FindTable(VMG_PTL_MAIT);
	if(Ptr)
	{
		int OldNumVTS = GetShort(Ptr,2);
		_ASSERT(OldNumVTS >= NumVTS);
		if(NumVTS > OldNumVTS) return;
		int NumCountries = GetShort(Ptr,0);
		uchar* PtrD;
		for(i=0;i<NumCountries;i++)
		{
			// For this country, find offset to PTL_MAIT.
			int Offset = GetShort(Ptr,8*(i+1)+4);
			uchar* PtrS = Ptr+Offset;	// Mask source
			PtrD = Ptr+Offset;	// Mask destination. I'm using the same offest.
			// For level L and vts V, the source mask is at Ptr + Offset + 2*L*(OldNumVTS+1)+V;
			int level;
			for(level=0;level<8;level++)
			{
				// Copy the masks corresponding to the present VTSs (2 bytes each).
				PtrS = Ptr+Offset+2*level*(OldNumVTS+1);
				memmove(PtrD,PtrS,(NumVTS+1)*2);
				// And update destination pointer.
				PtrD += (NumVTS+1)*2;
			}
		}
		// Finally, update end address;
		uint FOO = GetInt(Ptr,4);
		PutInt(Ptr,4,PtrD-1-Ptr);
		// And write new number of VTS.
		PutShort(Ptr,2,NumVTS);
	}

	Ptr = FindTable(VMG_VTS_ATRT);
	if(Ptr) 
	{
		PutShort(Ptr,0,NumVTS);
		if(NumVTS)
		{
			// I need to modify the end address of this table... First find offset to last VTS_ATRT table.
			uint LastOffset = GetInt(Ptr,8+4*(NumVTS-1));
			// Then read the new end address (end address of last VTS_ATRT table)
			uint NewEndAddress = LastOffset + GetInt(Ptr,LastOffset);
			// Then change the table size and write the size back.
			uint OldEndAddress = GetInt(Ptr,4);
			ChangeTableSize(VMG_VTS_ATRT,(int)NewEndAddress-(int)OldEndAddress);
			PutInt(Ptr,4,NewEndAddress);
		}
		else
			PutInt(Ptr,4,7);
	}


	// Now update TT_SRTP. If a title references a VTS that does not exist, remove it.
	Ptr = FindTable(TT_SRPT);
	// This will indicate whether there's a title defined in this VTS.
	bool TitleInVTS[100];
	memset(TitleInVTS,0,100*sizeof(bool));
	if(Ptr)
	{
		// This will contain the new map of titles. TitleMap[i] indicates the new title number that corresponds to
		// title i in the original. Indexes start at 1.
		uchar TitleMap[256];
		memset(TitleMap,0xFF,256);
		uchar* PtrS;
		uchar* PtrD;
		ushort NumTitles = GetShort(Ptr,0);
		ushort NewNumTitles = 0;
		PtrD = Ptr + 8;
		for(i=0;i<(int)NumTitles;i++)
		{
			// Get VTS number:
			uchar VTSNumber = Ptr[8+i*12+6];
			PtrS = Ptr + 8 + 12*i;
			if(VTSNumber <= NumVTS)
			{
				memmove(PtrD,PtrS,12);
				PtrD += 12;
				TitleMap[i+1] = NewNumTitles+1;
				NewNumTitles++;
			}
			if(VTSNumber < 100 && VTSNumber >= 0)
				TitleInVTS[VTSNumber] = true;
		}
		// Update number of titles.
		PutShort(Ptr,0,NewNumTitles);
		// Update end byte of table.
		PutInt(Ptr,4,PtrD-1-Ptr);
		// Now check that all the VTS have a title in them.
		for(i=1;i<NumVTS;i++)
		{
			if(TitleInVTS[i] == false)
			{
				MESSAGE2("No title was defined for VTS %d -- Adding one!",i);
				// Add an entry to TT_SRPT.
				ChangeTableSize(TT_SRPT,12);
				uchar TitleInfo[12] = {0x14,0x01,0x00,0x01,0x00,0x00,i,0x01,0x00,0x00,0x00,0x00};
				memcpy(Ptr + 8 + 12*NewNumTitles,TitleInfo,12);
				NewNumTitles++;
				PutShort(Ptr,0,NewNumTitles);
				AddInt(Ptr,4,12);
			}
		}
		// Now, go through all commands in all the PGCs of the manager, and change all Jump_TT to reflect the remapping!
		ushort pgc,lu;
		uchar* Commands;
		uchar* ThisCommand;
		ushort cmd,NumCommands;
		for(lu=0;lu<=GetNumLU();lu++)
		{
			for(pgc=0;pgc<GetNumPGC(lu);pgc++)
			{
				NumCommands = GetNumCommands(lu,pgc+1,3);
				GetCommands(lu,pgc+1,0,&Commands);
				for(cmd=0;cmd<NumCommands && Commands != 0;cmd++)
				{
					// See if this is a JumpTT
					ThisCommand = Commands+cmd*8;
					ushort OpCode = GetShort(ThisCommand,0);
					if(TYP == 1 && DIR == 1 && CMD == 2)
					{
						uchar OldTTN = ThisCommand[5];
						if(OldTTN > NumTitles || OldTTN == 0 || TitleMap[OldTTN] == 0xFF)
						{
							// This title was removed. Change the JumpTT command to a nop!
							memset(ThisCommand,0,8);
							MESSAGE5("Changing JumpTT %d to NOP in LU %d PGC %d Cmd %d",OldTTN,lu,pgc,cmd);
						}
						else
						{
							// This is a title that wasn't removed, but simply remapped.
							ThisCommand[5] = TitleMap[OldTTN];
							if(TitleMap[OldTTN] != OldTTN)
								MESSAGE6("Remapping JumpTT %d to %d in LU %d PGC %d Cmd %d",OldTTN,TitleMap[OldTTN],
									lu,pgc,cmd);
						}
					}
					// Also, if this is a JUMPSS_VTSM...
					else if(TYP == 1 && DIR == 1 && CMD == 6 && ThisCommand[4] > NumVTS )
					{
						memset(ThisCommand,0,8);
						MESSAGE5("Changing JUMPSS_VTSM %d to NOP in LU %d PGC %d Cmd %d",ThisCommand[4],lu,pgc,cmd);
					}
				}
			}
		}

	}
}

uchar CIFOLoader::GetRegionCodeMask()
{
	return m_Buffer ? m_Buffer[0x23] : 0;	
}

int CIFOLoader::BypassPGCPlayback(ushort LU, ushort PGC)
{
	// Procedure:
	// - Copy post commands at end of precommands with a nop in between.
	// - Change all breaks to gotos
	// - Change all LinkCN and LinkPG to gotos.

	int i;
	// Can't do it if the total number of commands will exceed 128.
	if(GetNumCommands(LU,PGC,3)+GetNumCommands(LU,PGC,1)+1 >= 128)
	{
		MESSAGE3("Can't bypass LU %d PGC %d, too many commands",LU,PGC);
		return -1;
	}
	// TODO: Check cell commands for Jump or Calls...
	uchar*	CellCommands;
	ushort NumCell = 0;
	if(GetCommands(LU, PGC, 2, &CellCommands))
		NumCell = GetShort(GetCommands(LU, PGC, 2, &CellCommands),0);
	for(i=0;i<NumCell;i++)
	{
		ushort OpCode = GetShort(CellCommands+i*8,0);
		// If one of the cell commands is Exit or Jump or Call we can't bypass.
		if(TYP == 1 && DIR == 1 && (CMD > 0 && CMD <= 8))
		{
			MESSAGE4("Can't bypass LU %d PGC %d, command %d jumps out of PGC",LU,PGC,i+1);
			return -1;
		}
	}

	// Add nop command at end of pre commands.
	uchar NopCmd[8] = {0,0,0,0,0,0,0,0};
	AddCommand(LU, PGC, NopCmd, 0);
	
	// Now get numbers of pre/postcommands and pointers...
	uchar*	PreCommands;
	ushort NumPre = GetShort(GetCommands(LU, PGC, 0, &PreCommands),0);
	uchar*	PostCommands;
	ushort NumPost = GetShort(GetCommands(LU, PGC, 1, &PostCommands),0);
	uchar TargetLine = (uchar)NumPre;
	uchar CmdHex[8];
	// Add all postcommands at end of precommands.
	for(i=0;i<NumPost;i++)
	{
		// You need to re-read that because the address changes.
		GetCommands(LU, PGC, 1, &PostCommands);
		// I need to do that because PostCommand points to the absolute location of the post commands,
		// which will change when AddCommand is called!
		// Add NumPre to all gotos!
		// A goto has TYP=0, CMD=1
		memcpy(CmdHex,PostCommands+i*8,8);
		ushort OpCode = ::GetShort(CmdHex,0);
		if(TYP == 0 && CMD == 1) 
			CmdHex[7] += (char) NumPre;
		AddCommand(LU, PGC, CmdHex, 0);
	}

	// Now go through precommands and replace what needs to be!
	for(i=0;i<NumPre;i++)
	{
		uchar* Cmd = PreCommands+i*8;
		ushort OpCode = GetShort(Cmd,0);
		if(TYP == 0 && CMD == 2)
		{
			// To change break into goto, change CMD (B1b3-0) to 1 and line (B7b7-0) to line #.
			Cmd[1] &= 0xF0; Cmd[1] |= 0x01;
			Cmd[7] = TargetLine;
		}
		else if(TYP == 1 && DIR == 0 && (CMD == 7 || CMD == 6))
		{
			// To change LinkCN or LinkPG into goto, change CMD (B1b3-0) to 1 and line (B7b7-0) to line #.
			// And TYP (B0b7-5) to 0, and B6 (button highlight) to 0.
			Cmd[1] &= 0xF0; Cmd[1] |= 0x01; Cmd[6] = 0;
			Cmd[7] = TargetLine; Cmd[0] &= 0x1F;
		}
	}
//	WriteIFO();
	return 0;
}

void CIFOLoader::RemoveVTSTitle(int title)
{
	uchar* VTSTitles = FindTable(VTS_PTT_SRPT);
	if(!VTSTitles) return ;
	int NumTitles = GetShort(VTSTitles,0);
	if(NumTitles == 1)
	{
		RemoveTable(VTS_PTT_SRPT);
		return;
	}
	uint EndOffset = GetInt(VTSTitles,4)+1;
	uint Offset = GetInt(VTSTitles,8+(title-1)*4);
	uint NextOffset = title==NumTitles ? EndOffset : GetInt(VTSTitles,8+(title)*4) ;
	int NumChapter = (NextOffset - Offset)/4;
	// Move data.
	MoveDataInTable(VTS_PTT_SRPT,VTSTitles+NextOffset,-4*NumChapter);
	AddShort(VTSTitles,0,-1);
	// Change pointers of title past it.
	int tit;
	for(tit=title;tit<NumTitles;tit++)
	{
		uint Offset = GetInt(VTSTitles,8+(tit-1)*4);
		uint NextOffset = GetInt(VTSTitles,8+tit*4);
		PutInt(VTSTitles,8+(tit-1)*4,NextOffset-4*NumChapter);
	}
}

void CIFOLoader::RemoveDVDTitle(int title)
{
	_ASSERT(m_IsManager);
	if(!m_IsManager || FindTable(TT_SRPT) == 0) return;

	uchar* Ptr = FindTable(TT_SRPT);
	int NumTitles = GetShort(Ptr,0);
	if(title > NumTitles) return;
	MoveDataInTable(TT_SRPT,Ptr+8+(title)*12,-12);
	AddShort(Ptr,0,-1);
	if(NumTitles == 1)
		RemoveTable(TT_SRPT);
}


void CIFOLoader::RemovePGC(ushort LU, ushort PGC)
{
	_ASSERT(LU==0);
	if(LU) return;
	ushort NumPGC = GetNumPGC(LU);
	if(PGC > NumPGC || PGC == 0) return;

	// If this PGC was referenced by a title, let's remove the title...
	// But only if the PGC is an entry PGC.
	uchar*	PGCI = FindTable(VTS_PGCI);
	if(PGCI == 0) return;
	// Set the title number.
	bool IsEntry = (PGCI[8+(PGC-1)*8]&0x80) != 0;

	uchar* VTSTitles = FindTable(VTS_PTT_SRPT);
	if(VTSTitles && IsEntry)
	{
		int NumTitles = GetShort(VTSTitles,0);
		int title;
		for(title=NumTitles-1;title>=0;title--)
		{
			uint Offset = GetInt(VTSTitles,8+title*4);
			uint EndOffset = GetInt(VTSTitles,4)+1;
			uint NextOffset = title<NumTitles-1 ? GetInt(VTSTitles,8+(title+1)*4) : EndOffset;
			int NumChapter = NextOffset - Offset;
			NumChapter /= 4;
			int chap;
			for(chap=0;chap<NumChapter;chap++)
			{
				if(GetShort(VTSTitles,Offset+chap*4) == PGC)
					break;
			}
			if(chap == NumChapter) continue;
			MESSAGE2("Removing title %d",title+1);
			RemoveVTSTitle(title+1);
			NumTitles --;
		}
	}

	
	IFOTable Table = LU == 0 ? VTS_PGCI : VTSM_PGCI_UT;
	int Delta;
	uchar* Ptr = FindTable(Table);
	// To remove a PGC, you simply move everything starting at the PGC after it.
	if(PGC < NumPGC && Ptr != 0)
	{
		uchar* Ptr2 = (uchar*)GetPGCHeader(LU,PGC+1);
		uchar* Ptr1 = (uchar*)GetPGCHeader(LU,PGC);
		Delta = (int)Ptr1-(int)Ptr2;
		MoveDataInTable(Table, Ptr2, (int)Ptr1-(int)Ptr2);
		// Move the pointers in VTS_PGCI and change offset to PGC headers...
		int i;
		for(i=PGC;i<=NumPGC;i++)
		{
			if(i<NumPGC)
			{
				memcpy(Ptr+i*8,Ptr+(i+1)*8,8);
				AddInt(Ptr+i*8+4,0,Delta);
			}
			// Now, all the PGCs after PGC have to be renumbered... Fix the table of titles...
			uchar* VTSTitles = FindTable(VTS_PTT_SRPT);
			if(!VTSTitles) continue;
			uint NumTitles = GetShort(VTSTitles,0);
			uint EndOffset = GetInt(VTSTitles,4)+1;
			uint title;
			for(title=0;title<NumTitles;title++)
			{
				uint Offset = GetInt(VTSTitles,8+title*4);
				int NumChapter = ((title<NumTitles-1 ? GetInt(VTSTitles,8+(title+1)*4) : EndOffset) - Offset);
				NumChapter /= 4;
				int chap;
				for(chap=0;chap<NumChapter;chap++)
				{
					if(GetShort(VTSTitles,Offset+chap*4) == i)
					{	
					// Decrement value.
						AddShort(VTSTitles,Offset+chap*4,-1);
					}
				}
			}
		}
	}
	else
	{
		// Last PGC, we don't have as much to do...
		uchar* Ptr1 = (uchar*)GetPGCHeader(LU,PGC);
		uchar* Ptr2 = FindTable(Table) + GetEndAddress(Table) + 1;
		MoveDataInTable(Table, Ptr2, (int)Ptr1-(int)Ptr2);
	}


	// Decrement number of PGCs.
	AddShort(Ptr,0,-1);
	NumPGC--;
	// If none left, remove table entirely.
	if(NumPGC == 0)
		RemoveTable(Table);
}


