/*
 *  DVDSubEdit and this code is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  DVDSubEdit and this code are distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
*/

// TEST DEFINES:
//#define FORCE_EVEN_HEIGHT
//#define ADD_TWO_BLANK_LINES

#include "SubPic.h"
#include "Cell.h"
#include "math.h"
#include <io.h>

// This is insanely annoying! If you don't define that, parts of the structures are padded with
// zeros, which is catastrophic for predefined headers.
#pragma pack(2)
// I'm defining these here, rather than including a windows file. Don't like these!
// They're used for bimap files.
typedef struct tagBITMAPFILEHEADER {
	ushort    bfType;
	ulong	  bfSize;
	ushort    bfReserved1;
	ushort    bfReserved2;
	ulong	  bfOffBits;
} BITMAPFILEHEADER;

typedef struct tagBITMAPINFOHEADER{
	ulong      biSize;
	long       biWidth;
	long       biHeight;
	ushort     biPlanes;
	ushort     biBitCount;
	ulong      biCompression;
	ulong      biSizeImage;
	long       biXPelsPerMeter;
	long       biYPelsPerMeter;
	ulong      biClrUsed;
	ulong      biClrImportant;
} BITMAPINFOHEADER;
#define BI_RGB    0L	// RGB uncompressed bitmap format.

#ifndef _ASSERT
#include <crtdbg.h>
#endif

#ifdef NOMESSAGES
#undef MESSAGE1
#undef ERROR1
#undef ERROR2
#undef ERROR3
#undef ERROR4
#define ERROR1 
#define ERROR2
#define ERROR3
#define ERROR4
#endif


#define min(A,B)((A < B)?A:B)
#define max(A,B)((A > B)?A:B)

#define MAX_WIDTH 800
#define MAX_HEIGHT 600

// This is static, shared by all subpics.
static uchar PixelArray[MAX_WIDTH*MAX_HEIGHT];
uchar*	CSubPic::m_Pixels;

static int MyRead(void* Buffer, int Count1, int Count2, int Fd)
{
	return _read(Fd,Buffer,Count1*Count2);
}

static int MyWrite(void* Buffer, int Count1, int Count2, int Fd)
{
	return _write(Fd,Buffer,Count1*Count2);
}


/////////////////////////////////////////////////////////////////
int CSubPic::m_DisplayHeight = 576;
int CSubPic::m_DisplayWidth = 720;
uchar CSubPic::m_DefColors[12] = {0,255,0,0,0,0,0,0,0,0,0,0};
uchar CSubPic::m_BackColor[3] = {0,0,0};
int CSubPic::m_MoveCmdPostCrop = 0;
int CSubPic::m_ForceEvenMove = 0;
int CSubPic::m_CropBackBorderX = 1;
int CSubPic::m_CropBackBorderY = 1;
int CSubPic::m_PixTempAlpha = 0;
uchar CSubPic::m_ShowState = 1;
char CSubPic::m_SelButton = 0;
uchar CSubPic::m_ButColorMap[12];
uchar CSubPic::m_ButAlphaMap[4];



CSubPic::CSubPic()
{
	// Create OCR output storage.
	m_OCRLines = new uchar[32][128];
	m_OCRLinesOrig = new uchar[32][128];
	m_CharCoords = 0;
	Reset();
	m_StartOfLine = new ushort[MAX_HEIGHT];
	m_Data = 0;
	m_Pixels = PixelArray;
	m_NextSPUInStream = 0;
}

CSubPic::~CSubPic()
{
	delete[] m_Data;
	delete[] m_StartOfLine;
	delete[] m_OCRLines;
	delete[] m_OCRLinesOrig;
	delete[] m_CharCoords;
}

void CSubPic::Reset(bool ResetAll)
{
	// Reset a bunch of values.
	m_NavPackLBA = 0;
	m_CurrentBit = 0;		
	m_CurrentByte = 0;	
	m_PixelCount = 0;	
	m_CurrentPixelX = 0;	
	m_CurrentPixelY = 0;	
	m_SubPicWidth = 0;		
	m_SubPicHeight = 0;
	m_ModSubPicWidth = 0;
	m_SubPicSize = 0;
	m_RemainingSpace = 0;
	m_SpaceBeforeCmd = -1;
	m_DataBytesRead = 0;
	m_DataBytesWritten = 0;
	m_TopMostPixel = -1;
	m_BottomMostPixel = -1;
	m_TrueBottomLine = -1;
	m_BottomMostPixelT = -1;
	m_VerticalAdjust = 0;
	m_SubPicHasNoText = false;
	m_LastVMove = 0;
	m_LastHMove = 0;
	m_BackgroundPix = -1;
	m_BodyPix = -1;
	m_AntiAliasPixel1 = -1;
	m_AntiAliasPixel2 = -1;
	m_NumButtons = 0;
	m_VerticalBorders = 0;
	m_PTS = 0;
	m_TruePTS = 0;
	m_CLUT = 0;
	m_NumChars = 0;
	m_SUPByteOffest = -1;

	// By default a '\0' indicates no more data in the OCR.
	int i;
	for(i=0;i<32;i++)
	{
		m_OCRLines[i][0]='\0';
		m_OCRLinesOrig[i][0]='\0';
	}

	m_StartX = m_StartXOrig = 0;
	m_StartY = m_StartYOrig = 0;
	m_EndX = m_EndXOrig = 0;
	m_EndY = m_EndYOrig = 0;
	m_TopFieldOffset = 0;
	m_BotFieldOffset = 0;
	m_EndOfField = 0;

	m_AlphaPtr = 0;
	m_ColorPtr = 0;
	m_SelectedDCSQT = 0;
	m_ChangeAllDCSQT = 0;

	m_SPUReadOK = 0;
	m_SubPicDecoded = 0;
	m_RightMostPixel = 0;
	m_LeftMostPixel = 720;
	m_NumPackInSPU = 0;
	memset(m_FilesFrom,0,32*sizeof(int));
	memset(m_LBAFrom,0,32*sizeof(int));
	memset(m_CLBAFrom,0,32*sizeof(int));
	memset(m_NumPix,0,4*sizeof(ushort));
	memset(m_ButtonScheme,0,sizeof(m_ButtonScheme));
	memset(m_CSColorsSel,0,sizeof(m_CSColorsSel));
	memset(m_CSAlphasSel,0,sizeof(m_CSAlphasSel));
	memset(m_CSColorsAct,0,sizeof(m_CSColorsAct));
	memset(m_CSAlphasAct,0,sizeof(m_CSAlphasAct));

	memset(m_ButtonStartX,0,36*sizeof(ushort));
	memset(m_ButtonSizeX,0,36*sizeof(ushort));
	memset(m_ButtonStartY,0,36*sizeof(ushort));
	memset(m_ButtonSizeY,0,36*sizeof(ushort));

	memset(m_ButColorMap,0,sizeof(m_ButColorMap));
	memset(m_ButAlphaMap,0,sizeof(m_ButAlphaMap));

	m_Modified = 0;
	m_VOBID = 0;
	m_CELLID = 0;
	m_StreamID = 0;
	m_SPUNumber = 0;
	m_SPUNumberInStream = 0;
	m_SelButton = 0;
}

static bool WarnForDecodingError = true;

uchar CSubPic::pop2()
{
	_ASSERT(m_CurrentBit >= 1);
	uchar val = m_Data[m_CurrentByte];
	// Move val all the way to the right, then mask, keeping last 2 bits.
	val >>= (m_CurrentBit-1);
	val &= 0x03;
	// Decrementn m_CurrentBit, if needed go into next byte.
	m_CurrentBit -= 2;
	if(m_CurrentBit == -1)
	{
		m_CurrentBit = 7;
		m_CurrentByte ++;
	}
	if(m_CurrentByte > m_EndOfField && WarnForDecodingError)
	{
		_ASSERT(0);
		WarnForDecodingError = 0;
	}
	return val;
}

uchar CSubPic::pop4()
{
	// This will be a bit slow, but it's much much easier this way!
	uchar val = pop2();
	return val*4+pop2();
}

uchar CSubPic::pop8()
{
	// This will be a bit slow, but it's much much easier this way!
	uchar val = pop4();
	return val*16+pop4();
}

int CSubPic::Out(int n, int c)
{
	int i;
	static int LeftMostPixelForThisLine = -1;

	if(n == 0)
	{
		// Special case: output until end of line.
		n = m_SubPicWidth-m_CurrentPixelX;
	}

	// The background pixel type is the first one we see. This should work most of the time, but not always.
	if(!m_SubPicDecoded && m_PixelCount == 0 && m_BackgroundPix == -1)
		m_BackgroundPix = c;

	// If this isn't a background pix, remember if this is the first one we've seen.
	if(!m_SubPicDecoded && c != m_BackgroundPix && LeftMostPixelForThisLine == -1)
		LeftMostPixelForThisLine = m_CurrentPixelX;

	// Output pixel m_Data. Also count how many we see of each pixel.
	for(i=0;i<n;i++)
		m_Pixels[m_PixelCount++] = c;
	m_NumPix[c] += n;
	m_CurrentPixelX += n;

	// If this is not a background pixel, make a note if it's the rightmost one seen so far.
	if(!m_SubPicDecoded && c != m_BackgroundPix) 
	{
		m_RightMostPixel = max(m_RightMostPixel,m_CurrentPixelX-1);
		if(m_TopMostPixel == -1) m_TopMostPixel = m_CurrentPixelY;
		m_BottomMostPixelT = m_CurrentPixelY;
	}

	_ASSERT(m_CurrentPixelX <= m_SubPicWidth);
	if(m_CurrentPixelX == m_SubPicWidth)
	{
		// End of line! Keep track of the leftmost pixel.
		if(!m_SubPicDecoded && LeftMostPixelForThisLine != -1)
			m_LeftMostPixel = min(LeftMostPixelForThisLine,m_LeftMostPixel);
		LeftMostPixelForThisLine = -1;
		m_CurrentPixelX = 0;
		m_CurrentPixelY += 2;
		m_PixelCount += m_SubPicWidth;

		// If at the end of a line, we're not at a byte boundary, 4 fill bits of 0 are added.
		_ASSERT(m_CurrentBit == 7 || m_CurrentBit == 3);
		if(m_CurrentBit == 3)
		{
			m_CurrentBit = 7;
			m_CurrentByte ++;
		}
		if(!m_SubPicDecoded)
			m_StartOfLine[m_CurrentPixelY] = m_CurrentByte;
	}
	if(m_CurrentPixelY >= m_SubPicHeight)
		return -1;
	else 
		return 0;
}

// RLE decoding... See http://www.mpucoder.com/DVD/spu.html
void CSubPic::RLEDecode(int Field)
{
	int n;
	int c;
	int ret = 0;
	m_CurrentBit = 7;
	WarnForDecodingError=true;
	if(Field == 0)
	{
		m_PixelCount = 0;
		m_CurrentPixelX = m_CurrentPixelY = 0;
		m_CurrentByte = m_TopFieldOffset;
		m_EndOfField = m_TopFieldOffset < m_BotFieldOffset ? m_BotFieldOffset : m_OffsetToCmd;
	}
	else
	{
		m_PixelCount = m_SubPicWidth;
		m_CurrentPixelX = 0;
		m_CurrentPixelY = 1;
		m_CurrentByte = m_BotFieldOffset;
		m_EndOfField = m_TopFieldOffset < m_BotFieldOffset ? m_OffsetToCmd : m_TopFieldOffset;
	}

	while(ret == 0)
	{
		n = pop2();
		if(n != 0)
		{
			// 4 bits: nncc
			c = pop2();
			ret = Out(n,c);
		}
		else
		{
			// 00...
			n=pop2();
			if(n != 0)
			{
				// 8 bits: 00nnnncc
				n = 4*n+pop2();
				c = pop2();
				ret = Out(n,c);
			}
			else
			{
				// 0000...
				n=pop2();
				if(n != 0)
				{
					// 12 bits: 0000nnnnnncc
					n = n*16+pop4();
					_ASSERT(n>0);
					c = pop2();
					ret = Out(n,c);
				}
				else
				{
					// 16 bits: 000000nnnnnnnncc
					n = pop8();
					c = pop2();
					ret = Out(n,c);
				}
			}
		}
	}
}

// Push functions to encode pixel data.
void CSubPic::push4(ushort Code)
{
	if(m_CurrentBit == 0)
		m_Data[m_CurrentByte] = Code << 4;
	else
		m_Data[m_CurrentByte] |= Code;
	
	m_CurrentBit += 4;
	if(m_CurrentBit == 8)
		m_CurrentBit = 0, m_CurrentByte++;
}

void CSubPic::push8(ushort Code)
{
	// Not optimal, but easy!
	push4(Code >> 4);
	push4(Code & 0x000F);
}

void CSubPic::push12(ushort Code)
{
	// Not optimal, but easy!
	push4(Code >> 8);
	push8(Code & 0x00FF);
}

void CSubPic::push16(ushort Code)
{
	// Not optimal, but easy!
	push8(Code >> 8);
	push8(Code & 0x00FF);
}

void CSubPic::push(int count, uchar c) // Encode character c n times.
{
	// Special case, n==0;
	ushort Code = (ushort) c;
	Code |= count << 2;
	if(count==0 || count >= 64)
	{
		push16(Code);
		return;
	}
	if(count >= 16)
	{
		push12(Code);
		return;
	}
	if(count >= 4)
		push8(Code);
	else
		push4(Code);
}

void CSubPic::RLEEncode(int Field, int StartCol, int EndCol, int StartLine, int EndLine, uchar* Store)
{
	int RepeatCount;
	uchar PixelType;
	int Line = (Field == 0) ? StartLine : StartLine+1;
	EndLine = min(EndLine,m_SubPicHeight-1);
	EndCol = min(EndCol,m_SubPicWidth-1);
	int Col = 0;
	uchar* Ptr = m_Pixels + Line*m_SubPicWidth;
	// This is a bit ugly, but it allows me to save to a different buffer without having to pass the buffer
	// argument to all the push() functions.
	uchar* OrigDataPtr = m_Data;
	if(Store)
		m_Data = Store;

	m_CurrentBit = 0;
	if(Field == 0)
	{
		m_CurrentByte = 4;		// 4 bytes for the size of the data, and the offset to commands.
		m_TopFieldOffset = 4;
	}
	else m_CurrentByte = m_BotFieldOffset;

	int PixCount=0;
	for(;Line<=EndLine;Line+=2)
	{
		Ptr = m_Pixels + Line*m_SubPicWidth + StartCol;
		for(Col = StartCol;Col<=EndCol;)
		{
			PixelType = *Ptr;
			RepeatCount = 0;
			// Check whether this character is repeated until the end of line.
			int Col2;
			uchar* Ptr2 = Ptr;
			for(Col2=Col;Col2<=EndCol;Col2++,RepeatCount++,Ptr2++)
				if(*Ptr2 != PixelType) break;
			if(Col2 > EndCol)
			{
				// Same character until end of line. If it appears more than 64 times, it's more
				// advantageous to output a special 16bit code. Otherwise, go with the 12bit code.
				push(RepeatCount >= 64 ? 0 : RepeatCount, PixelType);
//				push(RepeatCount, PixelType);
//				push(0, PixelType);
				Ptr = Ptr2;
				// We're done with this line.
				break;
			}
			// This pixel type does not span the rest of the line. Update our pointers etc.
			RepeatCount = min(RepeatCount,255);
			Ptr += RepeatCount;
			Col += RepeatCount;
			_ASSERT(Col <= EndCol);
			// Here, either RepeatCount == 255 or we have a new char.
			// In any case, output data: We saw PixelType RepeatCount times.
			push(RepeatCount,PixelType);
			// And continue until we're done with this line.
		}
		// End of line. If we're not at a byte boundary, output 4 "0" bytes.
		if(m_CurrentBit != 0)
			push4(0);
	}
#ifdef ADD_TWO_BLANK_LINES
push(0,m_BackgroundPix);
#endif
	// Set the pointer for the other field, if Field == 0;
	if(Field == 0)
		m_BotFieldOffset = m_CurrentByte;
	else
		m_SpaceBeforeCmd = m_OffsetToCmd - m_CurrentByte;

	// Restore data buffer.
	m_Data = OrigDataPtr;
}

int CSubPic::DecodeOrig()
{
	if(!m_SubPicDecoded)
	{
		if(DecodeSubpic() == -1)
			return -1;
	}
	// Undo any move and decode the data as if it hadn't been modified at all.
	m_TopFieldOffset = m_StartOfLine[0];
	m_BotFieldOffset = m_StartOfLine[1];
	m_StartY = m_StartYOrig;
	m_EndY = m_EndYOrig;
	m_StartX = m_StartXOrig;
	m_EndX = m_EndXOrig;
	m_SubPicHeight = min(m_EndY,m_DisplayHeight-1) - m_StartY + 1;
	m_SubPicWidth = m_EndX-m_StartX+1;
	DecodeSubpic();
	return 0;
}

int CSubPic::CropBackground(bool DecodeFirst)
{
//	if(ContainsHighlights())
//		return -1;
	if(DecodeFirst && DecodeOrig() != 0)
			return -1;

	// Start colomns and lines.
	int StartCol = max(0,m_LeftMostPixel-m_CropBackBorderX);
	StartCol = max(StartCol,0);
	int EndCol = min(m_RightMostPixel+m_CropBackBorderX,m_SubPicWidth-1);
	
	int StartLine = max(0,m_TopMostPixel-m_CropBackBorderY);
	StartLine = 2*(StartLine/2);
	int EndLine = min(m_BottomMostPixel+m_CropBackBorderY,m_SubPicHeight-1);

#ifdef FORCE_EVEN_HEIGHT
	// For an even height, and because startline is even, endline should be odd.
	EndLine = max(0,2*((EndLine+1)/2)-1);
#endif

	// RLE-Encode the two fields into a temporary buffer. Also, remember the top and bottom field offsets,
	// in case we don't have enough space to encode.
	uchar* TempData = new uchar[1024*16];
	memset(TempData,0,1024*16);
	ushort TFOffset = m_TopFieldOffset;
	ushort BFOffset = m_BotFieldOffset;
	// NOTE: It's ESSENTIAL to encode field 0 first, otherwise m_SpaceBeforeCmd won't be right!
	RLEEncode(0,StartCol,EndCol,StartLine,EndLine,TempData);
	RLEEncode(1,StartCol,EndCol,StartLine,EndLine,TempData);

	if(m_CurrentByte > m_OffsetToCmd)
	{
		// Not enough space to encode into available buffer.
		delete[] TempData;
		// Restore these two guys. 
		m_TopFieldOffset = TFOffset;
		m_BotFieldOffset = BFOffset ;
		VMoveSubPic(m_LastVMove);
		HMoveSubPic(m_LastHMove);
		return -1;
	}
	else
	{
		// Enough space. Let's copy it there. I'm copying the data after the end of the picture, to
		// erase anything left until the first command.
		int ToCopy = min(m_OffsetToCmd - m_TopFieldOffset, 1024*16 - m_TopFieldOffset);
		memcpy(m_Data+m_TopFieldOffset,TempData+m_TopFieldOffset,ToCopy);
	}
	delete[] TempData;

	// Now this is needed to place the subpic exactly where it was before cropping.
	m_StartXOrig += StartCol;
	m_StartX = m_StartXOrig;
	m_SubPicWidth = EndCol-StartCol+1;
	m_ModSubPicWidth = ((3+m_SubPicWidth)/4)*4;
	m_StartYOrig += StartLine;
	m_StartY = m_StartYOrig;
	m_SubPicHeight = EndLine-StartLine+1;
	m_EndX = m_EndXOrig = m_StartXOrig + m_SubPicWidth -1;
	m_EndY = m_EndYOrig = m_StartYOrig + m_SubPicHeight -1;
	WriteDAREA(m_StartXOrig, m_StartYOrig, m_EndXOrig, m_EndYOrig);
	int Offset = FindCommand(SET_DSPXA);
	PutShort(m_Data,Offset+1,m_TopFieldOffset);
	PutShort(m_Data,Offset+3,m_BotFieldOffset);

	m_Modified = true;
	// This is very important: it will ensure that the m_StartOfLine array will be updated.
	m_SubPicDecoded = 0;
	m_BottomMostPixelT = m_BottomMostPixel = m_TopMostPixel = -1;
	m_RightMostPixel = 0; m_LeftMostPixel = 720;
	DecodeSubpic();
	// Reapply moves.
	VMoveSubPic(m_LastVMove);
	HMoveSubPic(m_LastHMove);

	return 0;
}

int CSubPic::ReadThisPack(uchar* Data, int fd, int LBA, int NavPackLBA, ushort VOBID, 
						  uchar CELLID, uchar StreamID, int CLBA, int VobFileNumber, int SPUNumber)
{
Start:
	// This could be a stuffing pack! Let's find out...
	if(GetInt(Data,14) == 0x000001BE)
	{
		// This is a stuffing pack! Let's grab it anyway!
		m_FilesFrom[m_NumPackInSPU] = fd;
		m_LBAFrom[m_NumPackInSPU] = LBA;
		m_CLBAFrom[m_NumPackInSPU] = CLBA;
		m_NumPackInSPU++;
		return 0;
	}
	_ASSERT(m_SPUReadOK == 0);
	m_FilesFrom[m_NumPackInSPU] = fd;
	m_LBAFrom[m_NumPackInSPU] = LBA;
	m_CLBAFrom[m_NumPackInSPU] = CLBA;
	uint len;
	int ToRead;
	if(m_DataBytesRead == 0)
	{
		// Find start of subpic data... It's at 0x18+len...I also make sure this has a PTS value!
		// That's the only reliable way (?) to make sure this is truly the start of a SPU.
		len = (uint) (Data[0x16]);
		_ASSERT(len > 0 && (Data[0x15] & 0x80));
		if(len == 0)
		{
			static int CountWarnings = 0;
			MESSAGE2("Len is zero or there's no PTS at LBA %d!",LBA);
//			ERROR2("Serious error while reading subpic pack at LBA %d\nSkipping this pack",LBA);
			if(CountWarnings <= 3)
			{
				sprintf(Message,"Serious error while reading SPU # %d starting at LBA %d (LBA %d of VTS_XX_%d.VOB)\nReplacing with a stuffing pack...\n\
VCID: %d/%d StreamID 0x%x\n",
				SPUNumber, m_CLBAFrom[0], m_LBAFrom[0], VobFileNumber+1, VOBID, CELLID, m_StreamID+0x20);
			}
			if(CountWarnings == 3)
			{
				sprintf(Message+strlen(Message),"\nI won't report this error again");
			}
			if(CountWarnings <= 3)
				ERROR1(Message);
			CountWarnings++;
			return -1;
		}
		m_PTS = m_TruePTS = ParsePTS(Data+0x17);
		// Also, save the first 64 bytes of the subpic pack.
		memcpy(m_PackStartData,Data,64);
		m_SubPicSize = GetShort(Data,0x18+len);

		// I'm giving myself some room to grow the subpic size all the way to the end of the last sector.
		m_Data = new uchar[m_SubPicSize+SECTOR_SIZE];

		ToRead = min(SECTOR_SIZE-0x18-len, m_SubPicSize);

		// This is to deal with a padding stream in the middle of our data!
		// I need to read the length at offset 0x12 as it tells me the length of our stream.
		ushort StreamLength = GetShort(Data,0x12);
		int TR = StreamLength - 0x18-len+0x14;
		if(TR != ToRead)
			MESSAGE2("NOTE: There's probably a padding stream in the middle of the subpic data at LBA %d",LBA);
//		_ASSERT(TR == ToRead);
		ToRead = min(ToRead,TR);
		memcpy(m_Data,Data+0x18+len,ToRead);

		m_DataBytesRead += ToRead;
		// Now that we have the beginning of the data buffer, we can get the length and the
		// offset to commands.
//		m_SubPicSize = GetShort(m_Data,0);
		m_OffsetToCmd = GetShort(m_Data,2);
		_ASSERT(m_OffsetToCmd < m_SubPicSize);
		_ASSERT(m_OffsetToCmd > 0);
		if(m_OffsetToCmd >= m_SubPicSize)
		{
			ERROR2("Inconsistent data in subpic at LBA %d. Offset to commands past end of SPU",m_LBAFrom[0]);
			m_OffsetToCmd = -1;
		}
		m_NavPackLBA = NavPackLBA;
		m_VOBID = VOBID;
		m_CELLID = CELLID;
		m_StreamID = StreamID;
	}
	else
	{
		_ASSERT(m_SubPicSize != 0 && m_DataBytesRead != 0);
		// Check that the PES header data length is 0 or that no PTS is present (that would indicate
		// we missed the first SPU pack!).
		len = (uint) (Data[0x16]);
//		_ASSERT(len == 0 || (Data[0x15] & 0x80)==0);
		if(len != 0 && (Data[0x15] & 0x80) != 0)
		{

			// This is the beginning of a SPU! We're missing a pack from the previous one!
			sprintf(Message,"SPU # %d starting at LBA %d (LBA %d of VTS_XX_%d.VOB) is incomplete! Dropping it...\n\
VCID: %d/%d StreamID 0x%x\n",
				SPUNumber, m_CLBAFrom[0], m_LBAFrom[0], VobFileNumber+1, m_VOBID, m_CELLID, m_StreamID+0x20);
			ERROR1(Message);
			// Best thing to do is forget the preceding one and start again with this one...
			Reset();
			goto Start;
		}
		// Read as much as you can, except if we have enough data.
		ToRead = min(SECTOR_SIZE-0x18-len, m_SubPicSize - m_DataBytesRead);
		memcpy(m_Data+m_DataBytesRead,Data+0x18+len,ToRead);
		m_DataBytesRead += ToRead;
	}
	m_NumPackInSPU++;
	if(m_SubPicSize == m_DataBytesRead)
	{
		// We're done! Now, let's read a few things.
		m_SPUReadOK = true;
		int Ret = ParseCommands();
		if(Ret == 0)
		{
			SelectDCSQT(m_SelectedDCSQT);
			// Also find out how much space we have behind the end of the SPU. Removing 6 for the padding space.
			long remain = SECTOR_SIZE-0x18-len - ToRead - 6;
			m_RemainingSpace = remain > 0 ? (ushort) remain : 0;
		}
		return Ret == 0 ? 1 : Ret;
	}
	else return 0;
}

int CSubPic::ReadFromSUPFile(int SupFileFD, int SPUNumber)
{
	// Now read PTS value;
	int ByteOffset = _tell(SupFileFD);
	int PTS;
	if(MyRead(&PTS,4,1,SupFileFD) != 4) return -1;
	m_PTS = m_TruePTS = PTS;
	m_StreamID = 0;
	if(MyRead(&PTS,4,1,SupFileFD) != 4) return -1;
	// Read data... First, Size and Offset to commands.
	ushort Size, Offset;
	if(MyRead(&Size,2,1,SupFileFD) != 2) return -1; 
	Size = FixEndian(Size);
	if(MyRead(&Offset,2,1,SupFileFD) != 2) return -1; 
	Offset = FixEndian(Offset);
	// Set a few member variables.
	m_SubPicSize = Size;
	// When reading from a sup, we don't really have a limit to how large the subpic can grow.
	m_RemainingSpace = 2048;
	m_SpaceBeforeCmd = 0;
	m_OffsetToCmd = Offset;
	m_SPUReadOK = true;
	m_SPUNumber = SPUNumber;
	m_SPUNumberInStream = SPUNumber;
	m_SUPByteOffest = ByteOffset;
	// Allocate data. I'm adding m_RemainingSpace bytes to give myself some space! seek back to start of Size.
	m_Data = new uchar[Size+m_RemainingSpace];
	_lseek(SupFileFD,-4,SEEK_CUR);
	// And read all the stuff in our bufffer.
	if(MyRead(m_Data,1,Size,SupFileFD) != Size) return -1;
	// Parse commands, etc.
	if(ParseCommands() == -1) return -1;
	SelectDCSQT(m_SelectedDCSQT);
	m_FilesFrom[0]=SupFileFD;
	return 0;
}


void CSubPic::SelectDCSQT(int DCSQT)
{
	if(DCSQT >= m_DCSQT.size())
		return;
	Commands *TheCommand;
	m_ColorPtr = NULL;
	m_AlphaPtr = NULL;
	// First, use DCST 0, this normally has color/contrast.
	if(FindCommand(SET_COLOR,0,&TheCommand) != -1)
		m_ColorPtr = TheCommand->m_Params;
	if(FindCommand(SET_CONTR,0,&TheCommand) != -1)
		m_AlphaPtr = TheCommand->m_Params;
	if(FindCommand(SET_COLOR,DCSQT,&TheCommand) != -1)
		m_ColorPtr = TheCommand->m_Params;
	if(FindCommand(SET_CONTR,DCSQT,&TheCommand) != -1)
		m_AlphaPtr = TheCommand->m_Params;
	_ASSERT(m_ColorPtr != NULL && m_AlphaPtr != NULL);
#if 0
	// Could use this to show scrolling effects...
	if(FindCommand(SET_DAREA,DCSQT,&TheCommand) != -1)
	{
	m_StartX = TheCommand->m_Params[3];
	m_EndX = TheCommand->m_Params[2]; 
	m_StartY = TheCommand->m_Params[1];
	m_EndY = TheCommand->m_Params[0]; 
	if(!m_SubPicDecoded)
	{ m_StartYOrig = m_StartY ; m_EndYOrig = m_EndY ; }
	_ASSERT(m_StartX < 1024 && m_EndX < 1024 && m_StartY < 1024 && m_EndY < 1024);
	m_SubPicWidth = m_EndX - m_StartX + 1;
	m_SubPicHeight = m_EndY - m_StartY + 1;
	m_ModSubPicWidth = ((3+m_SubPicWidth)/4)*4;
	}
#endif

	m_SelectedDCSQT = DCSQT;
}

bool CSubPic::FindString(const char* Search, bool CaseSensitive)
{
	int j;
	int k=0;
	char Str1[1024]; 
	strcpy(Str1,Search); 
	// Make search string lowercase.
	if(!CaseSensitive)
		_strlwr(Str1);
	char Str2[1024];
	for(j=0;;j++)
	{
		if(m_OCRLines[j][0] == '\0')
			return false;
		// Copy OCR output and make it lowercase.
		strcpy(Str2,(char*)m_OCRLines[j]); 
		if(!CaseSensitive)
			_strlwr(Str2);
		// Then search.
		if(strstr(Str2,Str1) != 0)
			return true;
	}
}

bool CSubPic::FindChars(const char* Search, bool CaseSensitive)
{
	int j;
	int k=0;
	char Str1[1024]; 
	strcpy(Str1,Search); 
	// Again, make it lowercase.
	if(!CaseSensitive)
		_strlwr(Str1);
	char Str2[1024];
	for(j=0;;j++)
	{
		if(m_OCRLines[j][0] == '\0')
			return false;
		// Copy OCR output and make it lowercase.
		strcpy(Str2,(char*)m_OCRLines[j]); 
		if(!CaseSensitive)
			_strlwr(Str2);
		// Look for characters.
		if(strpbrk(Str2,Str1) != 0)
			return true;
	}
}

static int Brack = 0;

bool TestBracketL(uchar Char, const char* Brackets) 
{
	if(Char == Brackets[0]) {Brack = 0; return true;}
	if(Char == Brackets[1]) {Brack = 1; return true;}
	if(Char == Brackets[2]) {Brack = 2; return true;}
	return false;
}

bool TestBracketR(uchar Char, const char* Brackets) 
{
//	return Char == Brackets[3] || Char == Brackets[4] || Char == Brackets[5];
	return Char == Brackets[5-Brack];
}

bool TestBracketR2(uchar Char, const char* Brackets) 
{
	return Char == Brackets[3] || Char == Brackets[4] || Char == Brackets[5];
}

#if 1
bool CSubPic::IsTextBracketed(const char* Brackets)
{
	int i,j;
	bool Found = false;
	for(j=0;;j++)
	{
		if(m_OCRLines[j][0] == '\0') break;
		// Find first non-space character.
		for(i=0;;i++)
			if(m_OCRLines[j][i] != ' ' && m_OCRLines[j][i] != '-' || m_OCRLines[j][i] == '\0') break;
		if(!TestBracketL(m_OCRLines[j][i],Brackets))
			return false;
		// Now find the matching closing bracket.
		Found = false;
		i++;
		while(1)
		{
			if(TestBracketR(m_OCRLines[j][i],Brackets)) break;
			if(m_OCRLines[j][i] == '\0' && i==0)
				return false;
			// You might have to go to the next line!
			if(m_OCRLines[j][i] == '\0')
				i=-1,j++;
			i++;
		}
		// The matching bracket was found.
		Found = true;
		// Now check that there's no non-space character after that.
		// JEAN: Yes, but what if you have -[asdf] \n -[asdf]
		for(i++;;i++)
		{
			if(m_OCRLines[j][i] == '\0') break;
			if(TestBracketR2(m_OCRLines[j][i],Brackets)) continue;
			if(m_OCRLines[j][i] != ' ') return false;
		}
		// Continue on to the next line if any...
	}
	return Found;
}

#else
bool CSubPic::IsTextBracketed(const char* Brackets)
{
	int j;
	// Check whether the first non-space character is a bracket.
	for(j=0;j<m_NumChars;j++)
	{
		if(m_CharCoords[j].c == '\0') return false;
		// Find first non-space character.
		if(m_CharCoords[j].c == ' ' || m_CharCoords[j].c == '-') continue;
		if(!TestBracketL(m_CharCoords[j].c,Brackets))
			return false;
		else break;
	}
	// Check whether the first non-space character is a bracket.
	for(j=m_NumChars-1;j>=0;j++)
	{
		// Find first non-space character.
		if(m_CharCoords[j].c == ' ' || m_CharCoords[j].c == '-' || m_CharCoords[j].c == '\0') continue;
		if(!TestBracketR2(m_CharCoords[j].c,Brackets))
			return false;
		else break;
	}
	return true;
}

#endif
void CSubPic::EraseAll()
{
	int i;
	DecodeOrig();

	for(i=0;i<m_PixelCount;i++)
			m_Pixels[i] = m_BackgroundPix;
	// Now call this to make sure we reencode.
	CropBackground(false);
}


void CSubPic::EraseSubPic(int FirstChar, int LastChar, bool Recenter, bool RealignBottom)
{
	int TrueBottomLineBefore = m_StartY + m_TrueBottomLine;
	// Call the helper function, making sure we decode first.
	int i,j;
	// We have to erase line by line because of the way DoEraseText() works. Only decode
	// the first time around.
	for(i=j=FirstChar;i<=LastChar;i++)
	{
		if(m_CharCoords[j].line != m_CharCoords[i].line)
		{
			DoEraseText(j,i-1,j==FirstChar, Recenter);
			j=i;
		}
	}
	if(i!=j) 
		DoEraseText(j,i-1,j==FirstChar,Recenter);

//	DoEraseText(FirstChar,LastChar,true, Recenter);
	// Now re-encode
	CropBackground(false);
	m_NumChars = - m_NumChars;
	// And readjust the vertical position...
	if(RealignBottom && !m_SubPicHasNoText && TrueBottomLineBefore > m_StartY + m_TrueBottomLine)
		VMoveSubPic(m_LastVMove - (TrueBottomLineBefore - m_StartY - m_TrueBottomLine));
}

int CSubPic::DoEraseText(int FirstChar, int LastChar, bool DecodeFirst, bool Recenter)
{
	int i;
	if(DecodeFirst)
		DecodeOrig();
	if(LastChar >= m_NumChars)
		return -1;

	// The '\n' char has very strange values for x0 and x1...
	if(m_CharCoords[LastChar].c == '\n') LastChar--;
	int StartLine = m_SubPicHeight;
	int EndLine = 0;
	// Find the topmost and bottomost pixels of the area that will be moved.
	for(i=FirstChar;i<m_NumChars && m_CharCoords[i].c != 0 && m_CharCoords[i].c != '\n';i++)
	{
		StartLine = min(StartLine,m_CharCoords[i].y0);
		EndLine = max(EndLine,m_CharCoords[i].y1);
	}
	for(i=FirstChar;i>=0 && m_CharCoords[i].c != 0 && m_CharCoords[i].c != '\n';i--)
	{
		StartLine = min(StartLine,m_CharCoords[i].y0);
		EndLine = max(EndLine,m_CharCoords[i].y1);
	}
	// This is needed because there are anti-aliasing pixels out there!
	StartLine = max(0,StartLine-3);
	EndLine = min(m_SubPicHeight,EndLine+3);

	// Find the start and end column of the area to move. Cutting in between consecutive characters.
	int StartCol = max(m_LeftMostPixel,m_CharCoords[FirstChar].x0-5);
	if(FirstChar > 0 && m_CharCoords[FirstChar-1].line == m_CharCoords[FirstChar].line)
		StartCol = (m_CharCoords[FirstChar].x0 + m_CharCoords[FirstChar-1].x1 + 1)/2;
	// I need to allow going beyond m_SubPicWidth for italics. 
	int EndCol = m_CharCoords[LastChar].x1+5;
	if(LastChar < m_NumChars-1 && m_CharCoords[LastChar+1].c != '\n')
		EndCol = (m_CharCoords[LastChar+1].x0 + m_CharCoords[LastChar].x1 + 1)/2-1;
//KLUDGE, when erasing next to a space character, this will make sure there's no leftover.
	if(m_CharCoords[FirstChar-1].c == ' ') StartCol-=2;
	if(m_CharCoords[LastChar+1].c == ' ') EndCol+=2;
	float Slope = m_CharCoords[FirstChar].Slope;
	int LineTops = m_CharCoords[FirstChar].LineTops;
	int Offset = 0;
	// Then move: move the pixels and replace by background pixels.
	for(i=StartLine;i<=EndLine;i++)
	{
		uchar* Ptr = m_Pixels+i*m_SubPicWidth;
		// This takes care of italics. Slope is the tilt (0.01 for 1%), and LineTops is the starting
		// line that was used when straightening the bitmap. We need to replicate that here.
		Offset = (int)((i-LineTops)*Slope);
		int le = min(StartCol-Offset,m_SubPicWidth-1);	// Left limit of what to erase.
		int ri = min(EndCol+1-Offset,m_SubPicWidth);	// Right limit
		// We're moving a wedge here /_|, not a parallelipiped /_/ so N increases with i.
		int N = m_SubPicWidth-ri;						// How many to move
		memmove(Ptr+le,Ptr+ri,max(N,0));
		// Now erase. Because the right edge of what we moved is straight, we erase a rectangle.
		memset(Ptr+m_SubPicWidth-(ri-le),m_BackgroundPix,max(ri-le,0));
		if(Recenter)
		{
			// Move entire line right by half of what was removed...
			int HowMuch = (EndCol-StartCol) / 2;
			_ASSERT(HowMuch >= 0);
			if(HowMuch >= 0)
			{
				memmove(Ptr+HowMuch,Ptr,m_SubPicWidth-HowMuch);
				memset(Ptr,m_BackgroundPix,HowMuch);
			}
		}
	}
	return 0;
}

void CSubPic::EraseBrackets(const char* Brackets, bool HorizRecenter, bool RealignBottom)
{
	if(m_NumChars <= 0)
		return;
	// Look for brackets;
	int i,k,FirstChar;
	int CountChars = 0;
	int CurChar = 0;
	bool NeedsDecode = true;
	int TrueBottomLineBefore = m_StartY + m_TrueBottomLine;
	
	while(1)
	{
start: 
		// Look for opening bracket.
		for(k=CurChar;k<m_NumChars;k++)
			if(TestBracketL(m_CharCoords[k].c,Brackets)) break;
		if(k >= m_NumChars) goto end;
		CurChar = k+1;
		FirstChar = k;

		// Look for closing bracket.
		for(k=CurChar;k<m_NumChars;k++)
		{
			if(m_CharCoords[k].c == '\n') 
			{
				// End of line while looking for ]. Continue from start...
				CurChar = k;
				goto start;
			}
			if(TestBracketR(m_CharCoords[k].c,Brackets)) break;
		}
		if(k >= m_NumChars) goto end;
		int LastChar = k;
		CurChar = k;
		// Now look for something on the right of the bracket
		for(i=k+1;i<m_NumChars;i++)
			if(m_CharCoords[i].c != ' ' && m_CharCoords[i].c != '#' && m_CharCoords[i].c != ':') break;
		if(i >= m_NumChars || m_CharCoords[i].c == '\0' || m_CharCoords[i].c == '\n')
		{
			// Nothing right of the ']' bracket, look if we need to erase left of '['
			// Go left and see if you see something else that " " or "-" or '#'
			for(i=FirstChar-1;i>=0;i--)
			{
				if(m_CharCoords[i].c == ' ' || m_CharCoords[i].c == '-' || m_CharCoords[i].c == '#') 
					FirstChar = i;
				else
					break;
			}
		}
//		else LastChar = i;
		DoEraseText(FirstChar,LastChar,NeedsDecode,HorizRecenter);
		// Next time around we don't need to decode.
		NeedsDecode = false;
		CurChar++;
	}
end:
	// DoEraseText was called at least once. We need to call CropBackground to do the RLE encoding.
	if(!NeedsDecode)
	{
		CropBackground(false);
		// This will prevent running this function twice, which could be desastrous!
		m_NumChars = - m_NumChars;
	}

	// Optionally, you can readjust the vertical position...
	if(RealignBottom && !m_SubPicHasNoText && TrueBottomLineBefore > m_StartY + m_TrueBottomLine)
		VMoveSubPic(m_LastVMove - (TrueBottomLineBefore - m_StartY - m_TrueBottomLine));
}

void CSubPic::EraseColumns(bool HorizRecenter, bool RealignBottom )
{
	if(m_NumChars <= 0)
		return;
	// Look for ':';
	int k,FirstChar,LastChar;
	int CountChars = 0;
	int CurChar = 0, SpaceCount = 0;
	bool NeedsDecode = true;
	int TrueBottomLineBefore = m_StartY + m_TrueBottomLine;
	
start: 
		// Look for 1 word followed by ':'.
	FirstChar = CurChar;
	bool AllCaps = true;
	for(k=CurChar;k<m_NumChars-1;k++)
	{
		// Check all caps. Avoid 'l' (small L) as it can be confused with I capital i.
		if(m_CharCoords[k].c >= 'a' && m_CharCoords[k].c <= 'z' && m_CharCoords[k].c != 'l') 
			AllCaps = false;
		if(m_CharCoords[k].c == ' ') SpaceCount++;
		if(m_CharCoords[k].c == '\n') goto nextline;
		if(m_CharCoords[k].c == ' ' && 
			m_CharCoords[k+1].c != ':' && m_CharCoords[k-1].c != '-') SpaceCount+=2;
		// A zero can be mis-read as a capital O.
		if(m_CharCoords[k].c == ':' && (SpaceCount <= 1 || AllCaps) && 
			(m_CharCoords[k+1].c < '0' || m_CharCoords[k+1].c > '9') && m_CharCoords[k+1].c != 'O')
		{
			// Found 1 word followed by ':'. Erase!
			if(m_CharCoords[k+1].c == ' ') k++;
			LastChar = k;
			if(m_CharCoords[FirstChar].c == '-') FirstChar++;
			if(m_CharCoords[FirstChar].c == ' ') FirstChar++;
			DoEraseText(FirstChar,LastChar,NeedsDecode,HorizRecenter);
			// Next time around we don't need to decode.
			NeedsDecode = false;
			CurChar = k++;
			SpaceCount = 2;
		}
		if(SpaceCount > 1 && !AllCaps)
		{
nextline:
			// Two words... Go to next line.
			for(k=CurChar+1;k<m_NumChars;k++)
			{
				AllCaps = true;
				if(m_CharCoords[k].c == '\n')
				{
					CurChar = k+1;
					SpaceCount = 0;
					goto start;
				}
			}
			goto end;
		}
	}
end:
	// DoEraseText was called at least once. We need to call CropBackground to do the RLE encoding.
	if(!NeedsDecode)
	{
		CropBackground(false);
		// This will prevent running this function twice, which could be desastrous!
		m_NumChars = - m_NumChars;
	}

	// Optionally, you can readjust the vertical position...
	if(RealignBottom && !m_SubPicHasNoText && TrueBottomLineBefore > m_StartY + m_TrueBottomLine)
		VMoveSubPic(m_LastVMove - (TrueBottomLineBefore - m_StartY - m_TrueBottomLine));
}

bool CSubPic::ContainsHighlights()
{
	// This is a simple, but not foolproof test. If the subpic's navpack defines buttons, and if
	// our start command is a forced start, chances are these are highlights. I could also test
	// transparency, but some highlights are not transparents.
	return (m_NumButtons && FindCommand(FSTA_DSP,0) != -1);
}

int CSubPic::SaveSPU()
{
	if(!m_SPUReadOK)
		return -1;
	int i;
	uchar Buffer[SECTOR_SIZE];
	m_DataBytesWritten = 0;
	SwapFieldsIfNeeded();
	// WHAT I SHOULD ALWAYS DO: 1) Move the commands all the way to the end (so there's no change of losing a pack)
	// and 2) Move the data down so there's no gap between data and commands.
	if(m_RemainingSpace > 1) MoveAllCommands(m_RemainingSpace);
	// If the option is on, move the encoded bitmap so there's no gap between the data and the commands.
	if(m_MoveCmdPostCrop && m_SpaceBeforeCmd > 0) MoveData(m_SpaceBeforeCmd);
	for(i=0;i<m_NumPackInSPU;i++)
	{
		// Retrieve the fd and the LBA for the subpic pack we're about to save.
		int fd = m_FilesFrom[i];
		int LBA = m_LBAFrom[i];
		// Seek there and read the sector...
		if(_lseeki64(fd,(__int64)LBA*SECTOR_SIZE,SEEK_SET) == -1L || MyRead(Buffer,1,SECTOR_SIZE,fd) != SECTOR_SIZE)
		{
			ERROR3("Failed to seek/read %d-th subpic pack at LBA %d\n",i,LBA);
			return -1;
		}
		// Find start of subpic data then save back our subpic data...
		uint len = (uint) (Buffer[0x16]);
		int ToWrite = min(SECTOR_SIZE-0x18-len, m_SubPicSize - m_DataBytesWritten);
		if(ToWrite == 0)
		{
			// We have nothing left to write! but we have one or more packs to write!
			// Write a stuffing pack.
			uchar StuffingPack[SECTOR_SIZE];
			uchar StuffingPackHead[20] = {0x00,0x00,0x01,0xBA,0x44,0x00,0x04,0x00,0x04,0x01,0x01,0x89,0xC3,0xF8,0x00,0x00,0x01,0xBE,0x07,0xEC};
			memset(StuffingPack,0xFF,SECTOR_SIZE);
			memcpy(StuffingPack,StuffingPackHead,sizeof(StuffingPackHead));
			// Skipping to 0x0e to preserve the SCR
			if(_lseeki64(fd,(__int64)LBA*SECTOR_SIZE+0x0e,SEEK_SET) == -1L  ||
				MyWrite(StuffingPack+0x0e,sizeof(char),SECTOR_SIZE-0x0e,fd) != SECTOR_SIZE-0x0e)
			{
				ERROR3("Failed to write %d-th subpic pack at LBA %d\n",i,LBA);
				return -1;
			}
			continue;		
		}
		memcpy(Buffer+0x18+len,m_Data+m_DataBytesWritten,ToWrite);
		// Write back the beginning of the first pack, in case we modified it.
		_ASSERT(0x18+len <= 64);
		if(i==0)
			memcpy(Buffer,m_PackStartData,0x18+len);
		// We'll need the padding stuff if what we have to write does not fill the sector.
		if(0x18+len+ToWrite+6 <= SECTOR_SIZE)
		{
			// Write the padding stuff!!!
			PutInt(Buffer,0x18+len+ToWrite,0x000001be);
			PutShort(Buffer,0x18+len+ToWrite+4,SECTOR_SIZE-(0x18+len+ToWrite)-6);
			memset(Buffer+0x18+len+ToWrite+6,0xFF,SECTOR_SIZE-(0x18+len+ToWrite+6));
			// Also, modify the length at offset 0x12! This counts the byte from offset 0x14 to the end of the
			// commands (which should be a FF) right before the padding code.
			PutShort(Buffer,0x12,0x18+len+ToWrite-0x14);
		}
		else if(0x18+len+ToWrite < SECTOR_SIZE)
		{
			// Not enough room for the padding stuff. Simply put 0s after the commands. Well as it turns out, r0lZ'
			// Kiss and sony players don't like that. The kiss wants the subpic to finish right at the end.
#if 0
			memset(Buffer+0x18+len+ToWrite,0,SECTOR_SIZE-(0x18+len+ToWrite));
#else
			// Move everything down by a few bytes so the end of the commands is just at the end of the pack, and
			// stuff the PES headerdata http://forum.doom9.org/showthread.php?p=616490&highlight=padding#post616490
			int MoveBy = SECTOR_SIZE-(0x18+len+ToWrite);
			_ASSERT(MoveBy + len < 16);	// The length must fit in 1 byte!
			if(MoveBy + len < 16 && i < m_NumPackInSPU-1)
			{
				// len is relative to 0x17.
				memmove(Buffer+0x17+len+MoveBy,Buffer+0x17+len,SECTOR_SIZE-(0x17+len+MoveBy));
				Buffer[0x16]+=MoveBy;
				memset(Buffer+0x17+len,0xFF,MoveBy);
			}
			else 
			{
				MESSAGE2("Warning: Not enough room at LBA %d for padding stream",LBA);
				memset(Buffer+0x18+len+ToWrite,0,SECTOR_SIZE-(0x18+len+ToWrite));
			}
			// Also, modify the length at offset 0x12! See above.
			PutShort(Buffer,0x12,0x18+len+ToWrite-0x14+MoveBy);
#endif
		}
		// This is necessary for those rare cases where we had a padding stream in the middle of our
		// data.
		else PutShort(Buffer,0x12,0x18+len+ToWrite-0x14);

#if 0
FILE* Fucker = fopen("\\temp\\subpic.VOB","wb");
fwrite(Buffer,sizeof(char),SECTOR_SIZE,Fucker);
fclose(Fucker);
#endif
		if(_lseeki64(fd,(__int64)LBA*SECTOR_SIZE,SEEK_SET) == -1L ||
			MyWrite(Buffer,sizeof(char),SECTOR_SIZE,fd) != SECTOR_SIZE)
		{
			ERROR3("Failed to write %d-th subpic pack at LBA %d\n",i,LBA);
			return -1;
		}
		m_DataBytesWritten += ToWrite;
	}

	m_Modified = 0;
	// Now, copy OCR results to "original".
	for(i=0;i<32;i++)
		strcpy((char*)m_OCRLinesOrig[i],(char*)m_OCRLines[i]);
	return 0;
}

int CSubPic::ReloadSPU()
{
	if(!m_SPUReadOK)
		return -1;
	int i;
	m_DataBytesRead = 0;
	m_SubPicDecoded = 0;
	m_RightMostPixel = 0;
	m_LeftMostPixel = 720;
	m_TopMostPixel = m_BottomMostPixel = -1;

	if(m_SUPByteOffest != -1)
	{
		// If this SPU was loaded from a sup file, seek back to the byte offset, and reload...
		_lseeki64(m_FilesFrom[0],(__int64)m_SUPByteOffest,SEEK_SET);
		ReadFromSUPFile(m_FilesFrom[0],m_SPUNumber);
		m_Modified = 0;
		m_LastVMove = 0;
		m_LastHMove = 0;
		m_VerticalBorders = 0;
		// In case this guy had been modified, m_NumChars would be negative.
		m_NumChars = abs(m_NumChars);
		return 0;
	}

	uchar Buffer[SECTOR_SIZE];
	for(i=0;i<m_NumPackInSPU;i++)
	{
		// Retrieve the fd and the LBA for the subpic pack we're about to reload.
		int fd = m_FilesFrom[i];
		int LBA = m_LBAFrom[i];
		// Seek there and read the sector...
		if(_lseeki64(fd,(__int64)LBA*SECTOR_SIZE,SEEK_SET) == -1L || MyRead(Buffer,1,SECTOR_SIZE,fd) != SECTOR_SIZE)
		{
			ERROR3("Failed to seek/read %d-th subpic pack at LBA %d\n",i,LBA);
			return -1;
		}
		// Find start of subpic data then copy the subpic data...
		uint len = (uint) (Buffer[0x16]);
		if(i==0) m_SubPicSize = GetShort(Buffer,0x18+len);
		if(i==0) m_OffsetToCmd = GetShort(Buffer,0x18+len+2);
		int ToRead = min(SECTOR_SIZE-0x18-len, m_SubPicSize - m_DataBytesRead);
		// This is to deal with a padding stream in the middle of our data!
		// I need to read the length at offset 0x12 as it tells me the length of our stream.
		ushort StreamLength = GetShort(Buffer,0x12);
		ToRead = min(ToRead,StreamLength - 0x18-len+0x14);
		memcpy(m_Data+m_DataBytesRead,Buffer+0x18+len,ToRead);
		m_DataBytesRead += ToRead;
		// The remaining space must allow for a stuffing stream (6 bytes), I'm putting 7 for good measure.
		if(i==m_NumPackInSPU-1)
			m_RemainingSpace = SECTOR_SIZE-0x18-len - ToRead - 7;
	}
	if(m_DataBytesRead != m_SubPicSize)
	{
		ERROR1("Failed to re-load enough data");
		return -1;
	}
	ParseCommands();
	// This is necessary, in case we now have fewer DCSQT we had before we reloaded
	SelectDCSQT(0);
	SelectDCSQT(m_SelectedDCSQT);
	m_Modified = 0;
	m_LastVMove = 0;
	m_LastHMove = 0;
	m_VerticalBorders = 0;
	return 0;
}

int CSubPic::FindTrueBottomLine()
{
	if(m_SPUReadOK == false)
		return -1;
	
	int NumActivePixels[1024];
	uchar* Ptr;
	int i,j;
	// Find the maximum number of set pixels in the last subpic line.
	int MaxNum = 0;
	for(i=m_BottomMostPixel;i>=0;i--)
	{
		NumActivePixels[i] = 0;
		Ptr = m_Pixels + i*m_SubPicWidth;
		// I'm using the bodypixel here to be sure that we find a space between lines.
		for(j=0;j<m_SubPicWidth;j++,Ptr++)
			if(*Ptr == m_BodyPix) NumActivePixels[i]++;
		// Stop at the first empty line, after we've found one.
		if(NumActivePixels[i] == 0 && MaxNum > 0)
			break;
		MaxNum = max(MaxNum,NumActivePixels[i]);
	}
	int TopLastLine = i;
	int StopLine = TopLastLine+(m_BottomMostPixel-TopLastLine)/2;

	// The true bottom line is the first line that has at least half as many set pixels as the largest
	// line
	for(i=m_BottomMostPixel;i>=StopLine;i--)
	{
		NumActivePixels[i] = 0;
		Ptr = m_Pixels + i*m_SubPicWidth;
		// But here I'm using the background pixel.
		for(j=0;j<m_SubPicWidth;j++,Ptr++)
			if(*Ptr != m_BackgroundPix) NumActivePixels[i]++;

		if(2*NumActivePixels[i] > MaxNum)
			break;
	}
	m_TrueBottomLine = (i>=StopLine) ? i : m_BottomMostPixel;
	return 0;
}

int CSubPic::DecodeSubpic()
{
	if(m_SPUReadOK == false)
		return -1;
	m_SubPicHasNoText = false;
	if(!m_SubPicDecoded)
	{
		// Only set these the first time around. This way we'll keep a memory of them.
		m_StartOfLine[0] = m_TopFieldOffset;
		m_StartOfLine[1] = m_BotFieldOffset;
		m_SpaceBeforeCmd = 0;
	}
	memset(m_NumPix,0,4*sizeof(ushort));
	RLEDecode(0);
	int MaxCurrentByte = m_CurrentByte;
	RLEDecode(1);
	MaxCurrentByte = max(MaxCurrentByte,m_CurrentByte);
//if(m_CurrentByte != m_OffsetToCmd)
//ERROR3("Gap between end of subpic data at byte %d and start of commands at byte %d",m_CurrentByte, m_OffsetToCmd);

	if(m_BottomMostPixel == -1)
	{
		m_BottomMostPixel = m_BottomMostPixelT;
//		// Now find the true bottom line of the subpic.
//		FindTrueBottomLine();
	}
	// It's possible that the subpic is entirely empty!
	if(m_TopMostPixel == -1) 
	{
		m_SubPicHasNoText = true;
		m_TopMostPixel = 0;
		m_BottomMostPixel = min(m_SubPicHeight-1,10);
		m_TrueBottomLine = m_BottomMostPixel;
		m_LeftMostPixel = 0;
		m_RightMostPixel = min(m_SubPicWidth-1,10);
	}

	if(!m_SubPicDecoded)
	{
		FindBodyPixel();
		// Now find the true bottom line of the subpic.
		FindTrueBottomLine();
		m_SpaceBeforeCmd = m_OffsetToCmd - MaxCurrentByte;
	}


	m_SubPicDecoded = true;
	return 0;
}

void CSubPic::FindBodyPixel()
{
	int i,j;
	if(m_NumButtons > 0)
	{
#if 0
		// Special code for button highlights. We don't really want to apply the same technique.
		// A simple test based on the number of pixels used.
		int MaxCount;
		for(i=MaxCount=0;i<4;i++)
		{
			if(i==m_BackgroundPix) continue;
			if(m_NumPix[i] > MaxCount) MaxCount = m_NumPix[i], m_BodyPix = i;
		}
		for(i=MaxCount=0;i<4;i++)
		{
			if(i==m_BackgroundPix || i==m_BodyPix) continue;
			if(m_NumPix[i] > MaxCount) MaxCount = m_NumPix[i], m_AntiAliasPixel1 = i;
		}
		for(i=MaxCount=0;i<4;i++)
		{
			if(i!=m_BackgroundPix && i!=m_BodyPix && i!=m_AntiAliasPixel1 )
				m_AntiAliasPixel2 = i;
		}
#else
		// Here's another version: leave background pixel alone, set the other ones arbitrarily.
		i=0;
		if(i == m_BackgroundPix) i++;
		m_BodyPix = i++;
		if(i == m_BackgroundPix) i++;
		m_AntiAliasPixel1 = i++;
		if(i == m_BackgroundPix) i++;
		m_AntiAliasPixel2 = i++;
#endif
		return;
	}

	// This is more complicated but more reliable. Exclude the background pixel, and the first 
	// anti-alias pixel found. Then pick between the two remaining ones (if there are two!)
	int line = m_TopMostPixel+1;
	uchar* Ptr = m_Pixels + line*m_SubPicWidth;
	// Find first non-background pixel.
	for(i=0;i<m_SubPicWidth;i++)
		if(Ptr[i] != m_BackgroundPix) break;
	if(i == m_SubPicWidth)
		return;
	// This has GOT to be an antialiasing pixel.
	m_AntiAliasPixel1 = Ptr[i];
	// If only three pixel types are used, we're pretty much done!
	uchar Others[2] = {-1,-1};
	for(i=j=0;i<4;i++)
	{
		if(i != m_BackgroundPix && i != m_AntiAliasPixel1 && m_NumPix[i] != 0) 
			Others[j++] = i;
	}
	if(j == 0)
	{
		// No other pixel type! Assign other pixel types so we have all 4.
		m_BodyPix = m_AntiAliasPixel2 = m_AntiAliasPixel1;
		for(i=0;i<4;i++)
			if(m_BackgroundPix != i && m_BodyPix != i) break;
		m_AntiAliasPixel1 = i;
		for(i=0;i<4;i++)
			if(m_BackgroundPix != i && m_BodyPix != i && m_AntiAliasPixel1 != i) break;
		m_AntiAliasPixel2 = i;
		return;
	}
	if(j == 1)
	{
		// Only 1 other pixel type, we're done.
		m_BodyPix = Others[0];
		for(i=0;i<4;i++)
			if(m_BackgroundPix != i && m_BodyPix != i && m_AntiAliasPixel1 != i) break;
		m_AntiAliasPixel2 = i;
		return;
	}
	// More difficult case. There are 2 other pixel types, and we need to decide which one is used as
	// the character body. Go through bitmap, and count how many times you see o1-o2-o1. This would
	// point to o2 being the body.
	uchar First = -1;
	uchar Second = -1;
	int N = (m_BottomMostPixel + 1)*m_SubPicWidth;
	m_BodyPix = -1;
	int Tally[4]={1,1,1,1};Tally[m_BackgroundPix]=Tally[m_AntiAliasPixel1]=0;
	for(i=line*m_SubPicWidth;i<N;i++)
	{
		// Find first non-background, non-AntiAlias1 pixel.
		for(;i<N;i++)
			if(m_Pixels[i] != m_BackgroundPix && m_Pixels[i] != m_AntiAliasPixel1) break;
		if(i == N) break;
		First = m_Pixels[i];
		// Now look for next pixel type;
		for(;i<N;i++)
			if(m_Pixels[i] != First) break;
		if(i == N) break;
		// Make sure we're seeing o2!
		if(m_Pixels[i] == m_BackgroundPix || m_Pixels[i] == m_AntiAliasPixel1)
			continue;
		Second = m_Pixels[i];
		// Now look for o1 again.
		for(;i<N;i++)
			if(m_Pixels[i] != Second) break;
		if(i == N) break;
		// Make sure we're seeing o1!
		if(m_Pixels[i] != First)
			continue;
		// Now check that the next pixels are NOT Second
		for(;i<N;i++)
			if(m_Pixels[i] != First) break;
		if(i == N) break;
		if(m_Pixels[i] == Second)
			continue;
		// We found it! Add that to our Tally!
		Tally[Second]++;
	}
	// Now the winner is the one whose Tally has the largest value!
	int MaxTally = 0;
	for(i=0;i<4;i++)
	{
		// Kludge, I'm including also how many pixels are found of a given type...
		if(Tally[i]*m_NumPix[i]/4 > MaxTally)
			MaxTally = Tally[i]*m_NumPix[i]/4, m_BodyPix = i;
	}
	for(i=0;i<4;i++)
	{
		if(i == m_BodyPix || i == m_BackgroundPix || i == m_AntiAliasPixel1) continue;
		m_AntiAliasPixel2 = i;
	}

	// For testing...
//	assert(m_BodyPix == 1);
}


int CSubPic::OutputPicToTxtFile(const char* FileName)
{
	// Output "bitmap?"
	if(m_SPUReadOK == false)
		return -1;
	FILE* foo = fopen("FileName","w");
	if(!foo)
		return -1;
	int i,j;
	uchar* Ptr = m_Pixels;
	for(i=0;i<m_SubPicHeight;i++)
	{
		for(j=0;j<m_SubPicWidth;j++)
			fprintf(foo,"%d ",*Ptr++);
		fprintf(foo,"\n");
	}
	fclose(foo);
	return 0;
}

int CSubPic::FindCommand(CommandID Command, int DCSQT, Commands** TheCommand)
{
	if(m_SPUReadOK == false)
		return -1;
	// Go to DCSQT
	_ASSERT(DCSQT < m_DCSQT.size());
	if(DCSQT >= m_DCSQT.size())
		return -1;

	SP_DCSQT* TheDCSQT = &m_DCSQT[DCSQT];
	int i;
	for(i=0;i<TheDCSQT->m_Commands.size();i++)
	{
		if(TheDCSQT->m_Commands[i].m_CommandID == Command)
		{
			if(TheCommand)
				*TheCommand = &TheDCSQT->m_Commands[i];
			return TheDCSQT->m_Commands[i].m_Offset;
		}
	}
	return -1;
}

int CSubPic::ChangeCommand(CommandID OldCommand, CommandID NewCommand, int DCSQT)
{
	int Offset = FindCommand(OldCommand,DCSQT);
	if(Offset == -1)
		return -1;
	// Now change the command!
	m_Data[Offset] = (uchar)NewCommand;
	ParseCommands();
	SelectDCSQT(m_SelectedDCSQT);
	if(NewCommand != OldCommand)
		m_Modified = true;
	return 0;
}

// This reads 4 color or alpha values and stores them into the Dest array.
void ReadColorAlpha(ushort* Dest, uchar* Src)
{
	Dest[3] = Src[0]>>4;
	Dest[2] = Src[0]&0x0F;
	Dest[1] = Src[1]>>4;
	Dest[0] = Src[1]&0x0F;
}

// This takes an array of 4 color or alpha values and stores them into the Dest array.
void WriteColorAlpha(uchar* Dest, ushort* Src)
{
	Dest[1] &= 0xF0; Dest[1] |= Src[0];		
	Dest[1] &= 0x0F; Dest[1] |= (Src[1] << 4);
	Dest[0] &= 0xF0; Dest[0] |= Src[2];		
	Dest[0] &= 0x0F; Dest[0] |= (Src[3] << 4);
}

int CSubPic::GetNumDCSQTCanAdd(CommandID* Command, int NumCmd)
{
	// This is how many we can add.
	int Size = GetDCSQTSize(Command,NumCmd);
	return Size ? (m_RemainingSpace / Size) : 0;
}

int CSubPic::GetRequiredSpaceForDCSQT(int NumDCSQT, CommandID* Command, int NumCmd)
{
	int Size = GetDCSQTSize(Command,NumCmd);
	return Size*NumDCSQT;
}


int CSubPic::GetDCSQTSize(CommandID* Command, int NumCmd)
{
	// Calculate how much space we need to add a DCSQT. We need 2 bites for the timing, 2 bytes for the
	// offset to the next DCSQT, 1 byte for the command, and some bytes for the arguments.
	int ArgumentSize = 0;
	int i;
	for(i=0;i<NumCmd;i++)
	{
		switch(Command[i])
		{
			// For now, I'm only allowing adding transparency and color commands.
		case FSTA_DSP : return 0; break;
		case STA_DSP : return 0; break;
		case STP_DSP  : ArgumentSize += 0; break;
		case SET_COLOR : ArgumentSize += 2; break;
		case SET_CONTR : ArgumentSize += 2; break;
		case SET_DAREA : return 0; break;
		case SET_DSPXA : return 0; break;
		case CHG_COLCON : return 0; break;
		case CMD_END : break;
		}
	}
	// This is the size.
	return (2+2+1*NumCmd+ArgumentSize);
}


int CSubPic::AddDCSQT(int DCSQTBefore, int DelayMs, CommandID* Command, ushort (*Parameters)[4], int NumCmd)
{
	// Check that we have space for this command.
	if(GetNumDCSQTCanAdd(Command,NumCmd) < 1 || DCSQTBefore >= m_DCSQT.size())
	{
		ERROR1("Not enough space in the subpic pack to add another DCSQT!");
		return -1;
	}
	// Allocate more memory, and copy stuff back into it.
	int DCSQTSize = GetDCSQTSize(Command,NumCmd);
	int NewSubPicSize = m_SubPicSize + DCSQTSize;
	SP_DCSQT NewDCSQT;
	if(DCSQTBefore < m_DCSQT.size()-1)
	{
		// Not the last one.
		// Move DCSQT that are below the insertion point.
		SP_DCSQT After = m_DCSQT[DCSQTBefore+1];
		memmove(m_Data+After.m_OffsetToThis+DCSQTSize,m_Data+After.m_OffsetToThis,
			m_SubPicSize - After.m_OffsetToThis);
		NewDCSQT.m_OffsetToThis = m_DCSQT[DCSQTBefore+1].m_OffsetToThis;
		// Modify the offsets to account for the move.
		int i;
		for(i=DCSQTBefore+1;i<m_DCSQT.size();i++)
		{
			m_DCSQT[i].m_OffsetToThis += DCSQTSize;
		}
	}
	else
	{
		// Adding at end.
		NewDCSQT.m_OffsetToThis = m_SubPicSize;
	}
	// Add offset, timing, command and argument. The delay has to be rounded to the next frame.
	// DelayMs * 90 / 3600 is the closest frame (in PAL). multiply by 3600/1024 to get the subpic unit.
	if(VideoFormat == FORMAT_PAL)
		NewDCSQT.m_Delay = ((ushort)(DelayMs * 90.0f / 3600 + 0.5f)) * 3600/1024;
	else
		NewDCSQT.m_Delay = ((ushort)(DelayMs * 90.0f / 3003 + 0.5f)) * 3003/1024;
	Commands TheCommand;
	int i;
	int CmdOffset = 4;
	for(i=0;i<NumCmd;i++)
	{
		TheCommand.m_CommandID = Command[i];
		TheCommand.m_Offset = CmdOffset;
		TheCommand.m_NumParams = (Command[i] == CMD_END || Command[i] == STP_DSP) ? 0 : 4;
		memcpy(TheCommand.m_Params,Parameters[i],TheCommand.m_NumParams*sizeof(ushort));
		NewDCSQT.m_Commands.push_back(TheCommand);
		// FIX!
		CmdOffset += (Command[i] == CMD_END || Command[i] == STP_DSP) ? 1 : 3;
	}
	// Insert in our vector.
	if(DCSQTBefore < m_DCSQT.size()-1)
		m_DCSQT.insert(m_DCSQT.begin()+DCSQTBefore+1,NewDCSQT);
	else m_DCSQT.push_back(NewDCSQT);
	// Now write that to m_Data.
	for(i=0;i<m_DCSQT.size();i++)
	{
		if(i < m_DCSQT.size()-1)
			m_DCSQT[i].m_OffsetToNext = m_DCSQT[i+1].m_OffsetToThis;
		else
			m_DCSQT[i].m_OffsetToNext = m_DCSQT[i].m_OffsetToThis;
		// Write these as we've modified them!
		PutShort(m_Data,m_DCSQT[i].m_OffsetToThis+2,m_DCSQT[i].m_OffsetToNext);
	}
	// And write the data in our DCSQT.
	int OurOffset = DCSQTBefore+1;
	PutShort(m_Data,m_DCSQT[OurOffset].m_OffsetToThis,m_DCSQT[OurOffset].m_Delay);
	for(i=0;i<NumCmd;i++)
	{
		TheCommand = m_DCSQT[OurOffset].m_Commands[i];
		m_Data[m_DCSQT[OurOffset].m_OffsetToThis+TheCommand.m_Offset]=TheCommand.m_CommandID;
		// FIX: WRITE PARAMETERS!
		if(TheCommand.m_NumParams == 4)
			WriteColorAlpha(m_Data+m_DCSQT[OurOffset].m_OffsetToThis+TheCommand.m_Offset+1, TheCommand.m_Params);
	}
	// Update the subpic size!
	m_SubPicSize = NewSubPicSize;
	PutShort(m_Data,0,m_SubPicSize);
	m_RemainingSpace -= DCSQTSize;
	m_Modified = true;
	// I shouldn't have to do that...
	ParseCommands();
	// Finally, select the DCSQT that was selected before.
	if(m_SelectedDCSQT > DCSQTBefore) m_SelectedDCSQT++;
	SelectDCSQT(m_SelectedDCSQT);
	return 0;
}

int CSubPic::RemoveDCSQT(int Which)
{
	if(m_DCSQT.size() <= 1)
	{
		ERROR1("I can't remove the last remaining DCSQT!");
		return -1;
	}
	if(Which == 0)
	{
		ERROR1("I can't remove the first DCSQT!");
		return -1;
	}
	if(Which == -1) Which = m_DCSQT.size() -1;
	if(Which >= m_DCSQT.size()) return -1;
	// Check that we're not removing the last col/constrast!
	int i;
	// Calculate by how much to move.
	int DCSQTSize = 0;
	if(Which < m_DCSQT.size() - 1)
	{
		// Not the last one.
		DCSQTSize = m_DCSQT[Which+1].m_OffsetToThis - m_DCSQT[Which].m_OffsetToThis;
		memmove(m_Data+m_DCSQT[Which].m_OffsetToThis,m_Data+m_DCSQT[Which+1].m_OffsetToThis,m_SubPicSize - m_DCSQT[Which+1].m_OffsetToThis);
	}
	else DCSQTSize = m_SubPicSize - m_DCSQT[Which].m_OffsetToThis;
	for(i=Which;i<m_DCSQT.size();i++)
	{
		m_DCSQT[i].m_OffsetToThis -= DCSQTSize;
	}
	m_DCSQT.erase(m_DCSQT.begin()+Which);
	for(i=0;i<m_DCSQT.size();i++)
	{
		if(i < m_DCSQT.size()-1) m_DCSQT[i].m_OffsetToNext = m_DCSQT[i+1].m_OffsetToThis;
		else m_DCSQT[i].m_OffsetToNext = m_DCSQT[i].m_OffsetToThis;
		int u = m_DCSQT[i].m_OffsetToThis;
		// Write these as we've modified them!
		PutShort(m_Data,m_DCSQT[i].m_OffsetToThis+2,m_DCSQT[i].m_OffsetToNext);
	}
	m_SubPicSize -= DCSQTSize;
	m_RemainingSpace += DCSQTSize;
	PutShort(m_Data,0,m_SubPicSize);
	m_Modified = true;
	ParseCommands();
	// Finally, select the DCSQT that was selected before.
	if(m_SelectedDCSQT > Which || m_SelectedDCSQT == m_DCSQT.size()-1) m_SelectedDCSQT--;
	if(m_SelectedDCSQT >= m_DCSQT.size()-1) m_SelectedDCSQT = 0;
	SelectDCSQT(m_SelectedDCSQT);
	return 0;
}

int CSubPic::AddStop()
{
	// No need to add if there's one!
	if(FindCommand(STP_DSP,m_DCSQT.size()-1,0) != -1) return -1;
	CommandID Cmd[2] = {STP_DSP,CMD_END};
	ushort (*Param)[4] = new ushort[2][4];

	int RequiredSpace = GetRequiredSpaceForDCSQT(1,Cmd,2);
	MakeRoomForCommands(RequiredSpace);
	int OneFrame = VideoFormat == FORMAT_PAL ? (3600/90) : (3003/90);
	int Delay = GetDCSQTDelay(m_DCSQT.size()-1)+OneFrame;
	AddDCSQT(m_DCSQT.size()-1,Delay,Cmd,Param,2);
	delete[] Param;
	return 0;
}

int CSubPic::MakeRoomForCommands(int NumBytes)
{
	if(!m_SubPicDecoded && DecodeSubpic() == -1) return -1;
	if(NumBytes > m_SpaceBeforeCmd)
	{
		int FieldOffset = min(m_StartOfLine[0],m_StartOfLine[1]);
		if(FieldOffset > 4) MoveData(4-FieldOffset);
	}
	if(NumBytes <= m_SpaceBeforeCmd)
	{
		MoveAllCommands(-NumBytes);
		return NumBytes;
	}
	else return 0;
}

int CSubPic::AddFadeIn(bool Silent)
{
	int NumToAdd = 16;
	if(m_DCSQT.size() != 2) return -1;
	CommandID Cmd[2] = {SET_CONTR,CMD_END};
	if(!m_SubPicDecoded && DecodeSubpic() == -1)
			return -1;

	// Let's find out how many DCSQTs we'd like to add! Fade-in to last 1/8th of Delay.
	int Delay = GetDCSQTDelay(-1);
	int NumFrames = VideoFormat == FORMAT_PAL ? Delay/8/(3600/90) : Delay/8/(3003/90);
	NumToAdd = min(NumFrames,NumToAdd);
	if(NumToAdd == 0)
	{
		if(!Silent) ERROR1("Duration too short to add fade-in");
		return -1;
	}

	// Find out how much space we need, and if we can make that space.
	int RequiredSpace = GetRequiredSpaceForDCSQT(NumToAdd,Cmd,2);
	MakeRoomForCommands(RequiredSpace);

	int NumCanAdd = GetNumDCSQTCanAdd(Cmd,2);
	if(NumCanAdd < 4)
	{
		if(!Silent) ERROR1("Not enough room to add fade-in");
		return -1;
	}

	NumToAdd = min(NumCanAdd,NumToAdd);
	if(m_DCSQT.size() > 1 && m_DCSQT[m_DCSQT.size()-1].m_Delay * 1024/90 < 1000) 
	{
		if(!Silent) ERROR1("Subtitle too short to add fade-in (minimum 1s)");
		return -1;
	}
	int DelayInc = Delay / NumToAdd / 8;
	int CurrentTransp = GetMeanTransparency();
	// Now add NumToAdd commands.
	int i;
	ushort (*Param)[4] = new ushort[2][4]; memset(Param,0,16);
	for(i=0,Delay=DelayInc;i<NumToAdd;i++,Delay+=DelayInc)
	{
		AddDCSQT(m_DCSQT.size()-2,Delay,Cmd,Param,2);
	}
	for(i=0;i<NumToAdd+1;i++)
	{
		SelectDCSQT(i);
		ChangeTransparency(CurrentTransp*i/NumToAdd, 4);
	}
	delete[] Param;
	return 0;
}

int CSubPic::RemoveFadeIn(bool Silent)
{
	// First, check that there is something to remove! We must have more than 2 DCSQT, with all but the first/last
	// having a single contrast command.
	if(m_DCSQT.size() <= 2) return -1;
	if(!m_SubPicDecoded && DecodeSubpic() == -1)
			return -1;
	int i;
	ushort MaxTransp = 0;
	for(i=1;i<m_DCSQT.size()-1;i++)
	{
		if(m_DCSQT[i].m_Commands.size() != 2 || m_DCSQT[i].m_Commands[0].m_CommandID != SET_CONTR)
			return -1;
		// Compute the maximum opacity;
		MaxTransp = max(MaxTransp,m_DCSQT[i].m_Commands[0].m_Params[0]);
		MaxTransp = max(MaxTransp,m_DCSQT[i].m_Commands[0].m_Params[1]);
		MaxTransp = max(MaxTransp,m_DCSQT[i].m_Commands[0].m_Params[2]);
		MaxTransp = max(MaxTransp,m_DCSQT[i].m_Commands[0].m_Params[3]);
	}
	// Remove all m_DCSQT except first and last.
	while(m_DCSQT.size() > 2)
	{
		if(RemoveDCSQT(1) != 0) return -1;
	}
	// Finally, set the opacity for the first DCSQT.
	ChangeTransparency(MaxTransp,4);
	return 0;
}


void CSubPic::GetMinMaxDCSQTDelay(int &MinDurMs, int &MaxDurMs, int DCSQT)
{
	if(DCSQT == -1) DCSQT = m_DCSQT.size()-1;
	DCSQT--;
	if(DCSQT < 0 || DCSQT >= m_DCSQT.size())
	{
		MinDurMs = MaxDurMs = 0;
		return;
	}
	// A CDSQT must be at least 1 frame duration later than the previous one.
	MinDurMs = DCSQT == 0 ? 0 : m_DCSQT[DCSQT].m_Delay * 1024/90;
	int FrameDur = (VideoFormat == FORMAT_PAL ? 40 : 33);
	MinDurMs += FrameDur;
	// There must be at least 1 frame duration between then last DCSQT and the next spu PTS.
	if(m_NextSPUInStream && m_NextSPUInStream->m_TruePTS > m_TruePTS)
		MaxDurMs = m_NextSPUInStream->m_TruePTS / 90 - m_TruePTS / 90 - FrameDur;
	else
		MaxDurMs = 64537 * 1024 / 90;
	MaxDurMs = max(MaxDurMs,MinDurMs);
	MaxDurMs = min(MaxDurMs,64537 * 1024 / 90);
}

int CSubPic::ChangeDCSQTDelay(int DurationMs, int DCSQT)
{
	// Get our boundaries, make sure the subpics don't step on one another.
	int MinDurMs, MaxDurMs;
	GetMinMaxDCSQTDelay(MinDurMs, MaxDurMs, DCSQT);
	DurationMs = max(DurationMs,MinDurMs);
	DurationMs = min(DurationMs,MaxDurMs);
	// Now adjust the delay of the DCSQT.
	if(DCSQT == -1) DCSQT = m_DCSQT.size()-1;
	if(DCSQT < 0 || DCSQT >= m_DCSQT.size())
		return -1;
	SP_DCSQT* TheDCSQT = &m_DCSQT[DCSQT];
	// The delay needs to be truncated to the nearest frame.
	if(VideoFormat == FORMAT_PAL)
		TheDCSQT->m_Delay = ((ushort)(DurationMs * 90.0f / 3600 + 0.5f)) * 3600/1024;
	else
		TheDCSQT->m_Delay = ((ushort)(DurationMs * 90.0f / 3003 + 0.5f)) * 3003/1024;
	uchar* Ptr = m_Data + TheDCSQT->m_OffsetToThis;
	PutShort(Ptr,0,TheDCSQT->m_Delay);
	m_Modified = true;
	return 0;
}

int CSubPic::GetDCSQTDelay(int DCSQT)
{
	if(DCSQT == -1) DCSQT = m_DCSQT.size()-1;
	SP_DCSQT* TheDCSQT = &m_DCSQT[DCSQT];
	// Round to next frame (that's really what's happening).
	float FrameDurMs = VideoFormat == FORMAT_PAL ? 3600.0f / 90 : 3003.0f / 90;
	return ceil(TheDCSQT->m_Delay * 1024.0 /90  / FrameDurMs) * FrameDurMs;
}

int CSubPic::GetEndPTS()
{
	if(m_DCSQT.size() == 1) 
	{
		if(m_NextSPUInStream)
			return m_NextSPUInStream->m_PTS;
		else return -1;
	}
	return m_PTS + GetDCSQTDelay()*90;
}

int CSubPic::ParseCommands(int NumDCSQT, int OffsetToDCSQT)
{
	static int CountWarnings = 0;
	if(m_SPUReadOK == false)
		return -1;
	_ASSERT(OffsetToDCSQT < m_SubPicSize);
	if(OffsetToDCSQT >= m_SubPicSize)
	{
		if(CountWarnings <= 3)
		{
			MESSAGE3("Error in Parse Command at LBA %d: Offset to DCSQT %d",m_LBAFrom[m_NumPackInSPU-1],OffsetToDCSQT);
			sprintf(Message,"Serious error while parsing commands at LBA %d\nSomething is wrong with this SPU\n",m_LBAFrom[m_NumPackInSPU-1]);
		}
		if(CountWarnings == 3)
		{
			sprintf(Message+strlen(Message),"\nI won't report this error again");
		}
		if(CountWarnings <= 3)
			ERROR1(Message);
		CountWarnings++;
		return -1;
	}
	if(NumDCSQT == 0)
	{
		m_DCSQT.clear();
		m_CHCOL.clear();
	}
	SP_DCSQT Foo;
	m_DCSQT.push_back(Foo);
	SP_DCSQT* TheDCSQT = &m_DCSQT[NumDCSQT];
	// Get time info and offset info.
	TheDCSQT->m_OffsetToThis = (NumDCSQT == 0) ? m_OffsetToCmd : OffsetToDCSQT;
	uchar* Ptr = m_Data + TheDCSQT->m_OffsetToThis;
	TheDCSQT->m_Delay = GetShort(Ptr,0);
	TheDCSQT->m_OffsetToNext = GetShort(Ptr,2);
	TheDCSQT->m_Commands.clear();

	// Now parse commands...
	Ptr += 4;
	bool Done = 0;
	while(!Done)
	{
		// Get command:
		// Add a command.
		Commands Temp;
		TheDCSQT->m_Commands.push_back(Temp);
		Commands *TheCmd = &TheDCSQT->m_Commands.back();
		TheCmd->m_CommandID = (CommandID) *Ptr;
		TheCmd->m_Offset = Ptr-m_Data;
		TheCmd->m_NumParams = 0;
		switch(*Ptr)
		{
		case FSTA_DSP : Ptr ++;  break;
		case STA_DSP : Ptr ++; break;
		case STP_DSP : Ptr ++; break;
		case SET_COLOR : 
			TheCmd->m_NumParams = 4;
			ReadColorAlpha(TheCmd->m_Params,Ptr+1);
			Ptr += 3; break;
		case SET_CONTR : 
			TheCmd->m_NumParams = 4;
			ReadColorAlpha(TheCmd->m_Params,Ptr+1);
			Ptr += 3; break;
		case SET_DAREA : 
			TheCmd->m_NumParams = 4;
			TheCmd->m_Params[3] = ((Ptr[1])<<4)|((Ptr[2])>>4);
			TheCmd->m_Params[2] = ((Ptr[2] &0x0F) <<8)|(Ptr[3]);
			TheCmd->m_Params[1] = ((Ptr[4])<<4)|((Ptr[5])>>4);
			TheCmd->m_Params[0] = ((Ptr[5] &0x0F) <<8)|(Ptr[6]);
			if(NumDCSQT == 0)
			{
				m_StartX = TheCmd->m_Params[3];
				m_EndX = TheCmd->m_Params[2]; 
				m_StartY = TheCmd->m_Params[1];
				m_EndY = TheCmd->m_Params[0]; 

				if(!m_SubPicDecoded)
				{ 
					m_StartYOrig = m_StartY ; m_EndYOrig = m_EndY ; 
					m_StartXOrig = m_StartX; m_EndXOrig = m_EndX;
				}
				_ASSERT(m_StartX < 1024 && m_EndX < 1024 && m_StartY < 1024 && m_EndY < 1024);
				m_SubPicWidth = m_EndX - m_StartX + 1;
				m_SubPicHeight = min(m_EndY,576-1) - m_StartY + 1;
//				m_SubPicHeight = min(m_EndY,m_DisplayHeight-1) - m_StartY + 1;
				m_ModSubPicWidth = ((3+m_SubPicWidth)/4)*4;
			}
			Ptr += 7; break;
		case SET_DSPXA : 
			TheCmd->m_NumParams = 2;
			TheCmd->m_Params[0] = m_TopFieldOffset = GetShort(Ptr,1);
			TheCmd->m_Params[1] = m_BotFieldOffset = GetShort(Ptr,3);
//			_ASSERT(m_TopFieldOffset < 2*m_SubPicSize / 3 && m_BotFieldOffset < 2*m_SubPicSize / 3);
			Ptr += 5; break;
		case CHG_COLCON :
			{
			Ptr++;
			CHG_COLCON_Command CHG_COLCONCommand;
			ParseCHG_COLCON(&Ptr,&CHG_COLCONCommand); 
			CHG_COLCONCommand.m_DCSQT = NumDCSQT;
			m_CHCOL.push_back(CHG_COLCONCommand);
			break;
			}
		case CMD_END : 	Done = 1; break;
		default : 
			// This should NOT happen! Something went wrong here!
			if(CountWarnings <= 3)
			{
				ERROR4("Inconsistent data in subpic %d at LBA %d. Unknown command %d\nIgnoring this subpic.", m_SPUNumber,
					m_LBAFrom[0],*Ptr);
			}
			CountWarnings++;
			// Setting this to ensure we're exiting!
			TheDCSQT->m_OffsetToNext = TheDCSQT->m_OffsetToThis;
			Done = 1; 
			return -1;
			break;
		}
	}
	if(TheDCSQT->m_OffsetToNext != TheDCSQT->m_OffsetToThis)
		return ParseCommands(m_DCSQT.size(), TheDCSQT->m_OffsetToNext);
//	SelectDCSQT(0);
	return 0;
}


int CSubPic::ParseCHG_COLCON(uchar** ThePtr, CHG_COLCON_Command* Command)
{
	int bands;
	int i;
	// First are two bytes indicating total size of parameter area, including size.
	uchar* Ptr = *ThePtr;
	ushort ParamAreaSize = GetShort(Ptr,0);

#if 0
	// Don't parse, return immediately!
	Ptr += ParamAreaSize;
	return 0;
#endif

	Ptr += 2;
	// http://www.mpucoder.com/DVD/spu.html#dcsq
	for(bands=0;bands<64;bands++)
	{
		LN_CTLI TheBand;
		// Look for termination code.
		if(GetInt(Ptr,0) == 0x0FFFFFFF)
			break;
		// 4 bytes. 0s ss nt tt where sss is m_CSLN n is m_NumPX_CTLI and ttt is m_CTLN
		TheBand.m_CSLN = GetShort(Ptr,0);				// 0x0sss
		Ptr += 2;
		TheBand.m_NumPX_CTLI = (Ptr[0]>>4);			// 0x0n
		TheBand.m_CTLN = GetShort(Ptr,0) & 0x0FFF;		// 0x0ttt
		Ptr += 2;
		
		_ASSERT(TheBand.m_NumPX_CTLI < 8);
		for(i=0;i<TheBand.m_NumPX_CTLI;i++)
		{
			// 6 bytes, 2 by 2. First the start column
			TheBand.m_Columns[i].m_StartColNum = GetShort(Ptr,0); 
			Ptr += 2;	// Then the new colors.
			ReadColorAlpha(TheBand.m_Columns[i].m_NewColors,Ptr);
			Ptr += 2;	// Then the alphas.
			ReadColorAlpha(TheBand.m_Columns[i].m_NewAlphas,Ptr);
			Ptr += 2;
		}
		Command->m_Bands.push_back(TheBand);
	}
	Ptr += 4;
	_ASSERT(Ptr - *ThePtr == ParamAreaSize);
	*ThePtr = Ptr;
	MESSAGE2("Found CHG_COLCON, %d bands defined",i);
	Command->m_NumBands = bands;

	return 0;
}


uchar Colors[3*4] = {0,0,0,255,255,255,0,255,0,0,0,255};

int CSubPic::SavePicAsPPM(const char* FileName, uchar *CLUT, float FactX, float FactY, uchar* DefColors)
{
	if(m_SPUReadOK == false)
		return -1;
	int NewHeight;
	int NewWidth;
	int TotalByteCount;
	int i,j;
	static uchar  ColorMap[12];

	PrepareColors(CLUT, DefColors, ColorMap);

	FILE* fd = fopen(FileName,"wb");
	if(fd == 0)
		return -1;
	
	uchar* BitMap = new uchar[(int)(m_SubPicHeight/FactY+1)*(int)(m_SubPicWidth/ FactX+1) * 3];
	uchar* dest = BitMap;

	NewWidth = (int)(m_SubPicWidth/FactX);

	for(i=NewHeight=0,TotalByteCount=0;;i++)
	{
		// Find start of new line. PPM files are stored top to bottom.
		int yindex = (int)(i * FactY + .5f);
		if(yindex >= m_SubPicHeight)
			break;
		uchar* src = m_Pixels + yindex * m_SubPicWidth;
		// Copy this line.
		for(j=0;j<NewWidth;j++)
		{
			// Find pixel location in line.
			int xindex = (int)(j * FactX + .5f);
			// Copy the three RGB components for that pixel. Note the order is different than for BMP files.
			uchar Pick = 3*src[xindex];
			_ASSERT(Pick < 12);
			*dest++ = ColorMap[Pick];
			*dest++ = ColorMap[Pick+1];
			*dest++ = ColorMap[Pick+2];
			TotalByteCount+=3;
		}
		NewHeight++;
	}

	fprintf (fd,"P6\n%d %d\n255\n", NewWidth, NewHeight);
	fwrite(BitMap,sizeof(char),TotalByteCount,fd);
	fclose(fd);
	delete[] BitMap;
	return 0;
}

// This finds, among the NumColors colors in Colors, the one that's closest to Pixels. Pixels is stored as BGR, while
// colors are RGB...
static int FindClosestColor(uchar* Pixels, uchar* Colors, int NumColors)
{
	int MinAbsDiff = 1e8;
	int MinDiff = -1;
	int i;
	for(i=0;i<NumColors;i++)
	{
		int AbsDiff = abs(Pixels[0] - Colors[3*i+2]);
		AbsDiff += abs(Pixels[1] - Colors[3*i+1]);
		AbsDiff += abs(Pixels[2] - Colors[3*i+0]);
		if(AbsDiff < MinAbsDiff) MinAbsDiff = AbsDiff, MinDiff = i;
	}
	_ASSERT(MinDiff != -1);
	return MinDiff;
}

int CSubPic::SavePicAsBMP(const char* FileName, int BitPerPixel, uchar *CLUT, uchar* DefColors)
{
	if(m_SPUReadOK == false)
		return -1;
	long Height = m_DisplayHeight;
	long Width = m_DisplayWidth;
	int i,j;
	FILE* fd = fopen(FileName,"wb");
	if(fd == 0)
		return -1;

	int TotalByteCount = Width*Height*3;
	uchar* Bitmap = new uchar[TotalByteCount];
	memset(Bitmap,0,TotalByteCount);

	TotalByteCount = Width*Height*BitPerPixel/8;

	// Color Array.
	uchar DefaultColors[12];
	uchar* Ptr = DefaultColors;
	PrepareColors(CLUT,DefColors,DefaultColors);

	// If we have buttons, use the button color map as our colors.
	if(m_NumButtons) memcpy(DefaultColors,m_ButColorMap,12);

	// Put background pixel into bitmap.
	for(i=0;i<Width*Height;i++)
	{
		Bitmap[3*i]=DefaultColors[3*m_BackgroundPix+2];
		Bitmap[3*i+1]=DefaultColors[3*m_BackgroundPix+1];
		Bitmap[3*i+2]=DefaultColors[3*m_BackgroundPix+0];
	}

	// "Mix" into the blank bitmap, but ignore alpha (no mixing then!).
	MixScaleIntoBitmap(Bitmap,Width,Height,CLUT,DefaultColors, true);

	int ColorTableSize = BitPerPixel == 24 ? 0 : 16*4;
	int ColorUsed = BitPerPixel == 24 ? 0 : 16;
	BITMAPFILEHEADER BMPHeader = {'MB',0,0,0,sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFOHEADER)+ColorTableSize};
	BITMAPINFOHEADER BMPInfoHeader = {0,Width,Height,1,BitPerPixel,BI_RGB,TotalByteCount,0,0,ColorUsed,ColorUsed};
	BMPHeader.bfSize = sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFOHEADER)+ColorTableSize+TotalByteCount;
	BMPInfoHeader.biSize = sizeof(BITMAPINFOHEADER);

	// Write Bitmap headers.
	fwrite(&BMPHeader,sizeof(BITMAPFILEHEADER),1,fd);
	fwrite(&BMPInfoHeader,sizeof(BITMAPINFOHEADER),1,fd);
	// If needed, write color table:
	if(BitPerPixel == 4)
	{
		uchar* ColorTable = new uchar[ColorTableSize];
		memset(ColorTable,0,ColorTableSize);
		for(i=0;i<4;i++)
		{
			ColorTable[4*i] = DefaultColors[3*i+2];
			ColorTable[4*i+1] = DefaultColors[3*i+1];
			ColorTable[4*i+2] = DefaultColors[3*i+0];
		}
		fwrite(ColorTable,1,ColorTableSize,fd);
		delete[] ColorTable;
	}
	// Then write data!
	if(BitPerPixel == 4)
	{
		char* Line = new char[Width/2];
		int k;
		// Go through bitmap, and go back to pixel type! Now, that's quite silly isn't it?
		for(i=0;i<Height;i++)
		{
			for(j=k=0;j<Width;j+=2,k++)
			{
				// This is somewhat silly! But I think that's the simplest way to do it at this point.
				int Val1 = FindClosestColor(Bitmap+3*i*Width+3*j,DefaultColors,4);
				int Val2 = FindClosestColor(Bitmap+3*i*Width+3*j+3,DefaultColors,4);
				Line[k] = (Val1 << 4) + Val2;
			}
			fwrite(Line,1,Width/2,fd);
		}
		delete[] Line;
	}
	else
		fwrite(Bitmap,sizeof(char),TotalByteCount,fd);
	fclose(fd);
	delete[] Bitmap;
	return 0;
}

int CSubPic::LoadBMP(const char* FileName, uchar *CLUT, uchar* DefColors)
{
	if(m_SPUReadOK == false)
		return -1;
	if(!m_SubPicDecoded && DecodeSubpic() == -1) return -1;
	int Ret = -1;
	FILE* fd = fopen(FileName,"rb");
	if(fd == 0) return -1;
	long Height = m_DisplayHeight;
	long Height2;
	long Width = m_DisplayWidth;
	int TotalByteCount = Width*Height*3;
	int i,j;
	int NewLeftMostPixel, NewRightMostPixel, NewTopMostPixel, NewBottomMostPixel, NewBackgroundPixel;
	uchar* Ptr = m_Pixels;
	uchar* Bitmap = 0;
	NewBackgroundPixel = -1;
	// We'll need to remember these in case the encoding does not work.
	ushort SX,SY,EX,EY;
	int LastV, LastH;
	bool UseColorIndex;
	uchar DefaultColors[12];
	uchar BitMapColors[16*4];	// Table of colors found in bitmap.
	int ColorMapping[16];		// Maps colors in table found in bitmap to our 4 colors.
	bool empty = true;
	int skip = 0;

	// First of move the commands to give us as much room as possible!
	if(m_RemainingSpace > 5) MoveAllCommands(m_RemainingSpace -5);
	// Also, move data all the way down.
	int FieldOffset = min(m_StartOfLine[0],m_StartOfLine[1]);
	if(FieldOffset > 4) MoveData(4-FieldOffset);

	// Load the bitmap header
	BITMAPFILEHEADER BMPHeader;
	BITMAPINFOHEADER BMPInfoHeader;
	if(fread(&BMPHeader,sizeof(BITMAPFILEHEADER),1,fd) != 1) goto end;
	if(fread(&BMPInfoHeader,sizeof(BITMAPINFOHEADER),1,fd) != 1) goto end;
	// Check that the size is right.
	if(BMPInfoHeader.biSize != sizeof(BITMAPINFOHEADER))
	{
		ERROR1("Error: The bitmap must be standard Windows V.3 bitmap - Not OS2 at the moment!"); goto end;
	}
	if(BMPInfoHeader.biWidth != m_DisplayWidth || BMPInfoHeader.biHeight != m_DisplayHeight)
	{
		ERROR1("Error: The bitmap size should match the picture size for this DVD"); goto end;
	}
	// Check the compression, and the bit width.
	if(BMPInfoHeader.biCompression != BI_RGB || (BMPInfoHeader.biBitCount != 24 && BMPInfoHeader.biBitCount != 4))
	{
		ERROR1("Error: The bitmap must be an uncompressed 24bit or 4bit RGB bitmap"); goto end;
	}
	// Allocate bitmap and read it.
	UseColorIndex = BMPInfoHeader.biBitCount == 4;
	if(UseColorIndex) 
	{
		TotalByteCount = Width*Height/2;
		// A value of 0 indicates 2 ^ number of bits...
		if(BMPInfoHeader.biClrUsed == 0) BMPInfoHeader.biClrUsed = 1 << BMPInfoHeader.biBitCount;
		// Read the color table, and drop the unused bytes.
		fread(BitMapColors,BMPInfoHeader.biClrUsed*4,1,fd);
		for(i=j=0;i<BMPInfoHeader.biClrUsed*4;i++)
		{
			if(i%4==3) i++;
			BitMapColors[j++]=BitMapColors[i];
		}
	}
	else TotalByteCount = Width*Height*3;
	Bitmap = new uchar[TotalByteCount];
	memset(Bitmap,0,TotalByteCount);
	if(fread(Bitmap,1,TotalByteCount,fd) != TotalByteCount)
	{
		ERROR1("Error Couldn't read enough data from bitmap file!"); goto end;
	}

	// Prepare colors, as if we were outputing a bitmap, so we can match incoming colors.
	PrepareColors(CLUT,DefColors,DefaultColors);
	// If we have buttons, use the button color map as our colors.
	if(m_NumButtons) memcpy(DefaultColors,m_ButColorMap,12);

	if(UseColorIndex) 
	{
		// Now's the time to find the best mapping between the colors found in the bitmap table, and our
		// four colors.
		for(i=j=0;i<BMPInfoHeader.biClrUsed;i++)
		{
			ColorMapping[i] = FindClosestColor(BitMapColors+i*3,DefaultColors,4);
		}
	}

	// Copy the bitmap into the m_Pixels array. Make sure we only have our 3 specific colors and translate
	NewLeftMostPixel = Width; NewRightMostPixel = 0;
	NewTopMostPixel = Height ; NewBottomMostPixel = 0;
//	for(i=Height-1;i>=0;i--)
	// This is tricky and messy. Upon reimporting, I will always skip the two first lines, and remove the last line
	// in PAL (as the height in PAL is maximum 573. Height2 is 478 in NTSC, and 573 in PAL.
	skip = 2;
	Height2 = VideoFormat == FORMAT_PAL ? m_DisplayHeight-3 : m_DisplayHeight -2;
	for(i=Height-3;i>=Height-Height2-2;i--)	// I hate this kind of crap!
	{
		for(j=0;j<Width;j++)
		{
			// Read bitmap pixel:
			if(UseColorIndex)
			{
				if(j % 2 == 0)
					*Ptr++ = ColorMapping[Bitmap[i*Width/2+j/2] >> 4];
				else
					*Ptr++ = ColorMapping[Bitmap[i*Width/2+j/2] & 0x07];
				_ASSERT(Ptr[-1] < 4);
			}
			else
			{
				*Ptr++ = FindClosestColor(Bitmap+3*i*Width+3*j,DefaultColors,4);
			}
			// Determine the background pixel, the left/rightmost and top/bottommost pixels.
			if(NewBackgroundPixel == -1) NewBackgroundPixel = Ptr[-1];
			if(Ptr[-1] != NewBackgroundPixel) 
			{
				empty = false;
				NewLeftMostPixel = min(NewLeftMostPixel, j);
				NewRightMostPixel = max(NewRightMostPixel, j);
				NewTopMostPixel = min(NewTopMostPixel,Height2-i);
				NewBottomMostPixel = max(NewBottomMostPixel,Height2-i);
			}
		}
	}
	if(empty)
	{
		NewLeftMostPixel = NewRightMostPixel = Width/2;
		NewTopMostPixel = NewBottomMostPixel = Height/2-2;
	}
	// Now, let's encode this new bitmap. We'll use the CropBackground() function, as this increases the chances
	// that the new encoded bitmap will fit in the buffer.
	m_LeftMostPixel = NewLeftMostPixel;
	m_RightMostPixel = NewRightMostPixel;
	m_TopMostPixel = NewTopMostPixel ;
	m_BottomMostPixel = NewBottomMostPixel;
	m_BackgroundPix = NewBackgroundPixel;
	SX=m_StartXOrig;
	SY=m_StartYOrig;
	EX=m_EndXOrig;
	EY=m_EndYOrig;
	m_StartX = m_StartXOrig = 0;
	m_StartY = m_StartYOrig = skip;	// I skipped skip lines, so reposition where needed.
	m_SubPicWidth = Width;
	m_SubPicHeight = Height2;
	m_EndX = m_EndXOrig = Width-1;
	m_EndY = m_EndYOrig = m_StartY + Height2-1;
	LastV = m_LastVMove;
	LastH = m_LastHMove;
	// Because the bitmap was saved in its "moved" position, we shouldn't re-apply the H/V move!
	m_LastVMove = 0;
	m_LastHMove = 0;

	// Then call crop.
	if(CropBackground(0) == -1)
	{
		// Argh! Not enough space to fit the new bitmap.
		ERROR1("Ooops! The modified encoded subpic does not fit in the available space on the DVD!");
		// This isn't very nice. I have to reset these because they were changed above.
		m_StartXOrig=SX;
		m_StartYOrig=SY;
		m_EndXOrig=EX;
		m_EndYOrig=EY;
		// These are necessary so these values are recomputed.
		m_BottomMostPixelT = m_BottomMostPixel = m_TopMostPixel = -1;
		m_RightMostPixel = 0; m_LeftMostPixel = 720; m_BackgroundPix = -1;
		// This takes care of resetting everything that needs to be reset.
		m_SubPicDecoded = 0;
		m_TopFieldOffset = m_StartOfLine[0];
		m_BotFieldOffset = m_StartOfLine[1];
		m_StartY = m_StartYOrig;
		m_EndY = m_EndYOrig;
		m_StartX = m_StartXOrig;
		m_EndX = m_EndXOrig;
		m_SubPicHeight = m_EndY - m_StartY + 1;
		m_SubPicWidth = m_EndX-m_StartX+1;
		DecodeSubpic();
		// We need to restore these...
		m_LastVMove = LastV;
		m_LastHMove = LastH;
		VMoveSubPic(m_LastVMove);
		HMoveSubPic(m_LastHMove);
		Ret = -1;
	}
	else Ret = 0;

end:
	fclose(fd);
	delete[] Bitmap;
	return Ret;
}

inline bool CSubPic::IsPixelInRect(int x, int y)
{
	// Checks whether pixel x,y is within the rectangle of the current button.
	if(m_NumButtons == 0 || m_SelButton < 0 || 	m_SelButton >= 36) return false;
	if(x < m_ButtonStartX[m_SelButton] || x > m_ButtonStartX[m_SelButton] + m_ButtonSizeX[m_SelButton]) return false;
	if(y < m_ButtonStartY[m_SelButton] || y > m_ButtonStartY[m_SelButton] + m_ButtonSizeY[m_SelButton]) return false;
	return true;
}


bool CSubPic::CheckForIdenticalColors(uchar* CLUT, uchar* DefColors)
{
	uchar DefaultColors[12];
	PrepareColors(CLUT,DefColors,DefaultColors);
	// If we have buttons, use the button color map as our colors.
	if(m_NumButtons) memcpy(DefaultColors,m_ButColorMap,12);
	int i,j;
	for(i=0;i<4;i++)
	{
		for(j=i+1;j<4;j++)
		{
			// Check that the ith color isn't the same as the jth.
			if(DefaultColors[i*3]==DefaultColors[j*3] && DefaultColors[i*3+1]==DefaultColors[j*3+1] && DefaultColors[i*3+2]==DefaultColors[j*3+2])
				return false;
		}
	}
	return true;
}

void CSubPic::PrepareColors(uchar* CLUT, uchar* DefColors, uchar* TargetColorMap)
{
	if(CLUT)
	{
		memcpy(TargetColorMap,CLUT+3*m_ColorPtr[0],3*sizeof(uchar));
		memcpy(TargetColorMap+3,CLUT+3*m_ColorPtr[1],3*sizeof(uchar));
		memcpy(TargetColorMap+6,CLUT+3*m_ColorPtr[2],3*sizeof(uchar));
		memcpy(TargetColorMap+9,CLUT+3*m_ColorPtr[3],3*sizeof(uchar));
	}
	else if(DefColors)
		memcpy(TargetColorMap,DefColors,12*sizeof(char));
	else
	{
		// Using m_BodyPix to mark the pixels.
		if(1 || m_NumButtons == 0)
		{
			for(int i=0;i<4;i++)
			{
				if(i==m_BodyPix)
				{ memcpy(TargetColorMap+3*i,m_DefColors,3);}
				else if(i==m_BackgroundPix)
				{ memcpy(TargetColorMap+3*i,m_DefColors+3,3);}
				else if(i==m_AntiAliasPixel1)
				{ memcpy(TargetColorMap+3*i,m_DefColors+6,3);}
				else if(i==m_AntiAliasPixel2)
				{ memcpy(TargetColorMap+3*i,m_DefColors+9,3);}
			}
		}
		else
			memcpy(TargetColorMap,m_DefColors,12*sizeof(char));
	}

	// JEAN EXPERIMENTAL!
	int ColScheme = m_SelButton < 36 ? m_ButtonScheme[m_SelButton]-1 : -1;
	if(m_NumButtons && CLUT && ColScheme >= 0 && ColScheme < 3)
	{
		if(m_ShowState == 1)
		{	
			// Activated!
			memcpy(m_ButColorMap,CLUT+3*m_CSColorsSel[ColScheme][0],3*sizeof(uchar));
			memcpy(m_ButColorMap+3,CLUT+3*m_CSColorsSel[ColScheme][1],3*sizeof(uchar));
			memcpy(m_ButColorMap+6,CLUT+3*m_CSColorsSel[ColScheme][2],3*sizeof(uchar));
			memcpy(m_ButColorMap+9,CLUT+3*m_CSColorsSel[ColScheme][3],3*sizeof(uchar));
		}
		else if(m_ShowState == 2)
		{
			memcpy(m_ButColorMap,CLUT+3*m_CSColorsAct[ColScheme][0],3*sizeof(uchar));
			memcpy(m_ButColorMap+3,CLUT+3*m_CSColorsAct[ColScheme][1],3*sizeof(uchar));
			memcpy(m_ButColorMap+6,CLUT+3*m_CSColorsAct[ColScheme][2],3*sizeof(uchar));
			memcpy(m_ButColorMap+9,CLUT+3*m_CSColorsAct[ColScheme][3],3*sizeof(uchar));
		}
		else if(m_ShowState == 0)
			memcpy(m_ButColorMap,TargetColorMap,12*sizeof(uchar));

	}
	// If there's no CLUT, use the colors computed above as our buttons.
	if(CLUT == 0 && m_NumButtons) memcpy(m_ButColorMap,TargetColorMap,12*sizeof(char));
}

void CSubPic::PrepareAlphas(uchar* TargetAlphaMap)
{
	// Get alpha and color tables.
	TargetAlphaMap[0]=m_AlphaPtr[0]*17;
	TargetAlphaMap[1]=m_AlphaPtr[1]*17;
	TargetAlphaMap[2]=m_AlphaPtr[2]*17;
	TargetAlphaMap[3]=m_AlphaPtr[3]*17;
	int ColScheme = m_SelButton < 36 ? m_ButtonScheme[m_SelButton]-1 : -1;
	if(m_NumButtons && ColScheme >= 0 && ColScheme < 3)
	{
		if(m_ShowState == 1)
		{	
			// Selected.
			m_ButAlphaMap[0]=m_CSAlphasSel[ColScheme][0]*17;
			m_ButAlphaMap[1]=m_CSAlphasSel[ColScheme][1]*17;
			m_ButAlphaMap[2]=m_CSAlphasSel[ColScheme][2]*17;
			m_ButAlphaMap[3]=m_CSAlphasSel[ColScheme][3]*17;
		}
		else if(m_ShowState == 2)
		{
			// Activated
			m_ButAlphaMap[0]=m_CSAlphasAct[ColScheme][0]*17;
			m_ButAlphaMap[1]=m_CSAlphasAct[ColScheme][1]*17;
			m_ButAlphaMap[2]=m_CSAlphasAct[ColScheme][2]*17;
			m_ButAlphaMap[3]=m_CSAlphasAct[ColScheme][3]*17;
		}
		else memcpy(m_ButAlphaMap,TargetAlphaMap,4);
	}

	int i;
	if((m_PixTempAlpha == 16))
	{
		// Showing button highlights, all pixels (except background) are opaque.
		for(i=0;i<4;i++)
			if(m_BackgroundPix != i) TargetAlphaMap[i] = 15*17;
	}
	else if(m_PixTempAlpha)
	{
		// Showing individual pixels as opaque (others as transparent).
		for(i=0;i<4;i++)
			TargetAlphaMap[i] = ((1<<i) & m_PixTempAlpha) ? 15*17 : 0;
	}
}

uchar* CSubPic::SaveToBitmap(uchar* Bitmap, long &Width, long &Height, float FactX, float FactY, 
							 uchar *CLUT, uchar* DefColors, bool OnlyNonBackground, int DisplayMode)
{
	uchar* dest = Bitmap;
	static uchar  ColorMap[12];
	uchar AlphaMap[4] = {255,255,255,255};
	int i,j;

	// Get Alpha and color tables...
	PrepareAlphas(AlphaMap);	
	PrepareColors(CLUT, DefColors, ColorMap);


	int StartLine = 0;
	int StartCol = 0;
	int EndLine = 0;
	int EndCol = m_SubPicWidth-1;
	int CurX, CurY;
	if(OnlyNonBackground)
	{
		// Only displaying the useful part of the subpics.
		// This is tricky: m_VerticalAdjust stores by how many lines the subpic was moved in the
		// coded domain. Tricky tricky... I don't like it very much. The problem is that SubPicHeight
		// gets modified when you move the subpics, but not Top and Bottom pixels. This makes things much
		// more complicated than they should be.
		StartLine = m_SubPicHeight-min(m_BottomMostPixel+4,m_SubPicHeight-m_VerticalAdjust)-m_VerticalAdjust;
		StartLine /= FactY;
		EndLine = max(0,m_TopMostPixel-3+m_VerticalAdjust);
		StartCol = m_LeftMostPixel / FactX;
		EndCol = m_RightMostPixel;
	}
	for(i=StartLine,Height=0;;i++)
	{
		CurY = m_StartY+m_SubPicHeight-1-(i)*FactY;
		if(DisplayMode == 1)
			CurY =   (m_StartY+m_SubPicHeight-1 - m_DisplayHeight/8 -i * FactY )*4/3; // Adjustment for LB!
		// Find start of new line. Bitmaps are stored bottom to top.
		int yindex = m_SubPicHeight - 1 - (int)(i* FactY + .5f);
		// This can happen if you're displaying a PAL subpic into a NTSC video!
		if(yindex-StartLine >= m_DisplayHeight) continue;
		if(yindex < EndLine)
			break;
		uchar* src = m_Pixels + yindex * m_SubPicWidth;
		// Copy this line.
		for(j=StartCol,Width=0;;j++)
		{
			uchar* CMap = ColorMap;
			CurX = m_StartX + (j) * FactX;
			if(DisplayMode == 2)
				CurX = (m_StartX + (j) * FactX) * 3/4 + m_DisplayWidth / 8;
			// Copy the three RGB components for that pixel. The order is BGR!
			int xindex = (int)(j * FactX + .5f);
			if(xindex > EndCol) break;
			uchar Pick = 3*src[xindex];
			ushort Alpha = AlphaMap[src[xindex]];
			// If m_PixTempAlpha is non zero, show all highlights no matter where they fall.
			if(m_NumButtons && (m_PixTempAlpha || IsPixelInRect(CurX,CurY)))
			{ 
				if(!m_PixTempAlpha) Alpha = m_ButAlphaMap[src[xindex]];
				CMap = m_ButColorMap;
			}
			_ASSERT(Pick < 12);
			// Mixing into a background painted with m_BackColor.
			*dest++ = m_BackColor[2] *(255-Alpha)/255 + CMap[Pick+2]*Alpha/255;	// B
			*dest++ = m_BackColor[1] *(255-Alpha)/255 + CMap[Pick+1]*Alpha/255;	// G
			*dest++ = m_BackColor[0] *(255-Alpha)/255 + CMap[Pick]*Alpha/255;	// R
			Width++;
		}
		// Here, I'm extending the bitmap with a few more pixels if needed.
		for(j;Width%4!=0;j++)
		{
			*dest++ = m_BackColor[2];
			*dest++ = m_BackColor[1];
			*dest++ = m_BackColor[0];
			Width++;
		}
		Height++;
	}
	return m_NumButtons ? m_ButColorMap : ColorMap;
}


uchar* CSubPic::MixScaleIntoBitmap(uchar* Bitmap, long Width, long Height, 
								   uchar *CLUT, uchar* DefColors, bool IgnoreAlpha, int DisplayMode)
{
	uchar* dest = Bitmap;
	static uchar  ColorMap[12];
	uchar AlphaMap[4] = {255,255,255,255};
	int i,j;

	// Get alpha and color tables.
	PrepareAlphas(AlphaMap);
	PrepareColors(CLUT, DefColors, ColorMap);

	// Rescale our subpic to match the target output size. 
	float ScaleX = (float)m_DisplayWidth / Width;
	float ScaleY = (float)m_DisplayHeight / Height;
	int StartX = (int)(m_StartX/ScaleX);
	int StartY = (int)(m_StartY/ScaleY);

	int CurX, CurY;
	for(i=0;;i++)
	{
		// Find start of new line. Bitmaps are stored bottom to top, but our pixel array is stored
		// from top to bottom. FIX: Use the startX and startY values here!!!
		uchar* src = m_Pixels + (int)(i*ScaleY+0.5) * m_SubPicWidth;
		if((int)(i*ScaleY+0.5) >= m_SubPicHeight || Height-i-1-((int)(StartY)) < 0)
			break;
		dest = Bitmap + 3*(Height-i-1-((int)(StartY)))*Width + 3*StartX;
		CurY = m_StartY+i*ScaleY;
		if(DisplayMode == 1)
			CurY = (m_StartY - m_DisplayHeight/8)*4/3  + (i * ScaleY )*4/3; // Adjustment for LB!
		// Mix this line.
		for(j=0;;j++,dest+=3)
		{
			CurX = m_StartX + j * ScaleX;
			if(DisplayMode == 2)
				CurX = (m_StartX + j * ScaleX)*3/4 + m_DisplayWidth / 8; // Adjustment for P&S!
			// Mix the three RGB components for that pixel. The order is BGR!
			int Ind = (int)(j*ScaleX);
			if(Ind >= m_SubPicWidth)
				break;
			uchar Pick = 3*src[Ind];
			_ASSERT(Pick < 12);
			ushort Alpha = IgnoreAlpha ? 255 : AlphaMap[src[Ind]];
			uchar* CMap = ColorMap;
			// If IgnoreAlpha or m_PixTempAlpha is non zero, show all highlights no matter where they fall.
			if(m_NumButtons && (IgnoreAlpha || m_PixTempAlpha || IsPixelInRect(CurX,CurY)))
			{ 
				if(!IgnoreAlpha && !m_PixTempAlpha) Alpha = m_ButAlphaMap[src[Ind]];
				CMap = m_ButColorMap;
			}
			// Don't do anything if our pixel is transparent!
			if(Alpha != 0)
			{
				// I could make this faster by doing Alpha++ and using 256 instead of 255..
				dest[0] = dest[0]*(255-Alpha)/255 + CMap[Pick+2]*Alpha/255;	// B
				dest[1] = dest[1]*(255-Alpha)/255 + CMap[Pick+1]*Alpha/255;	// G
				dest[2] = dest[2]*(255-Alpha)/255 + CMap[Pick]*Alpha/255;	// R
			}
		}
	}
	return m_NumButtons ? m_ButColorMap : ColorMap;
}



void CSubPic::WriteDAREA(ushort StartX, ushort StartY, ushort EndX, ushort EndY)
{
	Commands *TheCommand;
	int Offset = FindCommand(SET_DAREA,0,&TheCommand);
	if(Offset != -1)
	{
		m_Data[Offset+1] = (uchar)(StartX >> 4);
		m_Data[Offset+2] = (uchar)((StartX & 0x000F)<<4);
		m_Data[Offset+2] |= (uchar)(EndX >> 8);
		m_Data[Offset+3] = (uchar)(EndX & 0x00FF);
		m_Data[Offset+4] = (uchar)(StartY >> 4);
		m_Data[Offset+5] = (uchar)((StartY & 0x000F)<<4);
		m_Data[Offset+5] |= (uchar)(EndY >> 8);
		m_Data[Offset+6] = (uchar)(EndY & 0x00FF);

		TheCommand->m_Params[3] = StartX;
		TheCommand->m_Params[2] = EndX;
		TheCommand->m_Params[1] = StartY;
		TheCommand->m_Params[0] = EndY;
	}
}

void CSubPic::WriteDSPXA(ushort TopOffset, ushort BotOffset)
{
	int Offset = FindCommand(SET_DSPXA);
	if(Offset != -1)
	{
		PutShort(m_Data,Offset+1,TopOffset);
		PutShort(m_Data,Offset+3,BotOffset);
	}
}

void CSubPic::SwapFieldsIfNeeded()
{
	if(!m_SubPicDecoded && DecodeSubpic() == -1) return;
	// This is tricky, but needed by some players (such as mpc) which don't like the top field offset to be
	// larger than the bottom field offset.
	if(m_TopFieldOffset < m_BotFieldOffset) return;
	// Here, I'm swapping the entire original data (not just what's referenced by top and bottom offsets).
	// It's quite possible for m_StartOfLine[1] to be smaller than m_StartOfLine[0], if we've swapped the field
	// once before! This makes sure OrigTop is the larger offset, and OrigBot is the smaller one.
	int OrigTop = max(m_StartOfLine[0],m_StartOfLine[1]);
	int OrigBot = min(m_StartOfLine[0],m_StartOfLine[1]);
	// These are the length of the top and bottom fields. 
	int LengthBot = OrigTop-OrigBot;
	int LengthTop = m_OffsetToCmd-OrigTop-m_SpaceBeforeCmd;
	uchar* TempData = new uchar[LengthBot];
	// I move the bottom field (appearing first) into a temp array.
	memcpy(TempData,m_Data+OrigBot,LengthBot);
	memmove(m_Data+OrigBot,m_Data+OrigTop,LengthTop);
	memcpy(m_Data+OrigBot+LengthTop,TempData,LengthBot);
	// Now update the starts of lines: Even ones have been moved by 
	int i;
	if(m_StartOfLine[0] < m_StartOfLine[1])
	{
		for(i=0;i<MAX_HEIGHT;i+=2)
		{
			m_StartOfLine[i] += LengthTop;
			m_StartOfLine[i+1] -= LengthBot;
		}
	}
	else
	{
		for(i=0;i<MAX_HEIGHT;i+=2)
		{
			m_StartOfLine[i] -= LengthBot;
			m_StartOfLine[i+1] += LengthTop;
		}
	}

	m_TopFieldOffset -= LengthBot;
	m_BotFieldOffset += LengthTop;
	WriteDSPXA(m_TopFieldOffset,m_BotFieldOffset);
	delete[] TempData;
}

void CSubPic::MoveData(int NumBytes)
{
	if(!m_SubPicDecoded && DecodeSubpic() == -1) return;
	int FieldOffset = min(m_StartOfLine[0],m_StartOfLine[1]);
	int NumMove = NumBytes > 0 ? m_DCSQT[0].m_OffsetToThis - FieldOffset - NumBytes : 
		m_DCSQT[0].m_OffsetToThis - FieldOffset;
_ASSERT(NumMove >= 0);
	memmove(m_Data+FieldOffset+NumBytes,m_Data+FieldOffset,NumMove);
	// Zero out after you!
	if(NumBytes > 0) memset(m_Data+4,0,NumBytes);
	else memset(m_Data+m_DCSQT[0].m_OffsetToThis+NumBytes,0,-NumBytes);

	// Update offset to top and bottom fields! and the space before commands. The subpic size does not change.
	m_TopFieldOffset += NumBytes;
	m_BotFieldOffset += NumBytes;
	int i;
	for(i=0;i<MAX_HEIGHT;i++)
	{
		m_StartOfLine[i] += NumBytes;
	}
	_ASSERT(m_TopFieldOffset >= 4 && m_BotFieldOffset >= 4);
	WriteDSPXA(m_TopFieldOffset,m_BotFieldOffset);
	m_SpaceBeforeCmd -= NumBytes;
}


void CSubPic::MoveAllCommands(int NumBytes)
{
	int i;
	// One one player (a sony) moving commands by an odd number of bytes creates a problem, who knows why?!
	// So.... Let's constrain our move to the largest (in absolute magnitude) number below NumBytes
	NumBytes = 2*(NumBytes/2);
	if(NumBytes == 0) return;
	memmove(m_Data+m_DCSQT[0].m_OffsetToThis+NumBytes,m_Data+m_DCSQT[0].m_OffsetToThis,
		m_SubPicSize - m_DCSQT[0].m_OffsetToThis);
	for(i=0;i<m_DCSQT.size();i++)
	{
		m_DCSQT[i].m_OffsetToThis += NumBytes;
		m_DCSQT[i].m_OffsetToNext += NumBytes;
		PutShort(m_Data,m_DCSQT[i].m_OffsetToThis+2,m_DCSQT[i].m_OffsetToNext);
	}
	// Make sure you update the subpic size, and the offset to commands
	m_SubPicSize += NumBytes;
	PutShort(m_Data,0,m_SubPicSize);
	m_OffsetToCmd += NumBytes;
	PutShort(m_Data,2,m_OffsetToCmd);
	// Also update the remaining space (after commandes) and the space before commands.
	m_RemainingSpace -= NumBytes;
	m_SpaceBeforeCmd += NumBytes;
	int Selected = m_SelectedDCSQT;
	ParseCommands();
	SelectDCSQT(Selected);
}


int CSubPic::VMoveSubPic(int NumPixels)
{
	// Decode subpic if needed.
	if(!m_SubPicDecoded)
	{
		if(DecodeSubpic() == -1)
			return -1;
	}
	// First of, "reset" the subpic, i.e. undo any move that might have been done before...
	m_TopFieldOffset = m_StartOfLine[0];
	m_BotFieldOffset = m_StartOfLine[1];
	m_StartY = m_StartYOrig;
	m_EndY = m_EndYOrig;
	m_SubPicHeight = m_EndY - m_StartY + 1;
	// Make sure NumPixels is even!
	NumPixels = min(NumPixels,m_StartY-m_VerticalBorders+m_TopMostPixel);
	// We want the true bottommost pixel to be above the vertical border, AND the bottom most
	// pixel to be within the frame.
	NumPixels = max(m_StartY+m_TrueBottomLine-m_DisplayHeight+m_VerticalBorders,NumPixels);
	NumPixels = max(m_StartY+m_BottomMostPixel-m_DisplayHeight,NumPixels);

	if(m_ForceEvenMove)
		NumPixels = 2*(NumPixels/2);

	m_LastVMove = NumPixels;
//	if(NumPixels >= m_SubPicHeight || -NumPixels >= m_SubPicHeight)
//		return -1;

	// m_StartY should never go below 2, and should always be even.
	if(NumPixels > 0)
	{
		// Moving subpic up.
		// This is tricky. If m_StartY isn't at the top of the picture (say 1, for example),
		// then we can move the subpic simply by adjusting m_StartY and m_EndY without changing
		// the pointers. We can move that way by NumPix1. To move further, we have to change
		// the pointers. We move that way by NumPix2.
//		int NumPix1 = m_StartY;	// Must be even!
//		int NumPix1 = 2*((m_StartY)/2);	// Must be even!
		int NumPix1 = max(m_StartY-2,0);
//		NumPix1 = min(NumPix1,NumPixels);
		NumPix1 = min(NumPix1,2*(NumPixels/2));
		int NumPix2 = NumPixels-NumPix1;
		NumPix2 = min(NumPix2,m_SubPicHeight-3);	// so m_EndY >= m_StartY+2!
		_ASSERT(NumPix2+1 < m_SubPicHeight);
		// Moving subpics up. Set the offset to the top and bottom fields to skip the top
		// lines, then adjust the endy value...
		// First, find offset to NumPixels line.
		_ASSERT(NumPix2 >= 0);
		if(NumPix2 % 2 == 0)
		{
			m_TopFieldOffset = m_StartOfLine[NumPix2];
			m_BotFieldOffset = m_StartOfLine[NumPix2+1];
		}
		else
		{
			// When moving by an odd number of lines, we need to interchange these two pointers.
			m_TopFieldOffset = m_StartOfLine[NumPix2];
			m_BotFieldOffset = m_StartOfLine[NumPix2+1];
		}
		// Adjust EndY.
		_ASSERT(m_EndY > NumPix2+NumPix1);
		m_EndY -= (NumPix2+NumPix1);
		m_StartY -= NumPix1;
		_ASSERT((m_StartY %2) == 0);
		// Then write all this back!
		WriteDSPXA(m_TopFieldOffset,m_BotFieldOffset);
		WriteDAREA(m_StartX, m_StartY, m_EndX, m_EndY);
		m_SubPicHeight = m_EndY - m_StartY + 1;
	}
	else
	{
		// Moving subpics down. Apparently, we can't just increment StartY, because it seems to always need to be even.
		// So to move down by an odd number, we move down by a larger even number (more negative) then move up by 1
		// pixel.
		int NumPix1 = 2*((NumPixels-1)/2);
		m_StartY -= NumPix1;
		m_EndY -= NumPix1;
		int NumPix2 = NumPixels-NumPix1;
		_ASSERT(NumPix2 == 0 || NumPix2 == 1);
		if(NumPix2 == 1)
		{
			// Moving by an odd number of lines. 
			m_TopFieldOffset = m_StartOfLine[NumPix2];
			m_BotFieldOffset = m_StartOfLine[NumPix2+1];
			m_EndY --;
		}
		// Then write all this back!
		WriteDSPXA(m_TopFieldOffset,m_BotFieldOffset);
		WriteDAREA(m_StartX, m_StartY, m_EndX, m_EndY);
		// m_EndY is *inclusive* so it can't be equal to m_DisplayHeight!
		m_SubPicHeight = min(m_EndY,m_DisplayHeight-1) - m_StartY + 1;
	}
	if(NumPixels)
		m_Modified = 1;
	// This is tricky. m_VerticalAdjust indicate by how many lines the subpic was actually moved,
	// in the coded domain.
	m_VerticalAdjust = -m_LastVMove-(m_StartY-m_StartYOrig);
	return 0;
}

int CSubPic::VMoveSubPicInFrame(int MinTopMostPos, int FlushIt)
{
	// Avoid moving subpics that are actually highlights.
	if(m_NumButtons)
		return 0;
	if(!m_SubPicDecoded)
	{
		if(DecodeSubpic() == -1)
			return -1;
	}
	int BottomMostPixel = m_TrueBottomLine;
	// Bottom margin is the same as the top one.
	int MaxBottomMostPos = m_DisplayHeight - MinTopMostPos;
	m_VerticalBorders = MinTopMostPos;

	// m_StartY does not always indicate the current amount by which the subpics were
	// moved (for example if they were moved up by modifying the pointers).
	// Instead, m_StartYOrig - m_LastVMove tells me the true start Y
	// position, after the last move.
	long CurrentStartY = m_StartYOrig - m_LastVMove;

	// Make sure the subpic is either above our bottom limit or below our top one. If it's violating
	// both limits we can't do anything!
	_ASSERT(CurrentStartY+m_TopMostPixel >= MinTopMostPos || 
		CurrentStartY+BottomMostPixel <= MaxBottomMostPos);

	int MoveBy = 0;
	// Check whether our subpic is below MinTopMostPos.
	if(CurrentStartY+m_TopMostPixel < MinTopMostPos)
	{
		// We need to move this guy down! MoveBy will be negative.
		MoveBy = CurrentStartY+m_TopMostPixel - MinTopMostPos;
		// Check that by moving this guy down, we're not going past the bottom limit.
		if(CurrentStartY+BottomMostPixel-MoveBy > MaxBottomMostPos)
		{
			ERROR2("Can't apply border to subpic %d because it is too high!",m_SPUNumber);
			_ASSERT(0);
			return 0;
		}
	}
	else if(CurrentStartY+BottomMostPixel > MaxBottomMostPos)
	{
		// We need to move this guys up! MoveBy will be positive
		MoveBy = CurrentStartY+BottomMostPixel - MaxBottomMostPos;
		// Check that by moving this guy up, we're not going past the toplimit.
		if(CurrentStartY+m_TopMostPixel-MoveBy < MinTopMostPos)
		{
			ERROR2("Can't apply border to subpic %d because it is too high!",m_SPUNumber);
			_ASSERT(0);
			return 0;
		}
	}
	else if(FlushIt)
	{
		// The subpic is somewhere in the middle but we want to flush it!
		if(CurrentStartY + m_TopMostPixel - MinTopMostPos < MaxBottomMostPos - (CurrentStartY+BottomMostPixel))
			// We're closer to the top, move up by a large amount!
			MoveBy = 800;
		else
			// We're closer to the bottom, move down by a large amount.
			MoveBy = -800;
	}
	VMoveSubPic(m_LastVMove + MoveBy);
	return m_LastVMove + MoveBy;
}

int CSubPic::HMoveSubPic(int NumPixels)
{
	// Reset previous move.
	m_StartX = m_StartXOrig;
	m_EndX = m_EndXOrig;

	NumPixels = max(NumPixels,-m_StartXOrig);
	NumPixels = min(NumPixels,m_DisplayWidth-m_EndXOrig-1);

	m_StartX += NumPixels;
	m_EndX += NumPixels;
	m_LastHMove = NumPixels;
	WriteDAREA(m_StartX, m_StartY, m_EndX, m_EndY);
	if(NumPixels != 0)
		m_Modified = 1;
	return 0;
}

bool CSubPic::CanHMove()
{
	return (m_SubPicWidth < m_DisplayWidth - 10);	
}


int CSubPic::ChangeTransparency(int Transparency, int PixelSelect)
{
	Commands *TheCommand;
	int Start = m_ChangeAllDCSQT ? 0 : m_SelectedDCSQT;
	int End = m_ChangeAllDCSQT ? m_DCSQT.size() : m_SelectedDCSQT+1;
	int DCSQT;
	for(DCSQT=Start;DCSQT<End;DCSQT++)
	{
		SelectDCSQT(DCSQT);
		// Only change the transparency if this DCSQT has a transparency command.
		int Offset = FindCommand(SET_CONTR,DCSQT,&TheCommand);
		if(Offset == -1) continue;
		uchar* Ptr = m_Data+Offset;

		if(PixelSelect < 4)
			m_AlphaPtr[PixelSelect] = Transparency;
		else if(PixelSelect == 4)
		{
			// Decode subpic if needed. We need that for m_BackgroundPix.
			if(!m_SubPicDecoded && DecodeSubpic() == -1)
					return -1;
			// Change the transparency of all pixels, except for the background pixels.
			if(m_BackgroundPix != 0) m_AlphaPtr[0] = Transparency; 
			if(m_BackgroundPix != 1) m_AlphaPtr[1] = Transparency; 
			if(m_BackgroundPix != 2) m_AlphaPtr[2] = Transparency; 
			if(m_BackgroundPix != 3) m_AlphaPtr[3] = Transparency; 
		}
		else
			m_AlphaPtr[0] = m_AlphaPtr[1] = m_AlphaPtr[2] = m_AlphaPtr[3] = Transparency;
		
		// Put that in our data!
		WriteColorAlpha(Ptr+1,m_AlphaPtr);
	}
	m_Modified = true;
	return 0;
}


int CSubPic::GetMeanTransparency()
{
	if(!m_SubPicDecoded) DecodeSubpic();
	int Average = 0;
	int Count = 0;
	if(m_BackgroundPix != 0 && m_NumPix[0] != 0) Average += m_AlphaPtr[0], Count++;
	if(m_BackgroundPix != 1 && m_NumPix[1] != 0) Average += m_AlphaPtr[1], Count++;
	if(m_BackgroundPix != 2 && m_NumPix[2] != 0) Average += m_AlphaPtr[2], Count++;
	if(m_BackgroundPix != 3 && m_NumPix[3] != 0) Average += m_AlphaPtr[3], Count++;
	if(Count == 0) return 0;
	return (int)((Average+.5f)/Count);
}

int CSubPic::ChangeColor(int Color, int PixelSelect)
{
	Commands *TheCommand;
	int Start = m_ChangeAllDCSQT ? 0 : m_SelectedDCSQT;
	int End = m_ChangeAllDCSQT ? m_DCSQT.size() : m_SelectedDCSQT+1;
	int DCSQT;
	for(DCSQT=Start;DCSQT<End;DCSQT++)
	{
		SelectDCSQT(DCSQT);
		// Only change the color if this DCSQT has a color command.
		int Offset = FindCommand(SET_COLOR,DCSQT,&TheCommand);
		if(Offset == -1) continue;
		uchar* Ptr = m_Data+Offset;

		if(m_ColorPtr[PixelSelect] != Color)
			m_Modified = true;
		m_ColorPtr[PixelSelect] = Color;
		// Put that in our data!
		WriteColorAlpha(Ptr+1,m_ColorPtr);
	}
	return 0;
}

int CSubPic::GetCommand(char* Command, int &DCSQT, int &CommandIndex)
{
	if(DCSQT >= m_DCSQT.size())
		return -1;
	SP_DCSQT *TheDCSQT = &m_DCSQT[DCSQT];
	if(CommandIndex == -1)
	{
		if(DCSQT == 0)
			sprintf(Command,"DCSQT: %d\n",DCSQT);
		else
			sprintf(Command,"DCSQT: %d -- Delay %.3fs\n",DCSQT, (float)GetDCSQTDelay(DCSQT)*1e-3f);
		goto end;
	}
	
	Commands TheCommand = TheDCSQT->m_Commands[CommandIndex];
	switch(TheCommand.m_CommandID)
	{
	case FSTA_DSP : sprintf(Command,"  Forced Start\n"); break;
	case STA_DSP : sprintf(Command,"  Start\n"); break;
	case STP_DSP : sprintf(Command,"  Stop\n"); break;
	case SET_COLOR : sprintf(Command,"  Set Color: e2=%d e1=%d p=%d b=%d\n",
						 TheCommand.m_Params[3],TheCommand.m_Params[2],TheCommand.m_Params[1],TheCommand.m_Params[0]); break;
	case SET_CONTR : sprintf(Command,"  Set Transparency: e2=%d e1=%d p=%d b=%d\n",
						 TheCommand.m_Params[3],TheCommand.m_Params[2],TheCommand.m_Params[1],TheCommand.m_Params[0]); break;
	case SET_DAREA : sprintf(Command,"  Set Display Area: sx=%d ex=%d sy=%d ey=%d \n", 
						 TheCommand.m_Params[3],TheCommand.m_Params[2],TheCommand.m_Params[1],TheCommand.m_Params[0]); break;
	case SET_DSPXA : sprintf(Command,"  Set field indexes\n"); break;
	case CHG_COLCON : 
		{
			// JEAN: TOTO: FIX THIS.
			int i,j;
			// Search our array for the CHG_COLCON command that's in this DCSQT.
			for(i=0;i<m_CHCOL.size();i++)
				if(m_CHCOL[i].m_DCSQT == DCSQT) break;
			// There SHOULD be one!
			_ASSERT(i != m_CHCOL.size());
			if(i == m_CHCOL.size())
				break;
			CHG_COLCON_Command *TheCom = &m_CHCOL[i];
			int ret;
			ret = sprintf(Command,"  CHG_COLCON\n\\par  Number of bands %d\n\\par ",TheCom->m_NumBands);
			for(i=0;i<TheCom->m_NumBands;i++)
			{
				// Go through all the bands, and for each one, display the relevant info.
				ret += sprintf(Command+ret,"   %d - Number of columns %d\n\\par ",i,TheCom->m_Bands[i].m_NumPX_CTLI);
				ret += sprintf(Command+ret,"    Top Line %d\n\\par ",TheCom->m_Bands[i].m_CSLN);
				ret += sprintf(Command+ret,"    Bot. Line %d\n\\par ",TheCom->m_Bands[i].m_CTLN);
				for(j=0;j<TheCom->m_Bands[i].m_NumPX_CTLI;j++)
				{
					// Same thing for all the defined columns..
					PX_CTLI* TheColumn = &TheCom->m_Bands[i].m_Columns[j];
					ret += sprintf(Command+ret,"    %d -- Start column %d\n\\par ",j,TheColumn->m_StartColNum);
					ret += sprintf(Command+ret,"     New colors: %d %d %d %d\n\\par ",
						TheColumn->m_NewColors[0],TheColumn->m_NewColors[1],
						TheColumn->m_NewColors[2],TheColumn->m_NewColors[3]);
					ret += sprintf(Command+ret,"     New alphas: %d %d %d %d\n\\par ",
						TheColumn->m_NewAlphas[0],TheColumn->m_NewAlphas[1],
						TheColumn->m_NewAlphas[2],TheColumn->m_NewAlphas[3]);
				}
			}
			break;
		}
	case CMD_END : sprintf(Command,"  End of commands\n"); break;
	}

end:
	CommandIndex++;
	if(CommandIndex >= TheDCSQT->m_Commands.size())
	{
		CommandIndex = -1;
		if(DCSQT < m_DCSQT.size()-1) DCSQT++;
		else
		{
			DCSQT = 0; return -1;
		}
	}
	return 0;
}

void CSubPic::ChangePTS(int NumFrames)
{
	// Just an experiment! Should not be used as they subpics probably won't appear in a player.
	int DeltaPTS = NumFrames * ((VideoFormat == FORMAT_PAL) ? 3600 : 3003);
	DeltaPTS = max(-(int)m_PTS,DeltaPTS);
	m_PTS += DeltaPTS;
	// Write this tack into our little "start of pack" buffer.
	BuildPTS(m_PackStartData+0x17, m_PTS, 0x20);
	m_Modified = true;
}

void CSubPic::SetPTS(int PTS)
{	
	// I need to adjust m_TruePTS as well!
	m_TruePTS += PTS - m_PTS;
	m_PTS = PTS;
	// Write this tack into our little "start of pack" buffer.
	BuildPTS(m_PackStartData+0x17, m_PTS, 0x20);
}

