#include "stdafx.h"
#include "VOBUtils.h"
#include "utils.h"

	// Creates the class.
CVTSLoader::CVTSLoader()
{
	int i;
	for(i=0;i<10;i++)
		m_VobFD[i] = 0;
	m_NumVOBs = 0;
	Reset();
}


CVTSLoader::CVTSLoader(int VTSNum, const char* DirPath)
{
	int i;
	for(i=0;i<10;i++)
		m_VobFD[i] = 0;
	m_NumVOBs = 0;
	Reset();
	LoadVTS(VTSNum,DirPath);
}

CVTSLoader::~CVTSLoader()
{
	Reset();
}

void CVTSLoader::Reset()
{
	int i;

	for(i=0;i<m_NumVOBs;i++)
	{
		if(m_VobFD[i]) fclose(m_VobFD[i]);
		m_VobFD[i] = 0;
	}

	memset(m_VobSizeLBA,0,10*sizeof(int));
	memset(m_VobStartLBA,0,10*sizeof(int));

	m_LoadedOK = 0;
	m_IsReadOnly = 0;
	m_NumVOBs = 0;
	m_WhichVTS = 0;
	m_DirPath[0] = 0;
	m_TitlesetSize = 0;
}

char* CVTSLoader::GetFileName(int Index)
{
	static char FileName[1024];
	if(m_WhichVTS)
		sprintf(FileName,"VTS_%02d_%d.VOB",m_WhichVTS,Index);
	else 
		sprintf(FileName,"VIDEO_TS.VOB");
	return FileName;
}

char* CVTSLoader::GetDomainName(int VTS, int LU)
{
	static char Name[64];
	if(VTS == 0) sprintf(Name,"VMGM");
	else if(LU == 0) sprintf(Name, "VTST %d",VTS);
	else if(LU == -1) sprintf(Name,"VTS %d",VTS);
	else sprintf(Name,"VTSM %d LU %d",VTS,LU);
	return Name;
}

char* CVTSLoader::GetFileName(int VTS, int Index)
{
	static char FileName[1024];
	if(VTS)
		sprintf(FileName,"VTS_%02d_%d.VOB",VTS,Index);
	else if(Index == 0)
		sprintf(FileName,"VIDEO_TS.VOB");
	else
		FileName[0] = 0;
	return FileName;
}


int CVTSLoader::LoadVTS(int VTSNum, const char* DirPath)
{
	Reset();
	if(DirPath)
		strcpy(m_DirPath,IsDir(DirPath)?DirPath : GetDirPath(DirPath));
	else
		strcpy(m_DirPath,".");

	m_WhichVTS = VTSNum;
	// Now, try to load the vob files...
	int i;
	m_VobStartLBA[0]=0;
	for(i=0;i<10;i++)
	{
		m_VobFD[i] = fopen(CatPath(m_DirPath,GetFileName(i+1)),"r+b");
		if(m_VobFD[i] == 0)
		{
			m_VobFD[i] = fopen(CatPath(m_DirPath,GetFileName(i+1)),"rb");
			if(m_VobFD[i] != 0) m_IsReadOnly = true;
			else break;
		}
		m_VobSizeLBA[i] = GetFileSizekB(CatPath(m_DirPath,GetFileName(i+1)))/2;
		m_TitlesetSize += m_VobSizeLBA[i];
		if(i > 0)
			m_VobStartLBA[i] = m_VobStartLBA[i-1]+m_VobSizeLBA[i-1];
	}
	// Couldn't load anything? return -1.
	if(i==0) return -1;
	m_NumVOBs = i;

	return 0;
}

int CVTSLoader::GetIndexForLBA(int SectorLBA)
{
	int i;
	for(i=0;i<m_NumVOBs;i++)
	{
		if(SectorLBA >= m_VobStartLBA[i] && 
			SectorLBA < m_VobStartLBA[i]+m_VobSizeLBA[i]) break;
	}
//	_ASSERT(i<m_NumVOBs);
	if(i >= m_NumVOBs)
		return -1;
	else return i+1;
}


FILE* CVTSLoader::GetFdForLBA(int SectorLBA, bool Seek, int* RelativeLBA, int* RemainingLBAToEnd)
{
	// Determine which VOB the sector points to;
	int i = GetIndexForLBA(SectorLBA)-1;
	if(i < 0) return 0;
	if(Seek)
	{
		int Ret = fseek(m_VobFD[i],(SectorLBA-m_VobStartLBA[i])*SECTOR_SIZE,SEEK_SET);
		_ASSERT(Ret == 0);
	}
	if(RelativeLBA) *RelativeLBA = SectorLBA-m_VobStartLBA[i];
	if(RemainingLBAToEnd) *RemainingLBAToEnd = m_VobStartLBA[i]+m_VobSizeLBA[i] - SectorLBA;
	return m_VobFD[i];
}

int CVTSLoader::GetVobSize(int SectorLBA)
{
	int i = GetIndexForLBA(SectorLBA)-1;
	if(i < 0) return 0;
	return m_VobSizeLBA[i];
}

int CVTSLoader::GetVobStartLBA(int SectorLBA)
{
	int i = GetIndexForLBA(SectorLBA)-1;
	if(i < 0) return 0;
	return m_VobStartLBA[i];
}
