#ifndef VOBUTILS_H
#define VOBUTILS_H

#include <stdio.h>
#include "includes.h"

class CVTSLoader
{
public: 

	// Creates the class.
	CVTSLoader();
	CVTSLoader(int VTSNum, const char* DirPath = 0);
	~CVTSLoader();

	// Closes everything down, and resets the class.
	void		Reset();


	// This loads a given VTS, using the specified path to the VIDEO_TS directory. The current
	// directory is used if DirPath = 0. Return 0 if everything went well.
	int			LoadVTS(int VTSNum, const char* DirPath = 0);

	// This returns the file index for a given LBA. Index 1 corresponds to VTS_xx_1.VOB.
	int			GetIndexForLBA(int SectorLBA);
	// This returns the file descriptor to be used to access the given LBA relative to the start of the
	// domain (absolute lba). Upon return, RelativeLBA contains the LBA relative to the start of the file.
	FILE*		GetFdForLBA(int AbsSectorLBA, bool Seek = true, int* RelativeLBA = 0, int* RemainingLBAToEnd = 0);
	int			GetVobStartLBA(int SectorLBA);
	int			GetVobSize(int SectorLBA);

	// This returns a string that contains the file name for the given VTS and the given index.
	// For example, GetFileName(3) for VTS=1 yields VTS_01_3.VOB;
	char*		GetFileName(int Index);
	static char*	GetFileName(int VTS, int Index);
	// This returns the name of the domain (e.g., VMGM, VTSM etc).
	static char* GetDomainName(int VTS, int LU);

	bool		m_LoadedOK;					// Was the domain loaded OK?
	bool		m_IsReadOnly;				// Are any of the VOBs in the domain read-only?
	int			m_WhichVTS;					// Which VTS did we open?
	int			m_NumVOBs;					// How many vobs in the domain (vts_xx_1.VOB and up).

	char		m_DirPath[1024];
	FILE*		m_VobFD[10];				// File descriptors for all the VOBs.
	int			m_VobStartLBA[10];			// Start LBA of each VOBs (cumulative!).
	int			m_VobSizeLBA[10];			// Size of each VOBs.
	int			m_TitlesetSize;
};


#endif